// perl.cpp : implementation file
//


#include <string>
#include <map>
#include <iostream>
#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>
#include "eperl.h"
#ifdef WIN32
#include "../nvs/nvperl.h"
#else
#include "nvperl.h"
#endif


using namespace std;

#define RESULT_BUFFER_LENGTH	1024

EXTERN_C void xs_init (pTHX);
//EXTERN_C void boot_DynaLoader (pTHX_ CV* cv);
EXTERN_C XS(NV_SetErrorMode);
EXTERN_C XS(NV_SetError);
EXTERN_C XS(NV_Command);
EXTERN_C XS(NV_GetSessionId);
EXTERN_C XS(NV_GetClass);
EXTERN_C XS(NV_GetCommand);
EXTERN_C XS(NV_GetInContainer);
EXTERN_C XS(NV_GetOutContainer);
EXTERN_C XS(NV_GetLanguage);
EXTERN_C XS(NV_GetError);
EXTERN_C XS(NV_GetNumParameters);
EXTERN_C XS(NV_GetSource);
EXTERN_C XS(NV_ParameterExist);
EXTERN_C XS(NV_GetParameter);
EXTERN_C XS(NV_GetParameterName);


// Classes declared in this file
class NvPerlInterpretor;
class NvMutex;

NvPerlInterpretor* GetInterpretorFromCallBack();
//void LogPerl(char *Message);

map<void*, void*> Interpretors;

/////////////////////////////////////////////////////////////////////////////
// NvPerlInterpretor class
class NvPerlInterpretor
{
public:		
	NvPerlCommand *m_pCommand;					// Pointer to a Nirva command object.
	PerlInterpreter *perl_session;
	char *ResultBuffer;
	int ResultBufferSize;
	const char *ScriptFile;
	int ErrorModeScript;
	int GotError;

	NvPerlInterpretor();
	virtual ~NvPerlInterpretor();

	int SetBufferSize(int DataSize);

public:		

	// Creates the perl interpretor
	Init();

	// RunScript
	int RunScript();
	
	// Send a command to Nirva
	int Command(const char *CommandString, char **ResBuffer);

	// Set the error information to Nirva
	void SetError(const char *ErrorInfo);

	void GetSessionId(char **ResBuffer);
	void GetClass(char **ResBuffer);
	void GetCommand(char **ResBuffer);
	void GetInContainer(char **ResBuffer);
	void GetOutContainer(char **ResBuffer);
	void GetLanguage(char **ResBuffer);
	void GetError(const char *InfoType, char **ResBuffer);
	int GetNumParameters();
	int GetSource();
	int ParameterExist(const char *ParameterName);
	void GetParameter(const char *ParameterName, char **ResBuffer);
	void GetParameterName(int index, char **ResBuffer);
};

// NvMutex class
// A mutex can just be locked or unlocked
// It allows thread protection of some part of code
class NvMutex
{
private:

	friend class NvEvent;	// NvEvent should access private members
// mutext declaration
#ifdef WIN32
	HANDLE m_mMutex;
#else
	pthread_mutex_t m_mMutex;
#endif

public:
	// standard constructor
	NvMutex();

	// copy constructor
	// this in fact don't do anything but avoid copying of the mutex object handle
	NvMutex(const NvMutex& MutexSrc);

	// destructor
	virtual ~NvMutex();

	// Locks the mutex
	void Lock();

	// Unlocks the mutex
	void Unlock();
	
	// We overload the assignement operator in order to not allow copy of the mutex handle
	const NvMutex& operator = (const NvMutex& MutexSrc);
};

NvMutex ProtectInterpretors;

// NvMutex class

// constructor
NvMutex::NvMutex()
{
#ifdef WIN32
	m_mMutex = CreateMutex(NULL, 0, NULL);
#else
	pthread_mutex_init(&m_mMutex, NULL);
#endif
}


// constructor by copy
// we don't accept to copy a mutex, we create another one
NvMutex::NvMutex(const NvMutex& MutexSrc)
{
#ifdef WIN32
	m_mMutex = CreateMutex(NULL, 0, NULL);
#else
	pthread_mutex_init(&m_mMutex, NULL);
#endif
}


// destructor
NvMutex::~NvMutex()
{
#ifdef WIN32
	ReleaseMutex(m_mMutex);
	CloseHandle(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
	pthread_mutex_destroy(&m_mMutex);
#endif
}


// lock
void NvMutex::Lock()
{
#ifdef WIN32
	WaitForSingleObject(m_mMutex, INFINITE);
#else
	pthread_mutex_lock(&m_mMutex);
#endif
}


// unlock
void NvMutex::Unlock()
{
#ifdef WIN32
	ReleaseMutex(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
#endif
}

// Assignement operator
const NvMutex& NvMutex::operator = (const NvMutex& MutexSrc)
{
	return *this;	// we do nothing
}

// convert a string to upper case
void MakeUpper(string& sString)
{
	for(int i=0; i<sString.length(); i++)
		sString.at(i) = (char)(toupper(sString[i]));
}
// convert a string to lower case
void MakeLower(string& sString)
{
	for(int i=0; i<sString.length(); i++)
		sString.at(i) = (char)(tolower(sString[i]));
}


// make some cleaning on strings
void CleanString(string& StringToClean)
{
	if(!StringToClean.size())
		return;
	
	string::size_type index;
	
	// replace tab with space
	while((index = StringToClean.find('\t')) != string::npos)
		StringToClean.replace(index, 1, 1, ' ');
	// remove line feeds
	while((index = StringToClean.find('\n')) != string::npos)
		StringToClean.erase(index, 1);
	// remove carriage returns
	while((index = StringToClean.find('\r')) != string::npos)
		StringToClean.erase(index, 1);
	// concatenate 2 consecutive spaces
	while((index = StringToClean.find("  ")) != string::npos)
		StringToClean.replace(index, 2, " ");
	// eliminate first space
	if(StringToClean.size() && StringToClean[0] == ' ')
		StringToClean.erase(0, 1);
	// eliminate last space
	if(StringToClean.size() && StringToClean[StringToClean.size()-1] == ' ')
		StringToClean.erase(StringToClean.size()-1, 1);
	
}



/////////////////////////////////////////////////////////////////////////////
// NvPerlInterpretor

NvPerlInterpretor::NvPerlInterpretor()
{
	ResultBufferSize = 0;
	ErrorModeScript = 0;
	GotError = 0;
	perl_session = NULL;
}

NvPerlInterpretor::~NvPerlInterpretor()
{
	if(perl_session != NULL)
	{
//cout << "Tread detached\n";
		ProtectInterpretors.Lock();
//	PerlInterpreter *old_perl = (PerlInterpreter *)Perl_get_context();
//    Perl_set_context(perl_session);
		PERL_SET_THX(perl_session);
		try
		{
			perl_destruct(perl_session);
		}
		catch(...)
		{
			// Do nothing special
		}
//    Perl_set_context(old_perl);
		try
		{
			perl_free(perl_session);
		}
		catch(...)
		{
			// Do nothing special
		}
		ProtectInterpretors.Unlock();
	}

	if(ResultBufferSize)
		delete ResultBuffer;
}
	

// Run the perl script
int NvPerlInterpretor::RunScript()
{
	if(perl_session == NULL)
		return 0;

	if(ScriptFile == NULL)
		return 0;

	if(!strlen(ScriptFile))
		return 0;

	ResultBufferSize = 0;

	if(SetBufferSize(0) == 0)
		return 0;

	int SizeBuffer = strlen(ScriptFile) + 10;
	char *Buffer = new char[SizeBuffer];
	if(Buffer == NULL)
		return 0;
	strcpy(Buffer, ScriptFile);

	char *argv[5];
	argv[0] = "";
	argv[1] = Buffer;
	argv[2] = 0;

	int argc = 2;

	int RetCode = 0;
	bool RetVal = true;

//	ProtectInterpretors.Lock();
//	PerlInterpreter *old_perl = (PerlInterpreter *)Perl_get_context();
//    Perl_set_context(perl_session);
	try
	{
		perl_parse(perl_session, xs_init, argc, argv, (char **)NULL);
	}
	catch(...)
	{
		RetVal = false;
	}

//	ProtectInterpretors.Unlock();

//	PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
	PERL_SET_THX(perl_session);
	if(RetVal)
	{
		try
		{
//			RetCode = perl_run(perl_session);
		}
		catch(...)
		{
			RetVal = false;
		}
	}
//    Perl_set_context(old_perl);

	delete Buffer;

	if(ResultBufferSize)
		delete ResultBuffer;

	ResultBufferSize = 0;

	if(!RetVal)
		return 0;
	if(RetCode != 0)
		return 0;

	if(GotError)
		return 0;

	return 1;
}


// Creates the perl session if not existing
int NvPerlInterpretor::Init()
{
	if(perl_session != NULL)
		return 1;

	ProtectInterpretors.Lock();
	try
	{
		perl_session = perl_alloc();
	}
	catch(...)
	{
	}
	if(perl_session == NULL)
	{
		ProtectInterpretors.Unlock();
		return 0;
	}

//	PerlInterpreter *old_perl = (PerlInterpreter *)Perl_get_context();
//    Perl_set_context(perl_session);
	PERL_SET_THX(perl_session);
	bool RetVal = true;
	try
	{
		perl_construct(perl_session);
	}
	catch(...)
	{
		RetVal = false;
	}

	if(!RetVal)
	{
//		Perl_set_context(old_perl);
		try
		{
			perl_free(perl_session);
		}
		catch(...)
		{
		}
		perl_session = NULL;
		ProtectInterpretors.Unlock();
		return 0;
	}

//cout << "Tread attached\n";
	
	ProtectInterpretors.Unlock();

	return 1;
}

// Increases the output buffer if necessary
// returns 0 if failure
// returns 1 if no change or first allocation
// returns 2 if realloc
int NvPerlInterpretor::SetBufferSize(int DataSize)
{
	if(DataSize < ResultBufferSize)
		return 1;

	if(ResultBufferSize == 0)
	{
		ResultBuffer = new char[RESULT_BUFFER_LENGTH];
		if(ResultBuffer == NULL)
			return 0;
		ResultBufferSize = RESULT_BUFFER_LENGTH;
	}

	if(DataSize < ResultBufferSize)
		return 1;

	char *NewResultBuffer = new char[DataSize+1];
	if(NewResultBuffer == NULL)
		return 0;

	delete ResultBuffer;
	ResultBuffer = NewResultBuffer;
	ResultBufferSize = DataSize+1;

	return 2;
}


// Execute a NIRVA command
int NvPerlInterpretor::Command(const char *CommandString, char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	bool bControlErrorOnScript = false;
	if(ErrorModeScript)
		bControlErrorOnScript = true;
	if(!m_pCommand->Command(CommandString, ResultBuffer, ResultBufferSize, &DataSize, bControlErrorOnScript))
		return 0;
	int ResultSetBufferSize = SetBufferSize(DataSize);
	if(ResultSetBufferSize == 0)
		return 0;	// Error
	if(ResultSetBufferSize == 2)
	{
		// We got a reallocation so we must resend the order
		if(!m_pCommand->Command(CommandString, ResultBuffer, ResultBufferSize, &DataSize, bControlErrorOnScript))
			return 0;
	}
	return 1;
}


// SetError
void NvPerlInterpretor::SetError(const char *ErrorInfo)
{
	m_pCommand->SetError(ErrorInfo);
}


void NvPerlInterpretor::GetSessionId(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetSessionId(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetSessionId(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetClass(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetClass(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetClass(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetCommand(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetCommand(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetCommand(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetInContainer(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetInContainer(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetInContainer(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetOutContainer(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetOutContainer(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetOutContainer(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetLanguage(char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetLanguage(ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetLanguage(ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetError(const char *InfoType, char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetError(InfoType, ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetError(InfoType, ResultBuffer, ResultBufferSize, &DataSize);
}

int NvPerlInterpretor::GetNumParameters()
{
	return m_pCommand->GetNumParameters();
}

int NvPerlInterpretor::GetSource()
{
	return m_pCommand->GetSource();
}

int NvPerlInterpretor::ParameterExist(const char *ParameterName)
{
	if(m_pCommand->ParameterExist(ParameterName))
		return 1;

	return 0;
}

void NvPerlInterpretor::GetParameter(const char *ParameterName, char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetParameter(ParameterName, ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetParameter(ParameterName, ResultBuffer, ResultBufferSize, &DataSize);
}


void NvPerlInterpretor::GetParameterName(int index, char **ResBuffer)
{
	*ResBuffer = ResultBuffer;
	strcpy(ResultBuffer, "");
	int DataSize;
	m_pCommand->GetParameterName(index, ResultBuffer, ResultBufferSize, &DataSize);
	if(SetBufferSize(DataSize) == 2)
		m_pCommand->GetParameterName(index, ResultBuffer, ResultBufferSize, &DataSize);
}



#ifdef WIN32
// Main dll entry point for windows
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
		
	return 1;   // ok
}
#endif


int InitPerlLib()
{
	ExitPerlLib();
	return 1;
}

void ExitPerlLib()
{
	// Removes all the interpretors if not yet made
	ProtectInterpretors.Lock();
	map<void *,void *>::iterator pos;
	NvPerlInterpretor *PerlClass;
	for(pos = Interpretors.begin(); pos != Interpretors.end(); ++pos)
	{
		PerlClass = (NvPerlInterpretor *)(pos->second);
		if(PerlClass == NULL)
			continue;
		try
		{
			if(PerlClass != NULL)
				delete PerlClass;
		}
		catch(...)
		{
		}
	}
	Interpretors.clear();
	ProtectInterpretors.Unlock();

}


void PerlDetachThread()
{
		// we always have a unique perl interpretor by thread
#ifdef WIN32
		void *ThreadHandle = (void *)GetCurrentThreadId();			// Thread handle for window
#else
		void *ThreadHandle = (void *) pthread_self();				// Thread handle for unix
#endif
	ProtectInterpretors.Lock();
	NvPerlInterpretor *PerlClass = (NvPerlInterpretor *)(Interpretors[ThreadHandle]);
	ProtectInterpretors.Unlock();
	if(PerlClass != NULL)
	{
		try
		{
			if(PerlClass != NULL)
				delete PerlClass;
		}
		catch(...)
		{
		}
		ProtectInterpretors.Lock();
		Interpretors.erase(ThreadHandle);
		ProtectInterpretors.Unlock();
	}
}

NVPERLINTERPRETOR NvGetInterpretor()
{
	// We get the interpreter of the current thread if there is one
	NvPerlInterpretor *PerlInterpretor = GetInterpretorFromCallBack();

	// Otherwise, we create a new one
	if(PerlInterpretor == NULL)
	{
		PerlInterpretor = new NvPerlInterpretor;
		if(PerlInterpretor == NULL)
			return (NVPERLINTERPRETOR)NULL;
		if(!PerlInterpretor->Init())
		{
			delete PerlInterpretor;
			return (NVPERLINTERPRETOR)NULL;
		}
#ifdef WIN32
		void *ThreadHandle = (void *)GetCurrentThreadId();			// Thread handle for window
#else
		void *ThreadHandle = (void *) pthread_self();				// Thread handle for unix
#endif
		ProtectInterpretors.Lock();
		Interpretors[ThreadHandle] = (void *)PerlInterpretor;
		ProtectInterpretors.Unlock();
	}

	return (NVPERLINTERPRETOR)PerlInterpretor;
}


void NvFreeInterpretor(NVPERLINTERPRETOR Interpretor)
{
	// THis function doesn't do anything more
	// because we keep the interpreter for all the thread duration
	// So the job is made in PerlDetachThread
	return;
	NvPerlInterpretor *PerlInterpretor = (NvPerlInterpretor *)Interpretor;
	try
	{
		if(PerlInterpretor != NULL)
			delete PerlInterpretor;
	}
	catch(...)
	{
	}
}

int NvInterpretorRunScript(NVPERLINTERPRETOR Interpretor, const char* ScriptFileName, void *pCommand)
{
	if(ScriptFileName == NULL)
		return 0;
	NvPerlInterpretor *PerlInterpretor = (NvPerlInterpretor *)Interpretor;
	PerlInterpretor->m_pCommand = (NvPerlCommand*)pCommand;
	PerlInterpretor->ScriptFile = ScriptFileName;
	int RetVal = 1;
	try
	{
		return PerlInterpretor->RunScript();
	}
	catch(...)
	{
		RetVal = 0;
	}

	return RetVal;
}





// Called by perl after initialization
EXTERN_C void
xs_init(pTHX)
{
	char *file = __FILE__;
	dXSUB_SYS;


	// DynaLoader is a special case
//	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
	
	// We must install here our callbacks
	newXS("NV::Command", NV_Command, file);
	newXS("NV::SetError", NV_SetError, file);
	newXS("NV::SetErrorMode", NV_SetErrorMode, file);
	newXS("NV::GetSessionId", NV_GetSessionId, file);
	newXS("NV::GetClass", NV_GetClass, file);
	newXS("NV::GetCommand", NV_GetCommand, file);
	newXS("NV::GetInContainer", NV_GetInContainer, file);
	newXS("NV::GetOutContainer", NV_GetOutContainer, file);
	newXS("NV::GetLanguage", NV_GetLanguage, file);
	newXS("NV::GetError", NV_GetError, file);
	newXS("NV::GetNumParameters", NV_GetNumParameters, file);
	newXS("NV::GetSource", NV_GetSource, file);
	newXS("NV::ParameterExist", NV_ParameterExist, file);
	newXS("NV::GetParameter", NV_GetParameter, file);
	newXS("NV::GetParameterName", NV_GetParameterName, file);
}

NvPerlInterpretor* GetInterpretorFromCallBack()
{
	// We get back the pointer to the NvPerlInterpretor class that called the perl script
#ifdef WIN32
	void *ThreadHandle = (void *)GetCurrentThreadId();			// Thread handle for window
#else
	void *ThreadHandle = (void *) pthread_self();				// Thread handle for unix
#endif

	ProtectInterpretors.Lock();
	NvPerlInterpretor *PerlClass = (NvPerlInterpretor *)(Interpretors[ThreadHandle]);
	ProtectInterpretors.Unlock();

	if(PerlClass != NULL)
	{
		int ResultBufferSizeHere;
		// Try something to test if the PerClass is OK or not
		try
		{
			ResultBufferSizeHere = PerlClass->ResultBufferSize;
		}
		catch(...)
		{
			PerlClass = NULL;
		}
	}

	return PerlClass;
}


// Callback for NIRVA Command function
XS(NV_Command)
{
    dXSARGS;
	char *CommandString;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	CommandString = (char *)SvPV(ST(0), PL_na);
	if(CommandString == NULL)
	{
		if(!PerlClass->ErrorModeScript)
		{
			PerlClass->GotError = 1;
			croak("");
		}
		XSRETURN(0);
	}
	

	char *ResBuffer;

	int RetVal = 1;
	try
	{
		RetVal = PerlClass->Command(CommandString, &ResBuffer);
	}
	catch(...)
	{
		RetVal = 0;
	}
	if(!RetVal)
	{
		if(!PerlClass->ErrorModeScript)
		{
			PerlClass->GotError = 1;
			croak("");
		}
		XSRETURN(RetVal);
	}

	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(RetVal);
}



// Callback for NIRVA SetError function
XS(NV_SetError)
{
    dXSARGS;
	char *ErrorInfo;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);


	// Get the parameter
	ErrorInfo = (char *)SvPV(ST(0), PL_na);
	if(ErrorInfo == NULL)
		XSRETURN(0);

	int RetVal = 1;
	try
	{
		PerlClass->SetError(ErrorInfo);
	}
	catch(...)
	{
		XSRETURN(0);
	}


	XSRETURN(RetVal);
}






// Sets the error mode "SCRIPT" for control to scrip and "NIRVA" for control by NIRVA
XS(NV_SetErrorMode)
{
    dXSARGS;
	char *ErrorMode;

	int ErrorModeScriptHere = 0;

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	if(items > 0)
	{
		ErrorMode = (char *)SvPV(ST(0), PL_na);
		if(ErrorMode != NULL)
		{
			string ErrorModeHere = ErrorMode;
			MakeUpper(ErrorModeHere);
			CleanString(ErrorModeHere);
			if(ErrorModeHere == "SCRIPT")
				ErrorModeScriptHere = 1;
		}
	}

	PerlClass->ErrorModeScript = ErrorModeScriptHere;

	XSRETURN(1);
}




// Callback for NIRVA GetSessionId function
XS(NV_GetSessionId)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetSessionId(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}




// Callback for NIRVA GetClass function
XS(NV_GetClass)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetClass(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}




// Callback for NIRVA GetCommand function
XS(NV_GetCommand)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetCommand(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}




// Callback for NIRVA GetInContainer function
XS(NV_GetInContainer)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetInContainer(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}




// Callback for NIRVA GetOutContainer function
XS(NV_GetOutContainer)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetOutContainer(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}





// Callback for NIRVA GetLanguage function
XS(NV_GetLanguage)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetLanguage(&ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}


// Callback for NIRVA GetError function
XS(NV_GetError)
{
    dXSARGS;
	char *InfoType;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	InfoType = (char *)SvPV(ST(0), PL_na);
	if(InfoType == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetError(InfoType, &ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}


// Callback for NIRVA GetNumParameters function
XS(NV_GetNumParameters)
{
    dXSARGS;


	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);


	int RetVal = 0;

	try
	{
		RetVal = PerlClass->GetNumParameters();
	}
	catch(...)
	{
	}
	

	XSRETURN(RetVal);
}


// Callback for NIRVA GetSource function
XS(NV_GetSource)
{
    dXSARGS;

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	
	int SourceType = -1;

	try
	{
		SourceType = PerlClass->GetSource();
	}
	catch(...)
	{
	}


	string sResult = "UNKNOWN";
	switch(SourceType)
	{
		case NV_SOURCE_CLIENT:
			sResult = "CLIENT";
		case NV_SOURCE_BROWSER:
			sResult = "BROWSER";
		case NV_SOURCE_PROCEDURE:
			sResult = "PROCEDURE";
		case NV_SOURCE_SERVICE:
			sResult = "SERVICE";
		case NV_SOURCE_SERVER:
			sResult = "SERVER";
	}

	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, sResult.c_str());
	

	XSRETURN(1);
}




// Callback for NIRVA ParameterExist function
XS(NV_ParameterExist)
{
    dXSARGS;
	char *ParameterName;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	ParameterName = (char *)SvPV(ST(0), PL_na);
	if(ParameterName == NULL)
		XSRETURN(0);
	
	int RetVal = 0;

	try
	{
		RetVal = PerlClass->ParameterExist(ParameterName);
	}
	catch(...)
	{
	}



	XSRETURN(RetVal);
}



// Callback for NIRVA GetParameter function
XS(NV_GetParameter)
{
    dXSARGS;
	char *ParameterName;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	ParameterName = (char *)SvPV(ST(0), PL_na);
	if(ParameterName == NULL)
		XSRETURN(0);
	

	char *ResBuffer;

	try
	{
		PerlClass->GetParameter(ParameterName, &ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}



// Callback for NIRVA GetParameterName function
XS(NV_GetParameterName)
{
    dXSARGS;
	int ParamIndex = -1;

	if(items != 1)
		XSRETURN(0);

	NvPerlInterpretor *PerlClass = GetInterpretorFromCallBack();
	if(PerlClass == NULL)
		XSRETURN(0);

	// Get the parameter
	ParamIndex = SvIV(ST(0));
	

	char *ResBuffer;

	try
	{
		PerlClass->GetParameterName(ParamIndex, &ResBuffer);
	}
	catch(...)
	{
	}


	// Here is the way to set a string variable to the perl script
	SV* SVResult = get_sv("NV::RESULT", TRUE);
	sv_setpv(SVResult, ResBuffer);
	

	XSRETURN(1);
}


/*
void LogPerl(char *Message)
{
	strcat(Message, "\n");
	char PerlLogFile[256];
	strcpy(PerlLogFile, "c:\\perl.log");
	FILE *stream = fopen(PerlLogFile, "at");
	if(stream != NULL)
	{
		fwrite(Message, 1, strlen(Message), stream);
		fclose(stream);	  
	}
}
*/

