// ***************************************************************************************
// File: listener.h
// Description: manage nirva listeners
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: A Nirva listener is an object that runs a set of threads, each of
//           them calling a procedure in loop with a defined sleep time between
//           occurences of the loop.
//           A single thread takes one Nirva session that is opened when the
//           thread is created and closed when the thread ends.
//           Each application has its own listener list
// ***************************************************************************************

#ifndef _LISTENER_H
#define _LISTENER_H

#include "task.h"
#include "util.h"

// Classes declared in this file
class NvListenerThread;
class NvListener;
class NvListenerList;



class NvListenerThread : public NvTask
{
public:

	std::string SessionId;				// Current thread session id
	NvListener *m_lListener;			// Attached listener
	bool b_Initialized;
	bool RequestToLeave;				// Request for stopping thread

	NvListenerThread();


// Overridable callbacks
protected:
	virtual int Main();
	virtual void OnException(std::string sInfo);
	virtual bool OnInit();
	virtual void OnCleanup();

public:
	
	// Stops the listener
	void Stop();
	NvSession* GetSession();
};






class NvListener : public NvObject
{
public:
	std::list<NvListenerThread *> m_vTheads;	// listener threads
	int m_iNumThreads;							// Number of threads that listen

	std::string Name;						// Listener name
	std::string Ident;						// Listener identifier
	std::string Description;				// Listener description
	std::string ProcName;					// Listener procedure name
	std::string ProcNameFailed;				// Listener failed procedure name
	std::string UserName;					// User name
	std::string OpenProc;					// Session open procedure
	std::string CloseProc;					// Session close procedure
	std::string Language;				// Language used when creating the listener
	int m_iSleepTime;						// Sleep time
	bool m_bRunning;						// Running or stopped flag
	bool m_bEnable;							// Enable flag
	NvListenerList *ListenerList;			// List that owns this object

	NvListener();
	virtual ~NvListener();

public:
	// Start listening
	bool Start();
	
	// Stop listening
	bool Stop(bool Force = false, int WaitTime = 10);

	// Get running state
	bool IsRunning();
};


class NvListenerList : public NvObject
{
protected:
	NvListener* GetListener(const char *Name);

public:
	std::map<std::string, NvListener*> m_mListeners;	// Listener list
	NvApplication *Application;							// Application that owns the list

	NvListenerList();
	virtual ~NvListenerList();

	static void CleanListenerName(std::string& ListenerName);
	static std::string GetListenerIdent(std::string ListenerName);

	// Removes all Listeners from the list
	void Clear();
	
	// Load the Listener list from registry
	void Load();

	// Saves the Listener list to the system registry
	bool Save();
	
	// Creates a new Listener
	bool CreateListener(const char *Name, const char *Description = NULL, const char *ProcName = NULL, const char *ProcNameFailed = NULL, const char *UserName = NULL, const char *OpenProc = NULL, const char *CloseProc = NULL, int NumThreads = 1, int SleepTime = 100, bool bEnable = false, const char *Language = NULL);
	
	// Set Listener parameters
	bool SetListenerParam(const char *Name, const char *Description = NULL, const char *ProcName = NULL, const char *ProcNameFailed = NULL, const char *UserName = NULL, const char *OpenProc = NULL, const char *CloseProc = NULL, int NumThreads = -1, int SleepTime = -1);

	// Removes a Listener
	bool RemoveListener(const char *Name);

	// Start a Listener
	bool StartListener(const char *Name);

	// Stop a Listener
	bool StopListener(const char *Name, bool Force = false, int WaitTime = 10);

	// Enable a Listener
	bool EnableListener(const char *Name);

	// Disable a Listener
	bool DisableListener(const char *Name);

	// Test the Listener running state
	bool IsListenerRunning(const char *Name);

	// Get Listener list
	bool ListenerList(const char *Name, NvObjTable *ListenerInfo);

	// Start all enabled Listeners
	bool StartListeners();

	// Stop all started Listeners
	bool StopListeners();

	// Test if a Listener exist
	bool ListenerExist(const char *Name);
};





#endif //_LISTENER_H
