// ***************************************************************************************
// File: mqint.cpp
// Description: interface from nvs to websphere mq
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: the real mq interface is implemented in a dedicated library
//           named nvmq. This file is the layer that interfaces to this
//           library.
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvMqLib

NvMqLib::NvMqLib()
{
	hLibMq = NULL;
	MqLibInit = NULL;
	MqLibFree = NULL;
	MqOpen = NULL;
	MqClose = NULL;
	MqGet = NULL;
	MqPut = NULL;
	m_bAvailable = false;
}

NvMqLib::~NvMqLib()
{
	// Unload the mq interface library if loaded
	FreeMqLib();
}


bool NvMqLib::LoadMqLib(const char *NirvaDir, bool UseLocking)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectMq);
	if(hLibMq != NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return true;
	}

#ifdef WIN32
	// For windows we first check if these dlls are accessible
	// in order to avoid a message box if not existing
	if(MyLoadLibrary("imqb23vn.dll") == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return false;
	}
	if(MyLoadLibrary("imqc23vn.dll") == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return false;
	}

#endif

	// Load the library here
	string MqLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(MqLibPath);
	MqLibPath += "Bin";
	AddDirSuffix(MqLibPath);
	MqLibPath += "nvmq";
#ifdef WIN32
#ifdef NV_DEBUG
	MqLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(MqLibPath);
	MqLibPath += "Bin";
	AddDirSuffix(MqLibPath);
	MqLibPath += "nvmqd";
	//MqLibPath = GetMain()->GetNirvaDir();
	//AddDirSuffix(MqLibPath);
	//MqLibPath += "Sources";
	//AddDirSuffix(MqLibPath);
	//MqLibPath += "nvmq";
	//AddDirSuffix(MqLibPath);
	//MqLibPath += "Debug";
	//AddDirSuffix(MqLibPath);
	//MqLibPath += "nvmq";
#endif
	MqLibPath += ".dll";
#endif
#ifdef LINUX
	MqLibPath += ".so";
#endif
#ifdef HPUX
	MqLibPath += ".sl";
#endif
#ifdef HPUXI
	MqLibPath += ".so";
#endif
#ifdef SOLARIS
	MqLibPath += ".so";
#endif
#ifdef AIX
	MqLibPath += ".a";
#endif

	hLibMq = MyLoadLibrary(MqLibPath.c_str());
	if(hLibMq == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return false;
	}

	MqLibInit = (INITMQLIB)MyGetLibFunction(hLibMq, "InitMqLib");
	MqLibFree = (EXITMQLIB)MyGetLibFunction(hLibMq, "ExitMqLib");
	MqOpen = (MQ_OPEN)MyGetLibFunction(hLibMq, "NvMqOpen");
	MqClose = (MQ_CLOSE)MyGetLibFunction(hLibMq, "NvMqClose");
	MqGet = (MQ_GET)MyGetLibFunction(hLibMq, "NvMqGet");
	MqPut = (MQ_PUT)MyGetLibFunction(hLibMq, "NvMqPut");
	if((MqLibInit == NULL) || (MqLibFree == NULL) || (MqOpen == NULL) || (MqClose == NULL) || (MqGet == NULL) || (MqPut == NULL))
	{
		MyFreeLibrary(hLibMq);
		MqLibInit = NULL;
		MqLibFree = NULL;
		MqOpen = NULL;
		MqClose = NULL;
		MqGet = NULL;
		MqPut = NULL;
		hLibMq = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return false;
	}

	if((MqLibInit)(NirvaDir) == 0)
	{
		MyFreeLibrary(hLibMq);
		MqLibInit = NULL;
		MqLibFree = NULL;
		MqOpen = NULL;
		MqClose = NULL;
		MqGet = NULL;
		MqPut = NULL;
		hLibMq = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return false;
	}


	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectMq);

	return true;
}



void NvMqLib::FreeMqLib(bool UseLocking)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectMq);
	if(hLibMq == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectMq);
		return;
	}

	(MqLibFree)();
	MyFreeLibrary(hLibMq);
	MqLibInit = NULL;
	MqLibFree = NULL;
	MqOpen = NULL;
	MqClose = NULL;
	MqGet = NULL;
	MqPut = NULL;
	hLibMq = NULL;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectMq);
}



bool NvMqLib::IsMqLibLoaded(bool UseLocking)
{
	bool RetVal = true;
	if(UseLocking)
		NV_LOCKLOCK(&ProtectMq);
	if(hLibMq == NULL)
		RetVal = false;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectMq);
	return RetVal;
}

	



MyNvMqInMsg::MyNvMqInMsg()
{
	Data = NULL;
	Clear();
}



MyNvMqInMsg::~MyNvMqInMsg()
{
	Clear();
}



void MyNvMqInMsg::Clear()
{
	Reply = false;
	CopyCorrelationId = false;
	CopyMessageId = false;
	if(Data != NULL)
		free(Data);
	Data = NULL;
	DataSize = 0;
	memset(MessageId, 0, sizeof(MessageId));
	memset(CorrelationId, 0, sizeof(CorrelationId));
	Parameters = "";
	Persistence = NVMQ_PERSISTENCE_NONE;
	ReplyQueueName = "";
	Type = NVMQ_MSG_UNKNOWN;
	Expiry = NVMQ_EXPIRY_NONE;
}


char* MyNvMqInMsg::SetData(const char* Buffer, int BufferSize)
{
	if(Data != NULL)
		free(Data);
	Data = NULL;
	DataSize = 0;
	if(BufferSize > 0)
	{
		Data = (char *)malloc(BufferSize);
		if(Data != NULL)
		{
			memcpy(Data, Buffer, BufferSize);
			DataSize = BufferSize;
		}
	}
	return Data;
}


void MyNvMqInMsg::SetMessageId(const unsigned char* Buffer, int BufferSize)
{
	memset(MessageId, 0, sizeof(MessageId));
	int SizeHere = BufferSize;
	if(SizeHere > sizeof(MessageId))
		SizeHere = sizeof(MessageId);
	memcpy(MessageId, Buffer, SizeHere);
}


void MyNvMqInMsg::SetCorrelationId(const unsigned char* Buffer, int BufferSize)
{
	memset(CorrelationId, 0, sizeof(CorrelationId));
	int SizeHere = BufferSize;
	if(SizeHere > sizeof(CorrelationId))
		SizeHere = sizeof(CorrelationId);
	memcpy(CorrelationId, Buffer, SizeHere);
}



MyNvMqOutMsg::MyNvMqOutMsg()
{
	Data = NULL;
	Clear();
}



MyNvMqOutMsg::~MyNvMqOutMsg()
{
	Clear();
}



void MyNvMqOutMsg::Clear()
{
	CopyCorrelationId = false;
	CopyMessageId = false;
	WithMessageHeader = true;
	if(Data != NULL)
		free(Data);
	Data = NULL;
	DataSize = 0;
	memset(InputMessageId, 0, sizeof(InputMessageId));
	memset(InputCorrelationId, 0, sizeof(InputCorrelationId));
	Persistence = NVMQ_PERSISTENCE_NONE;
	QueueName = "";
	Type = NVMQ_MSG_UNKNOWN;
	ErrorCode = 0;
	ErrorService = "";
	ErrorClass = "";
	ErrorInfo = "";
	ErrorDesc = "";
	Expiry = NVMQ_EXPIRY_NONE;
}



char* MyNvMqOutMsg::SetData(const char* Buffer, int BufferSize)
{
	if(Data != NULL)
		free(Data);
	Data = NULL;
	DataSize = 0;
	if(BufferSize > 0)
	{
		Data = (char *)malloc(BufferSize);
		if(Data != NULL)
		{
			memcpy(Data, Buffer, BufferSize);
			DataSize = BufferSize;
		}
	}
	return Data;
}


void MyNvMqOutMsg::SetInputMessageId(const unsigned char* Buffer, int BufferSize)
{
	memset(InputMessageId, 0, sizeof(InputMessageId));
	int SizeHere = BufferSize;
	if(SizeHere > sizeof(InputMessageId))
		SizeHere = sizeof(InputMessageId);
	memcpy(InputMessageId, Buffer, SizeHere);
}


void MyNvMqOutMsg::SetInputCorrelationId(const unsigned char* Buffer, int BufferSize)
{
	memset(InputCorrelationId, 0, sizeof(InputCorrelationId));
	int SizeHere = BufferSize;
	if(SizeHere > sizeof(InputCorrelationId))
		SizeHere = sizeof(InputCorrelationId);
	memcpy(InputCorrelationId, Buffer, SizeHere);
}


void MyNvMqOutMsg::Init(MyNvMqInMsg *InMsg)
{
	Clear();
	CopyMessageId = InMsg->CopyMessageId;
	CopyCorrelationId = InMsg->CopyCorrelationId;
	memcpy(InputMessageId, InMsg->MessageId, sizeof(InMsg->MessageId));
	memcpy(InputCorrelationId, InMsg->CorrelationId, sizeof(InMsg->MessageId));
	Persistence = InMsg->Persistence;
	QueueName = InMsg->ReplyQueueName;
	Type = InMsg->Type;
	Expiry = InMsg->Expiry;
}


NvMqThread::NvMqThread()
{
	b_Initialized = false;
	RequestToLeave = false;
	MqIdent = NULL;
}



// Client thread main function
int NvMqThread::Main()
{
	bool LocalRequestToLeave = false;
	RequestToLeave = false;
	while(1)
	{
		NV_LOCKOBJECT(this);
		LocalRequestToLeave = RequestToLeave;
		NV_UNLOCKOBJECT(this);
		if(LocalRequestToLeave)
			break;	// We must leave
		
		// we try to read some input for a maximum of 2 seconds
		int Result = (MqLib->MqGet)(MqIdent, 2, &(InputMessage));
		if(Result < 0)
			break;

		if(Result == 1)	// something has been read
		{
			// we leave if requested
			NV_LOCKOBJECT(this);
			LocalRequestToLeave = RequestToLeave;
			NV_UNLOCKOBJECT(this);
			if(LocalRequestToLeave)
				break;	// We must leave

			// we clear the output message
			OutputMessage.Clear();

			bool bReply = InputMessage.Reply;

			// we initalize the output message from the input message
			// if necessary
			if(bReply)
				OutputMessage.Init(&InputMessage);

			// we process the message
			ProcessMessage();

			InputMessage.Clear();

			// We eventually send the output to client
			if(bReply)
				(MqLib->MqPut)(MqIdent, &(OutputMessage));

			// we clear the output message
			OutputMessage.Clear();
		}

	}
	return 0;
}



// called to stop the server
// must be called from another thread
void NvMqThread::Stop()
{
	NV_LOCKOBJECT(this);
	RequestToLeave = true;
	NV_UNLOCKOBJECT(this);
}


// called when an exception occurs (see class NvTask)
void NvMqThread::OnException(string sInfo)
{
//	throw this;
}



// called before the Main function (see class NvTask)
bool NvMqThread::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_MQ);

	char sThreadId[20];
	sprintf(sThreadId, "%d", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type MQ", NV_LOG_INFO, 1, sThreadId, "system");

	SetSession(NULL);
	RemoveSessionOnCleanup = false;

	// Initialize the random generator for this thread
#ifdef WIN32
	int ThreadHandle = GetCurrentThreadId();						// Thread handle for window
#else
	pthread_t ThreadHandle = pthread_self();					// Thread handle for unix
#endif

	unsigned int randinit = (unsigned int)ThreadHandle;
	randinit *= (unsigned int)time(NULL);
	srand(randinit);

	// Open the queue;
	if(!Connect())
	{
		if(!m_bSameThread)
			GetMain()->Logs.Write("Threads", "Error init thread (Connect queue)", "Type MQ", NV_LOG_ERROR, 1, sThreadId, "system");
		return false;
	}

	b_Initialized = true;
	
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type MQ", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}


// called after the Main function (see class NvTask)
void NvMqThread::OnCleanup()
{
	char sThreadId[20];
	sprintf(sThreadId, "%d", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type MQ", NV_LOG_INFO, 1, sThreadId, "system");

	if(b_Initialized)
	{
		// Close the queue
		Disconnect();
	
		// Cleanup the actual command
		Command.Clear();

		// Free the session if not made
		NvSession *Session = GetSession();
		if(Session != NULL)
		{
			if(Session->ActualCommand != NULL)
			{
				// The session is in a command so it's probably a crash
				if(Session->ActualCommand->m_sService != "SYSTEM")
					GetMain()->Services.FreeService(Session->ActualCommand->m_sService.c_str());
				Session->ActualCommand = NULL;
				Session->OriginalCommand = NULL;
				Session->InContainer = NULL;
				Session->OutContainer = NULL;
				Session->SetActualCommand("");
				Session->SetOriginalCommand("");
				NV_UNLOCKOBJECT(Session);
			}

			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			// We remove the session if requested
			if(RemoveSessionOnCleanup)
				GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		}

		SetSession(NULL);
		RemoveSessionOnCleanup = false;

		// Detach thread from java interface if not yet made
		if(!m_bSameThread)
			GetMain()->JavaInterface.DetachJavaThread();
	}

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type MQ", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}




bool NvMqThread::Connect()
{
	// Open the queue;
	MqIdent = (MqLib->MqOpen)(QueueManagerName.c_str(), QueueName.c_str(), Channel.c_str(), Server.c_str(), Filter.c_str());
	if(MqIdent == NULL)
		return false;
	return true;
}


void NvMqThread::Disconnect()
{
	// Disconnect the queue;
	if(MqIdent != NULL)
		(MqLib->MqClose)(MqIdent);
	MqIdent = NULL;
}



// Sets the actual session
void NvMqThread::SetSession(NvSession *Session)
{
	NV_LOCKOBJECT(this);
	m_sSession = Session;
	NV_UNLOCKOBJECT(this);
}

	
// Gets the actual session
NvSession* NvMqThread::GetSession()
{
	NvSession *RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sSession;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



void NvMqThread::ProcessMessage()
{
	// PM TODO
	RemoveSessionOnCleanup = false;

	// Cleanup the actual command
	Command.Clear();

	// Get command parameters
	Command.ParseParameters(InputMessage.Parameters.c_str());

	if(InputMessage.Type == NVMQ_MSG_WEBS)
		Command.bWebs = true;

	int iXmlCoding = XML_MODEL_SIMPLE;
	string NvXmlModel = Command.Parameters.GetValue("NV_XML_MODEL");
	CleanString(NvXmlModel, true);
	MakeUpper(NvXmlModel);
	string sNvSimpleCoding = Command.Parameters.GetValue("NV_XML_SIMPLE");
	CleanString(sNvSimpleCoding, true);
	MakeUpper(sNvSimpleCoding);
	if(NvXmlModel.length())
	{
		if(NvXmlModel == "SIMPLE")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(NvXmlModel == "NORMAL")
			iXmlCoding = XML_MODEL_NORMAL;
		if(NvXmlModel == "COMPACT")
			iXmlCoding = XML_MODEL_COMPACT;
		if(NvXmlModel == "TINY")
			iXmlCoding = XML_MODEL_TINY;
	}
	else
	{
		if(sNvSimpleCoding == "YES")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(sNvSimpleCoding == "NO")
			iXmlCoding = XML_MODEL_NORMAL;
	}

	if(InputMessage.Type == NVMQ_MSG_WEBS)
		iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services

	if(InputMessage.Type == NVMQ_MSG_UNKNOWN)
	{
		// Bad message type
		Command.Error.SetError("SYSTEM", "MQ", NV_MQ_BAD_MESSAGE_TYPE);
		SetErrorCommand(iXmlCoding);
		return;
	}

	// if we have SOAP input, we must extract the body part that
	// becomes the real XML data input
	NvObjBinary SoapCommandHeader;
	if((InputMessage.Type == NVMQ_MSG_SOAP) || (InputMessage.Type == NVMQ_MSG_WEBS))
	{
		if(!GetSoapCommandHeader(&SoapCommandHeader))
		{
			Command.Error.SetError("SYSTEM", "SOAP", NV_SOAP_ERROR_PARSING_MESSAGE);
			SetErrorCommand(iXmlCoding);
		}
		// Get SoapBody should extract the Soap message body
		if(!GetSoapBody())
		{
			Command.Error.SetError("SYSTEM", "SOAP", NV_SOAP_ERROR_PARSING_MESSAGE);
			SetErrorCommand(iXmlCoding);
		}
	}

	// If we have input XML or SOAP, we get the eventual command parameters from it
	// They override the command parameters found on the message in parameters
	bool bParseXmlSuccess = true;
	if((InputMessage.Type == NVMQ_MSG_XML) || (InputMessage.Type == NVMQ_MSG_SOAP) || (InputMessage.Type == NVMQ_MSG_WEBS))
	{
		// For XML input, we must populate the input container with the XML data
		// The ownership of input data is transfered from http part to the command
		Command.DataSize = InputMessage.DataSize;
		Command.Data = InputMessage.Data;
		Command.DataFile = "";
		Command.bDeleteDataFile = false;
		InputMessage.DataSize = 0;
		InputMessage.Data = NULL;

		Command.iXmlInputType = XML_IN_XML;
		if(InputMessage.Type == NVMQ_MSG_SOAP)
			Command.iXmlInputType = XML_IN_SOAP;
		if(InputMessage.Type == NVMQ_MSG_WEBS)
			Command.iXmlInputType = XML_IN_SOAP;
		Command.iXmlCoding = iXmlCoding;

		// We must first do the eventual XSLT transformation of the input
		// For that, we need at least the application name and the
		// NV_XML_XSL_IN parameter in the URL
		string XslIn = Command.Parameters.GetValue("NV_XML_XSL_IN");
		CleanString(XslIn);
		if(InputMessage.Type == NVMQ_MSG_SOAP)
			XslIn = "";		// This parameter has no meaning in the SOAP connector
		if(InputMessage.Type == NVMQ_MSG_WEBS)
			XslIn = "";		// This parameter has no meaning in the WEBS connector
		if(XslIn.length())
		{
			// We must get the application name
			string AppliNameHere = "";
			string::size_type index = XslIn.find(":");
			if(index != string::npos)
			{
				AppliNameHere = XslIn.substr(0, index);
				CleanString(AppliNameHere);
				XslIn = XslIn.substr(index+1);
				CleanString(XslIn);
			}

			if(!AppliNameHere.length())
			{
				if(Command.Parameters.KeyExist("NV_APPLICATION"))
					AppliNameHere = Command.Parameters.GetValue("NV_APPLICATION");
				CleanString(AppliNameHere);
			}
			MakeUpper(AppliNameHere);
			if(!AppliNameHere.length())
				AppliNameHere = "NVDEF";

			// Ok we have now the application name

			NvSession TempSession;
			bParseXmlSuccess = TempSession.XsltInputTransform(&Command, XslIn, AppliNameHere);
		}
		if(bParseXmlSuccess)
			bParseXmlSuccess = ParseXmlCommand();
		// The ParseXmlCommand may automatically select the XML model
		iXmlCoding = Command.iXmlCoding;
	}

	if((bParseXmlSuccess) && (SoapCommandHeader.GetSize()))
	{
		NvXml XmlParser;
		NvObjIndStringList Parameters;
		bParseXmlSuccess = XmlParser.GetSoapHeaderCommandParam(&Parameters, SoapCommandHeader.m_cdata, SoapCommandHeader.m_isize);
		map<string,string>::iterator pos;
		for(pos = Parameters.m_mList.begin(); pos != Parameters.m_mList.end(); ++pos)
			Command.Parameters.SetValue(pos->first, pos->second);
	}

	if(InputMessage.Type == NVMQ_MSG_XML)
	{
		if(Command.Parameters.KeyExist("NV_XML_NSREF"))
		{
			string NsPrefix = Command.Parameters.GetValue("NV_XML_NSPREFIX");
			CleanString(NsPrefix, true);
			string NsRef = Command.Parameters.GetValue("NV_XML_NSREF");
			CleanString(NsRef, true);
			if(NsRef.length())
			{
				Command.m_sXmlNameSpace = NsPrefix;
				if(NsPrefix.length())
					Command.m_sXmlNameSpace += "=";
				Command.m_sXmlNameSpace += NsRef;
			}
			else
				Command.m_sXmlNameSpace = "";
		}
	}

	// We search that again because the parameter may come from the XML data itself
	NvXmlModel = Command.Parameters.GetValue("NV_XML_MODEL");
	CleanString(NvXmlModel, true);
	MakeUpper(NvXmlModel);
	sNvSimpleCoding = Command.Parameters.GetValue("NV_XML_SIMPLE");
	CleanString(sNvSimpleCoding, true);
	MakeUpper(sNvSimpleCoding);
	if(NvXmlModel.length())
	{
		if(NvXmlModel == "SIMPLE")
			iXmlCoding = XML_MODEL_SIMPLE;
		if(NvXmlModel == "NORMAL")
			iXmlCoding = XML_MODEL_NORMAL;
		if(NvXmlModel == "COMPACT")
			iXmlCoding = XML_MODEL_COMPACT;
		if(NvXmlModel == "TINY")
			iXmlCoding = XML_MODEL_TINY;

		if(InputMessage.Type == NVMQ_MSG_WEBS)
			iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services
		Command.iXmlCoding = iXmlCoding;
	}
	else
	{
		if(sNvSimpleCoding.length())
		{
			if(sNvSimpleCoding == "YES")
				iXmlCoding = XML_MODEL_SIMPLE;
			if(sNvSimpleCoding == "NO")
				iXmlCoding = XML_MODEL_NORMAL;
			if(InputMessage.Type == NVMQ_MSG_WEBS)
				iXmlCoding = XML_MODEL_SIMPLE;			// Always use the simple coding for web services

			Command.iXmlCoding = iXmlCoding;
		}
	}

	// Get the language from command (useful here for error messages)
	string LanguageHere = Command.Parameters.GetValue("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
		LanguageHere = "ENGLISH";
	Command.m_sLanguage = LanguageHere;

	if(!bParseXmlSuccess)
	{
		if(Command.Error.m_iErrorCode == 0)
			Command.Error.SetError("SYSTEM", "XML", NV_XML_ERROR_PARSING_COMMAND);
		SetErrorCommand(iXmlCoding);
		return;
	}


	// Get the session ID from the parameters 
	string SessionIdHere = Command.Parameters.GetValue("NV_SESSION_ID");
	CleanString(SessionIdHere);

	// Get the no error flag
	string NoError = Command.Parameters.GetValue("NV_NO_ERROR");
	MakeUpper(NoError);
	CleanString(NoError);
	bool bNoError = false;
	if(NoError == "YES")
		bNoError = true;;
	
	// And try to get the session
	NvSession *Session = GetMain()->Sessions.GetSession(SessionIdHere.c_str());
	SetSession(Session);

	bool SessionCreated = false;
	if(Session == NULL)
	{
		// First test if requested to create a new session if the actual one is expired
		string NewIfExpired = Command.Parameters.GetValue("NV_NEW_IF_EXPIRED");
		CleanString(NewIfExpired);
		MakeUpper(NewIfExpired);
		if(NewIfExpired == "YES")
			SessionIdHere = "";

		if(SessionIdHere.length())
		{
			// The session has probably expired
			Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_FIND);
			SetErrorCommand(iXmlCoding);
			return;
		}

		// The requested session is blank (or not given)
		// So we try to create it
		if(GetMain()->CheckOverTime())
		{
			Command.Error.SetError("SYSTEM", "SYSTEM", NV_SYSTEM_OVER_TIME);
			SetErrorCommand(iXmlCoding);
			return;
		}
		Session = GetMain()->Sessions.GetSession(NULL);
		SetSession(Session);
		if(Session == NULL)
		{
			int MaxSessions = GetMain()->GetMaxSessions();
			if(MaxSessions != -1)
			{
				if(GetMain()->Sessions.GetNumSessions() >= MaxSessions)
				{
					string AppliNameHere = Command.Parameters.GetValue("NV_APPLICATION");
					CleanString(AppliNameHere);
					MakeUpper(AppliNameHere);
					if(!AppliNameHere.length())
						AppliNameHere = "NVDEF";
					string UserHere = Command.Parameters.GetValue("NV_USER");
					CleanString(UserHere);
					if(!UserHere.length())
						UserHere = "nvdef";
					MakeLower(UserHere);
					if((AppliNameHere == "NVDEF") && (UserHere == "nvadmin"))
					{
						Session = GetMain()->Sessions.GetSession(NULL, false);
						SetSession(Session);
						if(Session != NULL)
						{
							// This session will only be authorized to list and remove sessions
							Session->m_bOverMaxSession = true;
//							SetErrorCommand(iXmlCoding);
						}
					}
					else
					{
						Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_TOO_MANY);
						SetErrorCommand(iXmlCoding);
					}
				}
				else
				{
					Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
					SetErrorCommand(iXmlCoding);
				}
			}
			else
			{
				Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTCREATE);
				SetErrorCommand(iXmlCoding);
			}
		}
		if(Session == NULL)
			return;

		SessionCreated = true;

		// The session creation was OK
		// Now, we must queuein to it
		string AppliNameHere = Command.Parameters.GetValue("NV_APPLICATION");
		CleanString(AppliNameHere);
		if(!AppliNameHere.length())
			AppliNameHere = "NVDEF";
		string UserHere = Command.Parameters.GetValue("NV_USER");
		CleanString(UserHere);
		if(!UserHere.length())
			UserHere = "nvdef";
		string PasswordHere = Command.Parameters.GetValue("NV_PASSWORD");
		CleanString(PasswordHere);
		bool bWithNewPassword = false;
		if(Command.Parameters.KeyExist("NV_NEW_PASSWORD"))
			bWithNewPassword = true;
		string NewPasswordHere = Command.Parameters.GetValue("NV_NEW_PASSWORD");
		CleanString(NewPasswordHere);
		string NewPasswordHereConfirm = Command.Parameters.GetValue("NV_NEW_PASSWORD_CONFIRM");
		CleanString(NewPasswordHereConfirm);
		bool bWithNewPasswordConfirm = false;
		if(bWithNewPassword)
		{
			if(Command.Parameters.KeyExist("NV_NEW_PASSWORD_CONFIRM"))
				bWithNewPasswordConfirm = true;
		}
		string Hash = AppliNameHere + UserHere + PasswordHere;
		Hash = MD5_string(Hash);
		MakeUpper(Hash);
		string sSessionTimeOut = Command.Parameters.GetValue("NV_TIMEOUT");
		CleanString(sSessionTimeOut);
		int iSessionTimeOut = atoi(sSessionTimeOut.c_str());
		if(iSessionTimeOut < 0)
			iSessionTimeOut = 0;
		string OpenProc = Command.Parameters.GetValue("NV_SESSION_OPEN");
		CleanString(OpenProc);
		if(!OpenProc.length())
			OpenProc = "session_open";
		if(OpenProc == "NV_SESSION_OPEN_NONE")
			OpenProc = "";
		Session->m_sOpenProc = OpenProc;
		string CloseProc = Command.Parameters.GetValue("NV_SESSION_CLOSE");
		CleanString(CloseProc);
		if(!CloseProc.length())
			CloseProc = "session_close";
		if(CloseProc == "NV_SESSION_CLOSE_NONE")
			CloseProc = "";
		Session->m_sCloseProc = CloseProc;
		GetMain()->AttachRequestList.Get(Session, ATTACH_REQUEST_TYPE_MQ, AppliNameHere);
		int LoginRetVal = Session->Login(UserHere.c_str(), PasswordHere.c_str(), AppliNameHere.c_str(), SESSION_TYPE_CLIENT, NewPasswordHere.c_str(), NewPasswordHereConfirm.c_str(), bWithNewPassword, bWithNewPasswordConfirm, iSessionTimeOut, Hash.c_str(), LanguageHere.c_str(), &Command.Parameters.m_mList, &Command);
		if(LoginRetVal != 1)
		{
			// Login failed
			// We prepare error information
			string Info = "user=";
			if(UserHere.length())
				Info += UserHere;
			else
				Info += "(Default user)";
			Info += " - application=";
			if(AppliNameHere.length())
				Info += AppliNameHere;
			else
				Info += "(Default application)";

			if(LoginRetVal < 0)
			{
				if(LoginRetVal == -1)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_USER, Info);
				if(LoginRetVal == -2)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CANNOT_GET_CONTEXT, Info);
				if(LoginRetVal == -3)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD, Info);
				if(LoginRetVal == -4)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_APPLICATION, Info);
				if(LoginRetVal == -5)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_CHANGE_PASSWORD, Info);
				if(LoginRetVal == -6)
					Command.Error.SetError("SYSTEM", "SECURITY", NV_SECURITY_BAD_PASSWORD_SYNTAX, Info);
			}
			else
				Command.Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOTLOGIN, Info);
			
			SetErrorCommand(iXmlCoding);

			// We free the session
			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

			// And remove it
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

			Session = NULL;
			SetSession(Session);

			return;
		}
		string UserHereForCheck = UserHere;
		CleanString(UserHereForCheck, true);
		MakeLower(UserHereForCheck);
		if(Session != NULL)
			if(UserHereForCheck == "nvadmin")
				Session->m_bAdministrator = true; // Administrator user
	}

	string CommandHere = Command.Parameters.GetValue("NV_CMD");
	string ClassHere = "";
	string ServiceHere = "";
	if(CommandHere.length())
	{
		string::size_type index = CommandHere.rfind(':');
		if(index != string::npos)
		{
			ClassHere = CommandHere.substr(0, index);
			CommandHere = CommandHere.substr(index+1);
			index = ClassHere.rfind(':');
			if(index != string::npos)
			{
				ServiceHere = ClassHere.substr(0, index);
				ClassHere = ClassHere.substr(index+1);
			}
		}
	}
	else
	{
		ServiceHere = Command.Parameters.GetValue("NV_SERVICE");
		ClassHere = Command.Parameters.GetValue("NV_CLASS");
		CommandHere = Command.Parameters.GetValue("NV_COMMAND");
	}
	CleanString(ServiceHere, true);
	MakeUpper(ServiceHere);
	CleanString(ClassHere, true);
	MakeUpper(ClassHere);
	CleanString(CommandHere, true);
	MakeUpper(CommandHere);
	if((!ServiceHere.length()) && (!ClassHere.length()) && (!CommandHere.length()))
	{
		ServiceHere = "SYSTEM";
		ClassHere = "MISC";
		CommandHere = "NOP";
	}
	if(!ServiceHere.length())
		ServiceHere = "SYSTEM";
	if(!ClassHere.length())
		ClassHere = "MISC";
	if(!CommandHere.length())
		CommandHere = "NOP";

	if(InputMessage.Type == NVMQ_MSG_WEBS)
	{
		// For web services, the command is always WEBS EXECUTE with a parameter named OPERATION that
		ServiceHere = "SYSTEM";
		ClassHere = "WEBSERVICE";
		CommandHere = "EXECUTE";
		string Operation = Command.Parameters.GetValue("OPERATION");
		CleanString(Operation);
		MakeLower(Operation);
		if(!Operation.length())
			Operation = "nvdef";
		Command.Parameters.SetValue("OPERATION", Operation);

		string WebService = Command.Parameters.GetValue("WEBSERVICE");
		CleanString(WebService);
		MakeLower(WebService);
		if(!WebService.length())
			WebService = "nvdef";
		Command.Parameters.SetValue("WEBSERVICE", WebService);
	}

	// Get the input container name from command
	string InContainerHere = Command.Parameters.GetValue("NV_CONTAINER");
	string OutContainerHere = InContainerHere;

	// Get the input container name from command
	if(Command.Parameters.KeyExist("NV_IN_CONTAINER"))
		InContainerHere = Command.Parameters.GetValue("NV_IN_CONTAINER");
	MakeLower(InContainerHere);
	CleanString(InContainerHere);

	// Get the output container name from command
	if(Command.Parameters.KeyExist("NV_OUT_CONTAINER"))
		OutContainerHere = Command.Parameters.GetValue("NV_OUT_CONTAINER");
	MakeLower(OutContainerHere);
	CleanString(OutContainerHere);

	if(!InContainerHere.length())
		InContainerHere = NV_CONTAINER_DEFAULT_NAME;
	if(!OutContainerHere.length())
		OutContainerHere = NV_CONTAINER_DEFAULT_NAME;

	// Get the close session from command
	string CloseSessionCommand = Command.Parameters.GetValue("NV_CLOSE_SESSION");
	MakeUpper(CloseSessionCommand);
	CleanString(CloseSessionCommand);
	bool bCloseSessionFromCommand = false;
	if(SessionCreated)
		bCloseSessionFromCommand = true;	// default is yes for MQ connector except if a session ID has been given
	if(CloseSessionCommand == "YES")
		bCloseSessionFromCommand = true;;
	if(CloseSessionCommand == "NO")
		bCloseSessionFromCommand = false;;

	// Get the language from command again because now we may have the session
	// that gives the default language
	LanguageHere = Command.Parameters.GetValue("NV_LANGUAGE");
	MakeUpper(LanguageHere);
	CleanString(LanguageHere);
	if(!LanguageHere.length())
	{
		if(Session != NULL)
			LanguageHere = Session->m_sLanguage;
		else
			LanguageHere = "ENGLISH";
	}

	// Now we can prepare the command
	Command.m_iCommandSource = NV_SOURCE_BROWSER;
	Command.m_sService = ServiceHere;
	Command.m_sClass = ClassHere;
	Command.m_sCommand = CommandHere;
	Command.m_sInContainer = InContainerHere;
	Command.m_sOutContainer = OutContainerHere;
//	if(!Command.m_sOutContainer.length())
//		Command.m_sOutContainer = Command.m_sInContainer;
	Command.m_sHost = GetMain()->GetHostWebs();
	Command.m_sLanguage = LanguageHere;


	Command.Parameters.Remove("NV_SERVICE");
	Command.Parameters.Remove("NV_CLASS");
	Command.Parameters.Remove("NV_COMMAND");
	Command.Parameters.Remove("NV_CMD");
	Command.Parameters.Remove("NV_IN_CONTAINER");
	Command.Parameters.Remove("NV_OUT_CONTAINER");
	Command.Parameters.Remove("NV_LANGUAGE");
	Command.Parameters.Remove("NV_CLOSE_SESSION");
	Command.Parameters.Remove("NV_NO_ERROR");

	
	if((InputMessage.Type == NVMQ_MSG_XML) || (InputMessage.Type == NVMQ_MSG_SOAP) || (InputMessage.Type == NVMQ_MSG_WEBS))
	{
		Command.iXmlInputType = XML_IN_XML;
		if(InputMessage.Type == NVMQ_MSG_SOAP)
			Command.iXmlInputType = XML_IN_SOAP;
		if(InputMessage.Type == NVMQ_MSG_WEBS)
			Command.iXmlInputType = XML_IN_SOAP;
		Command.iXmlCoding = iXmlCoding;

		Command.m_sInputCharset = "UTF-8";

	}

	if((InputMessage.Type == NVMQ_MSG_BINARY) || (InputMessage.Type == NVMQ_MSG_FILE))
	{
		
		// Create the file or binary object here

		// First get object name
		string ObjName = Command.Parameters.GetValue("NV_MQ_OBJNAME");
		CleanString(ObjName, true);
		MakeLower(ObjName);
		if(!ObjName.length())
			ObjName = NV_OBJ_DEFAULT_NAME;
		MakeLower(ObjName);
		NvContainer *InContainer = Session->GetContainer(Command.m_sInContainer.c_str(), true, &Command);

		NvObjFile *Object = NULL;
		if(InContainer != NULL)
		{
			InContainer->RemoveObject(ObjName.c_str());
			if(InputMessage.Type == NVMQ_MSG_FILE)
			{
				NvObjFile *Object = NULL;
				InContainer->CreateObject(ObjName.c_str(), NV_OBJTYPE_FILE);
				string TargetDir = Session->Application->GetWorkDir();
				Object = (NvObjFile *)(InContainer->GetObject(ObjName.c_str()));
				if(Object != NULL)
					if(!Object->SetAutoFileName(TargetDir))
						InContainer->RemoveObject(ObjName.c_str());
				if(InputMessage.DataSize)
					Object->Populate(InputMessage.Data, InputMessage.DataSize);	
			}
			if(InputMessage.Type == NVMQ_MSG_BINARY)
			{
				NvObjBinary *Object = NULL;
				InContainer->CreateObject(ObjName.c_str(), NV_OBJTYPE_BINARY);
				Object = (NvObjBinary *)(InContainer->GetObject(ObjName.c_str()));
				if(InputMessage.DataSize)
					Object->Populate(InputMessage.Data, InputMessage.DataSize);	
			}
		}

		// We generate XML output for this message
		OutputMessage.Type = NVMQ_MSG_XML;

		// And the default model for output XML is the simple one
		if(!sNvSimpleCoding.length() && !NvXmlModel.length())
			Command.iXmlCoding = XML_MODEL_SIMPLE;
		// We don't need the input data any more
		if(InputMessage.DataSize)
		{
			free(InputMessage.Data);
			InputMessage.Data = NULL;
			InputMessage.DataSize = 0;
		}
		Session->FreeContainer(InContainer, Command.m_sInContainer.c_str(), true, &Command);
	}
	if(InputMessage.Type == NVMQ_MSG_COMMAND)
	{
		// We generate XML output for this message
		OutputMessage.Type = NVMQ_MSG_XML;
	}

	
	// We execute the command
	bool SessionRetVal = true;
	try
	{
		SessionRetVal = Session->Command(&Command);
	}
	catch(...)
	{
		string Message = Command.m_sService;
		Message += ":";
		Message += Command.m_sClass;
		Message += ":";
		Message += Command.m_sCommand;
		Command.Error.SetError("SYSTEM", "COMMAND", NV_COMMAND_EXCEPTION, Message);
		SessionRetVal = false;
	}

	// Get the with header option for MQ output message
	// the default is YES
	OutputMessage.WithMessageHeader = true;
	string WithHeader = Command.Parameters.GetValue("NV_MQ_MSG_HEADER");
	CleanString(WithHeader, true);
	MakeUpper(WithHeader);
	if(WithHeader == "NO")
		OutputMessage.WithMessageHeader = false;

	// Get the optional NV_MQ_OUTPUT parameter that allows
	// to change dynamically the queue output name
	if(InputMessage.Reply && Command.Parameters.KeyExist("NV_MQ_OUTPUT"))
	{
		string NewOutputQueue = Command.Parameters.GetValue("NV_MQ_OUTPUT");
		if(NewOutputQueue.length())
			OutputMessage.QueueName = NewOutputQueue;
	}

	if(SessionRetVal)
	{
		// now, we manage output data
		if(Command.iOutputType == OUTPUT_BINARY)
			OutputMessage.Type = NVMQ_MSG_BINARY;
		if(Command.iOutputType == OUTPUT_FILE)
			OutputMessage.Type = NVMQ_MSG_FILE;

		// The ownership of ouput data is transfered from command to message part
		if(Command.DataFile.length())
		{
			NvObjBinary TmpBinary;
			TmpBinary.LoadFromFile(Command.DataFile.c_str());
			RemoveFile(Command.DataFile.c_str());
			OutputMessage.DataSize = TmpBinary.m_isize;
			OutputMessage.Data = TmpBinary.m_cdata;
			TmpBinary.m_isize = 0;
			TmpBinary.m_cdata = NULL;
			Command.DataFile = "";
			Command.bDeleteDataFile = false;
		}
		else
		{
			OutputMessage.DataSize = Command.DataSize;
			OutputMessage.Data = Command.Data;
			Command.DataSize = 0;
			Command.Data = NULL;
		}
	}
	
	if(bNoError)
		Command.Error.SetError();


	if(Session != NULL)
		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
	
	if((SessionRetVal) && (Command.m_sService == "SYSTEM") && (Command.m_sClass == "SESSION") && (Command.m_sCommand == "CLOSE"))
	{
		// The order is to close the session
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		Session = NULL;
	}
	else if(bCloseSessionFromCommand)
	{
		// We also close the session if explicitly requested as a command parameter
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
		Session = NULL;
	}
//	else if(SessionCreated && Command.Error.m_iErrorCode)
//	{
//		// We also close the session if an error occured and the session has just been created
//		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
//		Session = NULL;
//	}

	Session = NULL;
	SetSession(Session);

	// The error information is generated at the end in order for it to eventually not contain the session ID
	// if closed
	if(Command.Error.m_iErrorCode)
	{
		SetErrorCommand(iXmlCoding);
//		SetHttpError(HTTP_ERROR_INTERNAL, Info);
	}
	
	// Cleanup the actual command
	Command.Clear();

}



void NvMqThread::SetErrorCommand(int XmlCoding)
{
	if(OutputMessage.Data != NULL)
		free(OutputMessage.Data);
	OutputMessage.Data = NULL;
	OutputMessage.DataSize = 0;

	if(!InputMessage.Reply)
		return;		// nothing to do because there is no reply requested
	
	string sEncoding = "";
	if(Command.iXmlInputType)
		sEncoding = Command.m_sInputCharset;
	if(Command.Parameters.KeyExist("NV_XML_ENCODING"))
		sEncoding = Command.Parameters.GetValue("NV_XML_ENCODING");
	CleanString(sEncoding);
	MakeUpper(sEncoding);
	if((sEncoding.length()) && (sEncoding != "ISO-8859-1") && (sEncoding != "UTF-8"))
		sEncoding = "";
	bool bUnicode = false;
	if(!sEncoding.length())
	{
		if(GetMain()->IsUnicode())
			sEncoding = "UTF-8";
		else
			sEncoding = "ISO-8859-1";
	}
	if(sEncoding == "UTF-8")
		bUnicode = true;

	string ErrorPage = "";
	NvXml XmlObjectForCoding;
	if(Command.Parameters.KeyExist("NV_XML_STRICT"))
	{
		string sStrict = Command.Parameters.GetValue("NV_XML_STRICT");
		CleanString(sStrict, true);
		MakeUpper(sStrict);
		if(sStrict == "YES")
			XmlObjectForCoding.bStrict = true;
	}
	OutputMessage.ErrorCode = Command.Error.m_iErrorCode;
	OutputMessage.ErrorService = XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
	OutputMessage.ErrorClass = XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
	OutputMessage.ErrorInfo = XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
	OutputMessage.ErrorDesc = XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);

	if(InputMessage.Type == NVMQ_MSG_XML)
	{
		// Write a simple XML data
		string NsPrefix = "";
		string NsHref = "";
		string NameSpaceHere = Command.m_sXmlNameSpace;
		if(NameSpaceHere.length())
		{
			string::size_type index;
			index = NameSpaceHere.find('=');
			if(index != string::npos)
			{
				if(index)
					NsPrefix = NameSpaceHere.substr(0, index);
				if(index < (NameSpaceHere.length()-1))
					NsHref = NameSpaceHere.substr(index+1);
			}
		}
		ErrorPage = "<?xml version=\"1.0\" encoding=\"";
		ErrorPage += sEncoding;
		ErrorPage += "\" ?>\n";
		ReserveString(ErrorPage, 1000);
		ErrorPage += "<";

		if(NsHref.length() && NsPrefix.length())
		{
			ErrorPage += NsPrefix;
			ErrorPage += ":";
		}
		if(XmlCoding == XML_MODEL_SIMPLE)
			ErrorPage += "nvfault";
		else
			ErrorPage += "NIRVA";
		if(NsHref.length())
		{
			ErrorPage += " xmlns";
			if(NsPrefix.length())
			{
				ErrorPage += ":";
				ErrorPage += NsPrefix;
			}
			ErrorPage += "=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(NsHref, bUnicode);
			ErrorPage += "\"";
		}
		if(NsPrefix.length())
			NsPrefix += ":";
	
		NvSession* SessionHere = GetSession();
		if(SessionHere != NULL)
		{
			ErrorPage += " session=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->m_sIdent, bUnicode);
			ErrorPage += "\" application=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetAppName(), bUnicode);
			ErrorPage += "\" user=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetUserName(), bUnicode);
			ErrorPage += "\"";
		}
		ErrorPage += ">\n";
		if(XmlCoding == XML_MODEL_SIMPLE)
		{
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "code>";
			char Buf[50];
			sprintf(Buf, "%d", Command.Error.m_iErrorCode);
			string sValue = Buf;
			ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "code>\n";
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "service>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "service>\n";
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "class>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "class>\n";
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "description>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "description>\n";
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "info>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "info>\n";
		}
		else
		{
			ErrorPage += "\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERROR>\n";
			ErrorPage += "\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCODE>";
			char Buf[50];
			sprintf(Buf, "%d", Command.Error.m_iErrorCode);
			string sValue = Buf;
			ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCODE>\n";
			ErrorPage += "\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORSERVICE>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORSERVICE>\n";
			ErrorPage += "\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCLASS>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCLASS>\n";
			ErrorPage += "\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORDESC>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORDESC>\n";
			ErrorPage += "\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORINFO>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORINFO>\n";
			ErrorPage += "\t</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERROR>\n";
		}
		ErrorPage += "</";
		ErrorPage += NsPrefix;
		if(XmlCoding == XML_MODEL_SIMPLE)
			ErrorPage += "nvfault>\n";
		else
			ErrorPage += "NIRVA>\n";
		string ErrorXsl = Command.Parameters.GetValue("NV_XML_XSL_ERROR");
		if(ErrorXsl.length())
		{
			// An XSL transformation is required
			string AppliNameHere = "";
			if(SessionHere != NULL)
				AppliNameHere = SessionHere->GetAppName();
			else
			{
				if(Command.Parameters.KeyExist("NV_APPLICATION"))
					AppliNameHere = Command.Parameters.GetValue("NV_APPLICATION");
				CleanString(AppliNameHere);
				MakeUpper(AppliNameHere);
				if(!AppliNameHere.length())
					AppliNameHere = "NVDEF";
			}
			string AppFileDir = "";
			NvApplication *Application = GetMain()->Applications.GetApplication(AppliNameHere.c_str());
			if(Application != NULL)
			{
				string XslFile = "";
				XslFile = Application->GetFileDir();
				GetMain()->Applications.FreeApplication(AppliNameHere.c_str());
				string::size_type index = ErrorXsl.rfind('.');
				if(index == string::npos)
					ErrorXsl += ".xsl";
				XslFile += ErrorXsl;
				if(GetNoCaseRealFileName(XslFile))
				{
					NvXml XmlParser;
					unsigned long OutputDataSize = (unsigned long)(ErrorPage.length());
					char *OutputData = (char*)malloc(OutputDataSize);
					if(OutputData != NULL)
					{
						if(XmlParser.XsltTransform(&OutputData, &OutputDataSize, XslFile.c_str()))
							ErrorPage = OutputData;
						free(OutputData);
					}
				}

			}
		}
	}
	else if((InputMessage.Type == NVMQ_MSG_SOAP) || (InputMessage.Type == NVMQ_MSG_WEBS))
	{
		// Write a simple SOAP XML data
		string NsPrefix = "";
		string NsHref = "";
		string NameSpaceHere = "nvs=http://www.nirva-systems.com/2004-03-29/nirva-xml-fault";
		if(NameSpaceHere.length())
		{
			string::size_type index;
			index = NameSpaceHere.find('=');
			if(index != string::npos)
			{
				if(index)
					NsPrefix = NameSpaceHere.substr(0, index);
				if(index < (NameSpaceHere.length()-1))
					NsHref = NameSpaceHere.substr(index+1);
			}
		}
		
		ErrorPage = "<?xml version=\"1.0\" encoding=\"";
		ErrorPage += sEncoding;
		ErrorPage += "\" ?>\n";
		ReserveString(ErrorPage, 2000);
		ErrorPage += "<env:Envelope xmlns:env=\"http://schemas.xmlsoap.org/soap/envelope/\">\n";
		ErrorPage += "\t<env:Body>\n";
		ErrorPage += "\t\t<env:faultcode>Server</env:faultcode>\n";
		ErrorPage += "\t\t<env:faultstring>";
		ErrorPage += "NIRVA-ERROR:";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
		ErrorPage += ":";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
		ErrorPage += ":";
		char Buf[50];
		sprintf(Buf, "%d", Command.Error.m_iErrorCode);
		string sValue = Buf;
		ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
		ErrorPage += ":";
		ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
		if(Command.Error.m_sErrorInfo.length())
		{
			ErrorPage += ":";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
		}
		ErrorPage += "\t\t</env:faultstring>\n";
		ErrorPage += "\t\t<env:faultactor>";
		ErrorPage += Command.m_sHost;
		ErrorPage += "\t\t</env:faultactor>\n";
		ErrorPage += "\t\t<env:detail>\n";

		ErrorPage += "\t\t\t<";

		if(NsHref.length() && NsPrefix.length())
		{
			ErrorPage += NsPrefix;
			ErrorPage += ":";
		}
		if(XmlCoding == XML_MODEL_SIMPLE)
			ErrorPage += "nvfault";
		else
			ErrorPage += "NVERROR";
		if(NsHref.length())
		{
			ErrorPage += " xmlns";
			if(NsPrefix.length())
			{
				ErrorPage += ":";
				ErrorPage += NsPrefix;
			}
			ErrorPage += "=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(NsHref, bUnicode);
			ErrorPage += "\"";
		}
		if(NsPrefix.length())
			NsPrefix += ":";
	
		NvSession* SessionHere = GetSession();
		if(SessionHere != NULL)
		{
			ErrorPage += " session=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->m_sIdent, bUnicode);
			ErrorPage += "\" application=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetAppName(), bUnicode);
			ErrorPage += "\" user=\"";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(SessionHere->GetUserName(), bUnicode);
			ErrorPage += "\"";
		}
		ErrorPage += ">\n";
		if(XmlCoding == XML_MODEL_SIMPLE)
		{
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "code>";
			sprintf(Buf, "%d", Command.Error.m_iErrorCode);
			sValue = Buf;
			ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "code>\n";
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "service>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "service>\n";
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "class>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "class>\n";
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "description>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "description>\n";
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "info>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "info>\n";
		}
		else
		{
			ErrorPage += "\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERROR>\n";
			ErrorPage += "\t\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCODE>";
			sprintf(Buf, "%d", Command.Error.m_iErrorCode);
			sValue = Buf;
			ErrorPage += XmlObjectForCoding.EncodeOutputString(sValue, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCODE>\n";
			ErrorPage += "\t\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORSERVICE>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorService, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORSERVICE>\n";
			ErrorPage += "\t\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCLASS>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorClass, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORCLASS>\n";
			ErrorPage += "\t\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORDESC>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.GetErrorDesc(Command.m_sLanguage.c_str()), bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORDESC>\n";
			ErrorPage += "\t\t\t\t\t<";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORINFO>";
			ErrorPage += XmlObjectForCoding.EncodeOutputString(Command.Error.m_sErrorInfo, bUnicode);
			ErrorPage += "</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERRORINFO>\n";
			ErrorPage += "\t\t\t\t</";
			ErrorPage += NsPrefix;
			ErrorPage += "NVERROR>\n";
		}
		ErrorPage += "\t\t\t</";
		ErrorPage += NsPrefix;
		if(XmlCoding == XML_MODEL_SIMPLE)
			ErrorPage += "nvfault>\n";
		else
			ErrorPage += "NVERROR>\n";
		ErrorPage += "\t\t</env:detail>\n";
		ErrorPage += "\t</env:Body>\n";
		ErrorPage += "</env:Envelope>\n";
	}

	if(ErrorPage.length())
	{
		OutputMessage.Data = (char *)malloc(ErrorPage.length());
		if(OutputMessage.Data != NULL)
		{
			OutputMessage.DataSize = (int)(ErrorPage.length());
			memcpy(OutputMessage.Data, ErrorPage.c_str(), ErrorPage.length());
		}
	}
}




bool NvMqThread::ParseXmlCommand(NvObjBinary *XmlDataSource)
{
	string::size_type index;
	int inputSize = InputMessage.DataSize;
	char *inputData = InputMessage.Data;
	if(XmlDataSource != NULL)	// Test if the source data is message or XmlDataSource
	{
		inputSize = XmlDataSource->GetSize();
		inputData = XmlDataSource->m_cdata;
	}


	// We read a maximum of 1024 characters for retrieving the <NVCOMMAND> tag
	// Except if the input comes from XmlDataSource (at this time we take all (SOAP headers probably))
	int MaxSizeToRead = inputSize;
	if(XmlDataSource != NULL)
		if(MaxSizeToRead > 1024)
			MaxSizeToRead = 1024;

	if(MaxSizeToRead < 5)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(!XmlData.Populate(inputData, MaxSizeToRead-1))
		return false;
	
	// We must first search for the root tag

	bool bCommandTagFound = false;
	bool bRootFound = false;
	string Root = "NIRVA";
	string CompleteRoot = "NIRVA";
	string CommandTag = "NVCOMMAND";
	string Line;
	string NameSpace = "";
	string StartFlow = "";

	while(XmlData.ReadLine(Line) > 0)
	{
		if(!bRootFound)
		{
			if(!Line.length())
				continue;
			if(Line.length() < 2)
				continue;
			index = Line.find("<?");
			if(index != string::npos)
			{
				StartFlow = Line.substr(index);
				while((index = Line.find(">")) == string::npos)
				{
					if(XmlData.ReadLine(Line) == 0)
						return false;	// Error
					StartFlow += Line;
				}
				index = StartFlow.find(">");
				if(index == string::npos)
					return false;	// Error
				if(index == StartFlow.length() - 1)
					Line = "";
				else
				{
					Line = StartFlow.substr(index+1);
					StartFlow = StartFlow.substr(0, index+1);
				}
				StartFlow += "\n";
			}

			index = Line.find("<");
			if(index != string::npos)
			{
				if(index < Line.length())
				{
					// Probably first tag
					Root = Line.substr(index+1);
					index = Root.find('>');
					if(index != string::npos)
						Root = Root.substr(0, index);
					else
					{
						while(1)
						{
							if(XmlData.ReadLine(Line) == 0)
								return false;
							index = Root.find('>');
							if(index == string::npos)
							{
								Root += Line;
								continue;
							}
							Root += Line.substr(0, index);
							Line = Line.substr(index+1);
							break;
						}
					}
					CompleteRoot = Root;
					index = Root.find('/');
					if(index != string::npos)
						Root = Root.substr(0, index);
					index = Root.find(' ');
					if(index != string::npos)
						Root = Root.substr(0, index);
					index = Root.find('\t');
					if(index != string::npos)
						Root = Root.substr(0, index);
					index = Root.find('\n');
					if(index != string::npos)
						Root = Root.substr(0, index);
					if(Root.length())
					{

						// We are checking now for a namespace qualified root tag
						index = Root.find(':');
						if(index != string::npos)
						{
							NameSpace = Root.substr(0, index+1);
							Root = Root.substr(index+1);
						}
						CleanString(NameSpace, true);
						CleanString(Root, true);
						if(Root.length())
							bRootFound = true;
						else
							NameSpace = "";
					}
				}
			}
		}

		if(!bRootFound)
			continue;
		
		CommandTag = "NVCOMMAND>";
		index = Line.find(CommandTag);
		if(index == string::npos)
		{
			CommandTag = "nvcommand>";
			index = Line.find(CommandTag);
			if(index == string::npos)
				continue;
		}

		// We suspect a NVCOMMAND tag
		// We must search for the start tag
		string PreTag = Line.substr(0, index);
		string::size_type index2 = PreTag.rfind('<');
		if(index2 == string::npos)
			continue;

		if(index2 == index-1)
		{
			// Real NVCOMMAND tag with no namespace
			bCommandTagFound = true;
			break;
		}

/*		// We are checking now for a namespace qualified NVCOMMAND tag
		if(PreTag.length() <= 1)
			continue;

		PreTag = PreTag.substr(index2+1);
		CleanString(PreTag);
		
		// We should not have spaces
		if(PreTag.find(' ') != string::npos)
			continue;

		if(PreTag[PreTag.length()-1] != ':')
			continue;
		
		NameSpace = PreTag;*/
		bCommandTagFound = true;
		break;
	}
	
	Command.m_sXmlNameSpace = NameSpace;

	XmlData.Clear();	// Not used for now

	if(!bCommandTagFound)
	{
		string DataFlow = "<?xml version=\"1.0\" ?>\n<";
		DataFlow += CompleteRoot;
		DataFlow += ">\n";
		DataFlow += "</";
		if(NameSpace.length())
			DataFlow += NameSpace;
		DataFlow += Root;
		DataFlow += ">\n";
		XmlData.Append(DataFlow.c_str(), (int)(DataFlow.length()));
	}
	else
	{
		NvObjBinary JustData;
		StartFlow += "<";
		StartFlow += NameSpace;
		StartFlow += Root;
		if(NameSpace.length())
		{
			StartFlow += " xmlns:";
			StartFlow += NameSpace;
			if(StartFlow.size() && StartFlow[StartFlow.size()-1] == ':')
				StartFlow.erase(StartFlow.size()-1, 1);
			StartFlow += "=\"http://zut.com\"";
		}
		StartFlow += ">\n";
		XmlData.Append(StartFlow.c_str(), (int)(StartFlow.length()));

		// Now we must search for the end tag </NVCOMMAND>
		string StartTagToSearch = "<";
		StartTagToSearch += NameSpace;
		StartTagToSearch += CommandTag;

		string EndTagToSearch = "</";
		EndTagToSearch += NameSpace;
		EndTagToSearch += CommandTag;
		string EndFlow = EndTagToSearch;
		EndFlow += "\n</";
		EndFlow += NameSpace;
		EndFlow += Root;
		EndFlow += ">\n";

		int SizeToRead = 0;
		int Offset = 0;

		JustData.m_cdata = inputData;
		JustData.m_isize = inputSize;
		JustData.m_iposition = 0;

		int NumBytesRead = 0;
		int TotalBytesRead = 0;
		// First search for the start tag
		while(NumBytesRead = JustData.ReadLine(Line))
		{
			TotalBytesRead += NumBytesRead;
			index = Line.find(StartTagToSearch);
			if(index != string::npos)
			{
				Offset = TotalBytesRead-NumBytesRead+(int)index;
				TotalBytesRead = NumBytesRead-(int)index;
				Line = Line.substr(index);
				break;
			}
		}

		// We have found it, we must verify that the end tag is not on the same line
		index = Line.find(EndTagToSearch);
		if(index != string::npos)
		{
			SizeToRead = (int)index;
		}
		else
		{
			while(NumBytesRead = JustData.ReadLine(Line))
			{
				TotalBytesRead += NumBytesRead;
				index = Line.find(EndTagToSearch);
				if(index != string::npos)
				{
					SizeToRead = TotalBytesRead-NumBytesRead+(int)index;
					break;
				}
			}
		}
		JustData.m_cdata = NULL;
		JustData.m_isize = 0;
		JustData.Clear();
		if(SizeToRead)
			if(!JustData.Populate(inputData+Offset, SizeToRead))
				return false;

		if(!SizeToRead)		// Nothing found
			return true;

		XmlData.AppendFrom(&JustData);
		XmlData.Append(EndFlow.c_str(), (int)(EndFlow.length()));
	}
	
	int iInputXmlCoding = XML_MODEL_SIMPLE;
	if(bRootFound && Root == "NIRVA")
		iInputXmlCoding = XML_MODEL_NORMAL;
	if(iInputXmlCoding != Command.iXmlCoding)
	{
		// We must verify if we apply automatic selection of XML output model or not
		string sSimpleModel = Command.Parameters.GetValue("NV_XML_SIMPLE");
		CleanString(sSimpleModel, true);
		MakeUpper(sSimpleModel);
		if((sSimpleModel != "NO") && (sSimpleModel != "YES"))
			Command.iXmlCoding = iInputXmlCoding;	// We apply automatic selection
	}


	// The XML flow is now in the XmlData object
	NvXml XmlParser;
	NvObjIndStringList Parameters;
	string NameSpaceInXml;
	if(!XmlParser.GetCommandParam(&Parameters, XmlData.m_cdata, XmlData.m_isize, NameSpaceInXml))
		return false;

	Command.m_sXmlNameSpace = NameSpaceInXml;

	map<string,string>::iterator pos;
	for(pos = Parameters.m_mList.begin(); pos != Parameters.m_mList.end(); ++pos)
		Command.Parameters.SetValue(pos->first, pos->second);

	return true;
}



// This function reads the SOAP body from the input data and replaces the input data by this
// body content.
bool NvMqThread::GetSoapBody()
{
	int inputSize = InputMessage.DataSize;
	// We read a maximum of 1024 characters for retrieving the <Envelope> tag
	int MaxSizeToRead = inputSize;
	if(MaxSizeToRead > 1024)
		MaxSizeToRead = 1024;

	if(MaxSizeToRead < 10)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(!XmlData.Populate(InputMessage.Data, MaxSizeToRead-1))
		return false;
	
	int index = 0;
	int Offset = 0;
	int indexstarttag = -1;
	string NameSpace = "";
	char *Data = XmlData.m_cdata;
	int DataSize = XmlData.m_isize;
	while(1)
	{
		index = findstring(Data, DataSize-Offset, "Envelope");
		if(index == -1)
			return false;	// not a SOAP message

		// We have found the Envelope string, we must verify that its a real XML tag
		// and get the eventual namespace
		
		index = Offset+index;	// Real index
		Offset = index+1;
		if(index == 0)
			continue;	// The data starts directly with Envelope, this is not SOAP
		if((Data[index-1] != ':') && (Data[index-1] != '<'))
			continue;	// This is not valid tag
		int i=0;
		for(i=index-1; i>0; i--)
		{
			if(Data[i] == ' ')	// No space authorized
				break;
			if(Data[i] == '<')
			{
				indexstarttag = i;
				break;
			}
		}
		if(indexstarttag != -1)
			break;
	}

	if(indexstarttag == -1)
		return false;

	// indexstarttag is the < character for the Envelope tag
	// index points to the start of the Envelope string
	Data[index] = '\0';
	if(Data[index-1] == ':')
	{
		Data[index-1] = '\0';
		NameSpace = Data+indexstarttag+1;
	}

	Data[indexstarttag] = '\0';
	string BeforeEnvelope = Data;

	XmlData.Clear();	// Not used for now

	// We have a SOAP message, we must now find for the SOAP body and extract it
	string StartTagToSearch = "<";
	if(NameSpace.length())
	{
		StartTagToSearch += NameSpace;
		StartTagToSearch += ":";
	}
	StartTagToSearch += "Body";

	string EndTagToSearch = "</";
	if(NameSpace.length())
	{
		EndTagToSearch += NameSpace;
		EndTagToSearch += ":";
	}
	EndTagToSearch += "Body";

	int SizeToRead = 0;
	int OffsetStart = -1;

	// Search the start body tag
	const char *MsgData = (const char*)(InputMessage.Data);
	int MsgDataSize = InputMessage.DataSize;
	int MsgOffset = findstring(MsgData, MsgDataSize, StartTagToSearch.c_str());
	if(MsgOffset == -1)
		return false;

	// Search the end of the start body tag
	int Offset2 = findstring(MsgData+MsgOffset, MsgDataSize-MsgOffset, ">");
	if(Offset2 == -1)
		return false;
	OffsetStart = MsgOffset+Offset2+1;
	if(OffsetStart >= MsgDataSize)
		return false;

	MsgOffset = findstring(MsgData+OffsetStart, MsgDataSize-OffsetStart, EndTagToSearch.c_str());
	if(MsgOffset == -1)
		return false;

	SizeToRead = MsgOffset;
	if(SizeToRead)
	{
		int NewSize = SizeToRead + (int)(BeforeEnvelope.length());
		char *NewData = (char*)malloc(NewSize);
		if(NewData == NULL)
			return false;
		if(BeforeEnvelope.length())
			memcpy(NewData, BeforeEnvelope.c_str(), BeforeEnvelope.length());
		memcpy(NewData+BeforeEnvelope.length(), InputMessage.Data + OffsetStart, SizeToRead);
		free(InputMessage.Data);
		InputMessage.Data = NewData;
		InputMessage.DataSize = NewSize;
	}
	else
	{
		free(InputMessage.Data);
		InputMessage.Data = NULL;
		InputMessage.DataSize = 0;
	}

	return true;
}


bool NvMqThread::GetSoapCommandHeader(NvObjBinary *SoapCommandHeader)
{
	SoapCommandHeader->Clear();
	int inputSize = InputMessage.DataSize;

	// We read a maximum of 1024 characters for retrieving the <Envelope> tag
	int MaxSizeToRead = inputSize;
	if(MaxSizeToRead > 1024)
		MaxSizeToRead = 1024;

	if(MaxSizeToRead < 10)
		return true;	// nothing to do

	NvObjBinary XmlData;
	if(!XmlData.Populate(InputMessage.Data, MaxSizeToRead-1))
		return false;
	
	int index = 0;
	int Offset = 0;
	int indexstarttag = -1;
	string NameSpace = "";
	char *Data = XmlData.m_cdata;
	int DataSize = XmlData.m_isize;
	while(1)
	{
		index = findstring(Data, DataSize-Offset, "Envelope");
		if(index == -1)
			return false;	// not a SOAP message

		// We have found the Envelope string, we must verify that its a real XML tag
		// and get the eventual namespace
		
		index = Offset+index;	// Real index
		Offset = index+1;
		if(index == 0)
			continue;	// The data starts directly with Envelope, this is not SOAP
		if((Data[index-1] != ':') && (Data[index-1] != '<'))
			continue;	// This is not valid tag
		int i=0;
		for(i=index-1; i>0; i--)
		{
			if(Data[i] == ' ')	// No space authorized
				break;
			if(Data[i] == '<')
			{
				indexstarttag = i;
				break;
			}
		}
		if(indexstarttag != -1)
			break;
	}

	if(indexstarttag == -1)
		return false;

	// indexstarttag is the < character for the Envelope tag
	// index points to the start of the Envelope string
	Data[index] = '\0';
	if(Data[index-1] == ':')
	{
		Data[index-1] = '\0';
		NameSpace = Data+indexstarttag+1;
	}

	Data[indexstarttag] = '\0';
	string BeforeEnvelope = Data;

	string SpecialStartRootTag = "<ZutdeZut>\n";
	string SpecialEndRootTag = "</ZutdeZut>\n";

	XmlData.Clear();	// Not used for now

	// We have a SOAP message, we must now find for the SOAP body and extract it
	string StartTagToSearch = "<";
	if(NameSpace.length())
	{
		StartTagToSearch += NameSpace;
		StartTagToSearch += ":";
	}
	StartTagToSearch += "Header";

	string EndTagToSearch = "</";
	if(NameSpace.length())
	{
		EndTagToSearch += NameSpace;
		EndTagToSearch += ":";
	}
	EndTagToSearch += "Header";

	int SizeToRead = 0;
	int OffsetStart = -1;
	// Search the start header tag
	const char *MsgData = (const char*)(InputMessage.Data);
	int MsgDataSize = InputMessage.DataSize;
	int MsgOffset = findstring(MsgData, MsgDataSize, StartTagToSearch.c_str());
	if(MsgOffset == -1)
		return true;	// No header

	// Search the end of the start body tag
	int Offset2 = findstring(MsgData+MsgOffset, MsgDataSize-MsgOffset, ">");
	if(Offset2 == -1)
		return false;
	OffsetStart = MsgOffset+Offset2+1;
	if(OffsetStart >= MsgDataSize)
		return false;

	MsgOffset = findstring(MsgData+OffsetStart, MsgDataSize-OffsetStart, EndTagToSearch.c_str());
	if(MsgOffset == -1)
		return false;

	SizeToRead = MsgOffset;
	if(SizeToRead)
	{
		if(BeforeEnvelope.length())
			SoapCommandHeader->Append(BeforeEnvelope);
		SoapCommandHeader->Append(SpecialStartRootTag);
		SoapCommandHeader->Append(InputMessage.Data + OffsetStart, SizeToRead);
		SoapCommandHeader->Append(SpecialEndRootTag);
	}

	return true;
}

	

NvMq::NvMq()
{
	m_iNumThreads = 1;
	QueueManagerName = "";
	QueueName = "";
	Filter = "";
	Channel = "";
	Server = "";
	m_bRunning = false;
	m_bAutoStart = false;
	MqLib = NULL;
	Name = "";
	Description = "";
	Ident = "";
}
	

NvMq::~NvMq()
{
	Stop(true);
}
	

bool NvMq::Start()
{
	NV_LOCKOBJECT(this);
	if(m_bRunning)
	{
		// Ever running
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(m_iNumThreads == 0)
	{
		// No thread to launch
		m_bRunning = true;
		NV_UNLOCKOBJECT(this);
		return true;
	}

	int i;
	bool bError = false;
	for(i=0; i<m_iNumThreads; i++)
	{
		NvMqThread *MqThread = new NvMqThread;
		if(MqThread == NULL)
		{
			bError = true;
			break;
		}
		MqThread->QueueManagerName = QueueManagerName;
		MqThread->QueueName = QueueName;
		MqThread->Filter = Filter;
		MqThread->Channel = Channel;
		MqThread->Server = Server;
		MqThread->MqLib = MqLib;
		bool bTest = MqThread->Connect();
		MqThread->Disconnect();
		if(!bTest)
		{
			delete MqThread;
			bError = true;
			break;
		}
		if(!MqThread->Run())
		{
			delete MqThread;
			bError = true;
			break;
		}
		m_vTheads.push_back(MqThread);
	}
	if(!bError)
		m_bRunning = true;
	if(bError && m_vTheads.size())
	{
		m_bRunning = true;
		NV_UNLOCKOBJECT(this);
		Stop(true);
		return false;
	}
	NV_UNLOCKOBJECT(this);
	if(bError)
		return false;
	return true;
}

	

bool NvMq::Stop(bool Force, int WaitTime)
{
	NV_LOCKOBJECT(this);
	if(!m_bRunning)
	{
		// not running
		NV_UNLOCKOBJECT(this);
		return true;
	}

	list<NvMqThread *>::iterator pos;
	for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
	{
		NvMqThread *MqThread = *pos;
		if(MqThread == NULL)
			continue;
		MqThread->Stop();
	}
	
	// wait for them to stop
	time_t initial_time = time(NULL);
	bool bAtLeastOneRunning = false;
	while(1)
	{
		SLEEP(200);
		bAtLeastOneRunning = false;
		for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
		{
			NvMqThread *MqThread = *pos;
			if(MqThread == NULL)
				continue;
			if(MqThread->IsThreadRunning())
			{
				bAtLeastOneRunning = true;
				break;
			}
		}
		if(!bAtLeastOneRunning)
			break;
		if((int)(time(NULL) - initial_time) > WaitTime)
			break;
	}
	if(bAtLeastOneRunning && Force)
	{
		// Let's kill hardly
		for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
		{
			NvMqThread *MqThread = *pos;
			if(MqThread == NULL)
				continue;
			if(MqThread->IsThreadRunning())
				MqThread->Kill();
		}
	
		// wait for them to stop
		time_t initial_time = time(NULL);
		bAtLeastOneRunning = false;
		while(1)
		{
			SLEEP(200);
			bAtLeastOneRunning = false;
			for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
			{
				NvMqThread *MqThread = *pos;
				if(MqThread == NULL)
					continue;
				if(MqThread->IsThreadRunning())
				{
					bAtLeastOneRunning = true;
					break;
				}
			}
			if(!bAtLeastOneRunning)
				break;
			if((int)(time(NULL) - initial_time) > WaitTime)
				break;
		}
	}
	
	if(bAtLeastOneRunning)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
	{
		NvMqThread *MqThread = *pos;
		if(MqThread != NULL)
			delete MqThread;
	}

	m_vTheads.clear();
	m_bRunning = false;
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvMq::IsRunning()
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = m_bRunning;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Constructor
NvMqList::NvMqList()
{
}

// Destructor
NvMqList::~NvMqList()
{
	Clear();
}



void NvMqList::CleanQueueName(string& QueueName)
{
	CleanString(QueueName);

	string::size_type index;

	// remove stars
	while((index = QueueName.find('*')) != string::npos)
		QueueName.erase(index, 1);

	// remove slashes
	while((index = QueueName.find('/')) != string::npos)
		QueueName.erase(index, 1);

	// remove backslashes
	while((index = QueueName.find('\\')) != string::npos)
		QueueName.erase(index, 1);
}



string NvMqList::GetQueueIdent(string QueueName)
{
	CleanQueueName(QueueName);
	MakeUpper(QueueName);
	string QueueIdent = MD5_string(QueueName);
	MakeUpper(QueueIdent);
	return QueueIdent;
}



// Removes all queues from the list
void NvMqList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,NvMq *>::iterator pos;
	for(pos = m_mQueues.begin(); pos != m_mQueues.end(); ++pos)
	{
		NvMq *Queue = pos->second;
		if(Queue != NULL)
			delete Queue;
	}
	m_mQueues.clear();
	NV_UNLOCKOBJECT(this);
}


// used to get a queue (only internaly)
NvMq* NvMqList::GetQueue(const char *Name)
{
	NvMq *FoundQueue = NULL;
	string NameHere = Name;
	string IdentHere = GetQueueIdent(NameHere);

	if(m_mQueues.find(IdentHere) != m_mQueues.end())
		FoundQueue = m_mQueues[IdentHere];

	return FoundQueue;
}


void NvMqList::Load()
{
	NvRegistryContainer *Registry;
	Registry = &(GetMain()->SystemRegistry);

	// load the queue table
	// The queue table is stored in a table object named "queues" in the
	// registry key system.queues
	NvContainer Container;
	if(Registry->GetKey("system.queues", &Container, NV_CONTAINER_COPY_NO_APPEND, "queues", false, true))
	{
		NvObj *Object = Container.GetObject("queues");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *QueueTable = (NvObjTable*)Object;
			int index_Name = QueueTable->GetColumnIndex("Name");
			int index_Desc = QueueTable->GetColumnIndex("Description");
			int index_QueueManagerName = QueueTable->GetColumnIndex("QueueManagerName");
			int index_QueueName = QueueTable->GetColumnIndex("QueueName");
			int index_Filter = QueueTable->GetColumnIndex("Filter");
			int index_NumThreads = QueueTable->GetColumnIndex("NumThreads");
			int index_AutoStart = QueueTable->GetColumnIndex("AutoStart");
			int index_Channel = QueueTable->GetColumnIndex("Channel");
			int index_Server = QueueTable->GetColumnIndex("Server");
			int row_index;
			for(row_index=0; row_index<QueueTable->GetNumRow(); row_index++)
			{
				// Get queue parameters
				string Name = QueueTable->GetCellLine(row_index, index_Name);
				NvMqList::CleanQueueName(Name);
				if(!Name.length()) // The queue is identified by its name
					continue;
				if(GetQueue(Name.c_str()) != NULL)	// Check doubles
					continue;
				string Description = QueueTable->GetCellLine(row_index, index_Desc);
				string QueueManagerName = QueueTable->GetCellLine(row_index, index_QueueManagerName);
				string QueueName = QueueTable->GetCellLine(row_index, index_QueueName);
				string Filter = QueueTable->GetCellLine(row_index, index_Filter);
		
				string NumThreads = QueueTable->GetCellLine(row_index, index_NumThreads);
				int iNumThreads = atoi(NumThreads.c_str());
				if(iNumThreads <= 0)
					iNumThreads = 1;

				string AutoStart = QueueTable->GetCellLine(row_index, index_AutoStart);
				int iAutoStart = 0;
				if(AutoStart == "YES")
					iAutoStart = 1;

				string Channel = QueueTable->GetCellLine(row_index, index_Channel);
				string Server = QueueTable->GetCellLine(row_index, index_Server);

				CreateQueue(Name.c_str(), Description.c_str(), QueueManagerName.c_str(), QueueName.c_str(), Filter.c_str(), iNumThreads, iAutoStart, Channel.c_str(), Server.c_str());
			}
		}
		// Resave in case of change
		Save();
	}
}


// Saves the queue list to the system registry
// This function is called internally so its not locked
bool NvMqList::Save()
{
	NvRegistryContainer *Registry;
	Registry = &(GetMain()->SystemRegistry);

	// First create the registry key "system.queues" if it doesn't exist
	Registry->SetKey("system.queues", NULL);

	NvContainer Container;
	if(!Registry->GetKey("system.queues", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occure here

	// Creates the queues object if it doesn't exist
	Container.CreateObject("queues", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("queues");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("queues");
		Container.CreateObject("queues", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("queues");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *QueueTable = (NvObjTable*)Object;
	QueueTable->Clear();
	QueueTable->InsertColumn("Name");
	QueueTable->InsertColumn("Description");
	QueueTable->InsertColumn("QueueManagerName");
	QueueTable->InsertColumn("QueueName");
	QueueTable->InsertColumn("Filter");
	QueueTable->InsertColumn("NumThreads");
	QueueTable->InsertColumn("AutoStart");
	QueueTable->InsertColumn("Channel");
	QueueTable->InsertColumn("Server");
	int index_Name = QueueTable->GetColumnIndex("Name");
	int index_Desc = QueueTable->GetColumnIndex("Description");
	int index_QueueManagerName = QueueTable->GetColumnIndex("QueueManagerName");
	int index_QueueName = QueueTable->GetColumnIndex("QueueName");
	int index_Filter = QueueTable->GetColumnIndex("Filter");
	int index_NumThreads = QueueTable->GetColumnIndex("NumThreads");
	int index_AutoStart = QueueTable->GetColumnIndex("AutoStart");
	int index_Channel = QueueTable->GetColumnIndex("Channel");
	int index_Server = QueueTable->GetColumnIndex("Server");

	int j;
	map<string,NvMq *>::iterator pos;
	for(pos = m_mQueues.begin(),j=0; pos != m_mQueues.end(); ++pos)
	{
		NvMq *Queue = pos->second;
		if(Queue == NULL)
			continue;

		QueueTable->InsertRow();
		QueueTable->InsertCellLine(j, index_Name, Queue->Name.c_str());
		QueueTable->InsertCellLine(j, index_Desc, Queue->Description.c_str());
		QueueTable->InsertCellLine(j, index_QueueManagerName, Queue->QueueManagerName.c_str());
		QueueTable->InsertCellLine(j, index_QueueName, Queue->QueueName.c_str());
		QueueTable->InsertCellLine(j, index_Filter, Queue->Filter.c_str());
		char Buf[50];
		sprintf(Buf, "%d", Queue->m_iNumThreads);
		QueueTable->InsertCellLine(j, index_NumThreads, Buf);
		if(Queue->m_bAutoStart)
			QueueTable->InsertCellLine(j, index_AutoStart, "YES");
		else
			QueueTable->InsertCellLine(j, index_AutoStart, "NO");
		QueueTable->InsertCellLine(j, index_Channel, Queue->Channel.c_str());
		QueueTable->InsertCellLine(j, index_Server, Queue->Server.c_str());
		j++;
	}
	
	if(!Registry->SetKey("system.queues", &Container))
		return false;

	return true;
}
	


bool NvMqList::CreateQueue(const char *Name, const char *Description, const char *QueueManagerName, const char *QueueName, const char *Filter, int NumThreads, int AutoStart, const char *Channel, const char *Server)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue != NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	FoundQueue = new NvMq;
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	FoundQueue->MqLib = MqLib;
	FoundQueue->Name = NameHere;
	FoundQueue->Ident = GetQueueIdent(NameHere);
	if(Description != NULL)
		FoundQueue->Description = Description;
	if(QueueManagerName != NULL)
		FoundQueue->QueueManagerName = QueueManagerName;
	if(QueueName != NULL)
		FoundQueue->QueueName = QueueName;
	if(Filter != NULL)
		FoundQueue->Filter = Filter;
	if(NumThreads != -1)
		FoundQueue->m_iNumThreads = NumThreads;
	if(FoundQueue->m_iNumThreads <= 0)
		FoundQueue->m_iNumThreads = 1;
	if(AutoStart == 0)
		FoundQueue->m_bAutoStart = false;
	if(AutoStart == 1)
		FoundQueue->m_bAutoStart = true;
	if(Channel != NULL)
		FoundQueue->Channel = Channel;
	if(Server != NULL)
		FoundQueue->Server = Server;

	m_mQueues[FoundQueue->Ident] = FoundQueue;

	// Save the new queue list
	if(!Save())
	{
		m_mQueues.erase(FoundQueue->Ident);
		delete FoundQueue;
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}


	


bool NvMqList::SetQueueParam(const char *Name, const char *Description, const char *QueueManagerName, const char *QueueName, const char *Filter, int NumThreads, int AutoStart, const char *Channel, const char *Server)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(FoundQueue->IsRunning())
	{
		// The queue must be stopped
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(Description != NULL)
		FoundQueue->Description = Description;
	if(QueueManagerName != NULL)
		FoundQueue->QueueManagerName = QueueManagerName;
	if(QueueName != NULL)
		FoundQueue->QueueName = QueueName;
	if(Filter != NULL)
		FoundQueue->Filter = Filter;
	if(NumThreads != -1)
		FoundQueue->m_iNumThreads = NumThreads;
	if(FoundQueue->m_iNumThreads <= 0)
		FoundQueue->m_iNumThreads = 1;
	if(AutoStart == 0)
		FoundQueue->m_bAutoStart = false;
	if(AutoStart == 1)
		FoundQueue->m_bAutoStart = true;
	if(Channel != NULL)
		FoundQueue->Channel = Channel;
	if(Server != NULL)
		FoundQueue->Server = Server;

	// Save the new queue list
	if(!Save())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}




bool NvMqList::RemoveQueue(const char *Name)
{
	if(Name ==NULL)
		return true;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return true;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(FoundQueue->IsRunning())
	{
		// The queue must be stopped
		NV_UNLOCKOBJECT(this);
		return false;
	}
	m_mQueues.erase(FoundQueue->Ident);
	delete FoundQueue;


	// Save the new queue list
	Save();

	NV_UNLOCKOBJECT(this);
	return true;
}



bool NvMqList::StartQueue(const char *Name)
{
	if(!MqLib->IsMqLibLoaded())
		return false;

	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(FoundQueue->IsRunning())
	{
		// The queue is ever running
		NV_UNLOCKOBJECT(this);
		return true;
	}

	bool RetVal = FoundQueue->Start();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvMqList::StopQueue(const char *Name, bool Force, int WaitTime)
{
	if(!MqLib->IsMqLibLoaded())
		return true;

	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(!FoundQueue->IsRunning())
	{
		// The queue is ever stopped
		NV_UNLOCKOBJECT(this);
		return true;
	}

	bool RetVal = FoundQueue->Stop(Force, WaitTime);

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvMqList::IsQueueRunning(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	bool RetVal = FoundQueue->IsRunning();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}





bool NvMqList::QueueList(const char *Name, NvObjTable *QueueInfo)
{
	QueueInfo->Clear();

	bool AllQueues = true;

	NV_LOCKOBJECT(this);

	NvMq *FoundQueue = NULL;
	if((Name != NULL) && strlen(Name))
	{
		// There is a request on a specific queue
		FoundQueue = GetQueue(Name);
		if(FoundQueue == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllQueues = false;
	}


	QueueInfo->InsertColumn("NAME");
	QueueInfo->InsertColumn("DESCRIPTION");
	QueueInfo->InsertColumn("STATUS");
	QueueInfo->InsertColumn("NUM_THREADS");
	QueueInfo->InsertColumn("AUTOSTART");
	QueueInfo->InsertColumn("QUEUE_MANAGER");
	QueueInfo->InsertColumn("QUEUE_NAME");
	QueueInfo->InsertColumn("FILTER");
	QueueInfo->InsertColumn("CHANNEL");
	QueueInfo->InsertColumn("SERVER");

	if(AllQueues)
	{
		int j;
		map<string,NvMq *>::iterator pos;
		for(pos = m_mQueues.begin(),j=0; pos != m_mQueues.end(); ++pos)
		{
			NvMq *Queue = pos->second;
			if(Queue == NULL)
				continue;
			QueueInfo->InsertRow();
			QueueInfo->InsertCellLine(j, 0, Queue->Name);
			QueueInfo->InsertCellLine(j, 1, Queue->Description);
			if(Queue->IsRunning())
				QueueInfo->InsertCellLine(j, 2, "RUNNING");
			else
				QueueInfo->InsertCellLine(j, 2, "STOPPED");
			char Buffer[20];
			sprintf(Buffer, "%d", Queue->m_iNumThreads);
			QueueInfo->InsertCellLine(j, 3, Buffer);
			if(Queue->m_bAutoStart)
				QueueInfo->InsertCellLine(j, 4, "YES");
			else
				QueueInfo->InsertCellLine(j, 4, "NO");
			QueueInfo->InsertCellLine(j, 5, Queue->QueueManagerName);
			QueueInfo->InsertCellLine(j, 6, Queue->QueueName);
			QueueInfo->InsertCellLine(j, 7, Queue->Filter);
			QueueInfo->InsertCellLine(j, 8, Queue->Channel);
			QueueInfo->InsertCellLine(j, 9, Queue->Server);
			j++;
		}
	}
	else
	{
		QueueInfo->InsertRow();
		QueueInfo->InsertCellLine(0, 0, FoundQueue->Name);
		QueueInfo->InsertCellLine(0, 1, FoundQueue->Description);
		if(FoundQueue->IsRunning())
			QueueInfo->InsertCellLine(0, 2, "RUNNING");
		else
			QueueInfo->InsertCellLine(0, 2, "STOPPED");
		char Buffer[20];
		sprintf(Buffer, "%d", FoundQueue->m_iNumThreads);
		QueueInfo->InsertCellLine(0, 3, Buffer);
		if(FoundQueue->m_bAutoStart)
			QueueInfo->InsertCellLine(0, 4, "YES");
		else
			QueueInfo->InsertCellLine(0, 4, "NO");
		QueueInfo->InsertCellLine(0, 5, FoundQueue->QueueManagerName);
		QueueInfo->InsertCellLine(0, 6, FoundQueue->QueueName);
		QueueInfo->InsertCellLine(0, 7, FoundQueue->Filter);
		QueueInfo->InsertCellLine(0, 8, FoundQueue->Channel);
		QueueInfo->InsertCellLine(0, 9, FoundQueue->Server);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}




bool NvMqList::AutoStartQueues()
{
	if(!MqLib->IsMqLibLoaded())
		return false;
	NV_LOCKOBJECT(this);
	map<string,NvMq *>::iterator pos;
	for(pos = m_mQueues.begin(); pos != m_mQueues.end(); ++pos)
	{
		NvMq *Queue = pos->second;
		if(Queue == NULL)
			continue;
		if(Queue->m_bAutoStart)
			Queue->Start();
	}
	NV_UNLOCKOBJECT(this);

	return true;
}




bool NvMqList::StopQueues()
{
	if(!MqLib->IsMqLibLoaded())
		return true;

	NV_LOCKOBJECT(this);
	map<string,NvMq *>::iterator pos;
	for(pos = m_mQueues.begin(); pos != m_mQueues.end(); ++pos)
	{
		NvMq *Queue = pos->second;
		if(Queue == NULL)
			continue;
		Queue->Stop(true, 30);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvMqList::QueueExist(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanQueueName(NameHere);

	if(!NameHere.length())
		return false;

	bool RetVal = false;
	NV_LOCKOBJECT(this);
	NvMq *FoundQueue = GetQueue(NameHere.c_str());
	if(FoundQueue != NULL)
		RetVal = true;

	NV_UNLOCKOBJECT(this);
	return RetVal;
}
