// ***************************************************************************************
// File: nvs.cpp
// Description: main program entry point, implementation of the NvsServer class
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************


#include "nvs.h"
#ifndef WIN32
#include <pwd.h>
#endif

#include <openssl/rsa.h>       /* SSLeay stuff */
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER


using namespace std;

#define SERVER_ADDRESS_CTRL "ctrl.nirva-systems.com:80"

// Affinity includes
#ifdef LINUX
#include <sched.h>
#endif
#ifdef SOLARIS
#include <sys/pset.h>
#endif
#ifdef HPUX
#include <sys/mpctl.h>
#endif
#ifdef HPUXI
#include <sys/mpctl.h>
#endif

#ifdef WIN32	// real main definition for window. The window specific start functions are in nvswin.cpp
int realmain(int argc, char* argv[]);	// fonction prototype for the real main
#endif	// end window service specific code

unsigned int WindowService = 0;					// set to 1 if ran as a window service
NvsServer *MainServer = NULL;					// Main server object. All is made in it
static bool ConsoleMode = false;
static bool WithTime = false;
static bool Verbose = false;
static bool CertificateMode = false;
static bool DefaultParam = false;				// Allow to reload default NIRVA parameters
static bool NoJava = false;						// Allow to use Nirva without java VM
static bool NoMq = false;						// Allow to use Nirva without MQ
static bool InternalUnicode = false;			// Allow to use unicode as internal coding
static bool ClusterMode = false;				// Allow to run in cluster mode
NvMutex ProtectConsoleLog;
const char* CertificateFile = NULL;		// Used in certificate registration mode
const char* KeyFile = NULL;				// Used in certificate registration mode
const char* ClusterServer = NULL;		// Address of the peer cluster
bool bOldObjectsUnicode = false;				// Old objects unicode

void CatchSignal(int SigNum);
int GetNumberOfProcessors();

#ifdef WIN32
LONG __stdcall ExceptCallBack(EXCEPTION_POINTERS *pExPtrs);
#endif


#ifdef WIN32
extern int GetNumberOfProcessorsWin32();
#endif

//**********************************************************************************
//**********************************************************************************
// Main function of the server
// This function checks for clients to connect until a request to close the server has been asked
// When a client connects, the function creates a separate thread that manages the client connection
//**********************************************************************************
//**********************************************************************************
#ifdef WIN32
int realmain(int argc, char* argv[])	// because of window service
#else
int main(int argc, char* argv[])
#endif
{
	bool RetCode;
#ifdef WIN32
	SetUnhandledExceptionFilter(ExceptCallBack);
#ifdef _DEBUG
  //InitAllocCheck(ACOutput_XML, TRUE,0);
#endif
#endif
{
	// Create the main server object
	MainServer = NVNEW NvsServer;


// The next feature is made in order to set the dynamic loader library path
// to point to the NIRVA Bin and NIRVA Java/bin directories under UNIX
// For that, we must restart an instance of nvs because on UNIX platforms and
// especially on LINUX and SOLARIS, the dynamic loader (ld) is using the LD_LIBRARY_PATH
// environment variable used when starting the process so we cannot dynamically change it.
// The solution is to re-create the same process after having changed the LD_LIBRARY_PATH
// environment variable
#ifndef WIN32
	string m_sNirvaDir = "/usr/nirva";
	char *envbuf = getenv("NIRVA");
	if(envbuf != NULL)
	{
		m_sNirvaDir = envbuf;
		CleanString(m_sNirvaDir);
	}
	if(!m_sNirvaDir.length())
		m_sNirvaDir = "/usr/nirva";

	RemoveDirSuffix(m_sNirvaDir);
	AddDirSuffix(m_sNirvaDir);

	// We must now if Nirva is using internal or external jvm here
	bool bInternalJvm = true;
	NvRegistryContainer BasicSystemRegistry;
	NvRegistryContainer SystemRegistry;
	string SystemRegistryRootDirectory = m_sNirvaDir;
	SystemRegistryRootDirectory += "Registry";
	// This directory must also exist
	if(!DirExist(SystemRegistryRootDirectory.c_str()))
	{
		printf("Bad registry directory\n");
		delete MainServer;
		return -1;
	}

	AddDirSuffix(SystemRegistryRootDirectory);

	// We set the system root directory for now to default one
	SystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());

	// We set the basic system root directory
	BasicSystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());
   
	
	NvContainer Container;
	if(BasicSystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "basics", false, true))
	{
		NvObj *Object = Container.GetObject("basics");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
		{
			NvObjIndStringList *Parameters = (NvObjIndStringList*)Object;
			string Value;
			
			// Get registry home directory
			Value = Parameters->GetValue("RegHome"); 
			if(Value.length())
				SystemRegistryRootDirectory = Value;
			CleanString(SystemRegistryRootDirectory);
			AddDirSuffix(SystemRegistryRootDirectory);
		}
	}

	SystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());

	string ExtJvmPath = "";
	if(SystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "parameters", false, true))
	{
		NvObj *Object = Container.GetObject("parameters");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
		{
			NvObjIndStringList *Parameters = (NvObjIndStringList*)Object;
			string Value;
		
			// Get internal jvm flag
			Value = Parameters->GetValue("InternalJvm"); 
			if(Value.length() && (Value == "0"))
			{
				bInternalJvm = false;
				ExtJvmPath = Parameters->GetValue("JVMPathName");
			}
		}
	}

	string EnvVar = "LD_LIBRARY_PATH";
	string PathToAdd = "./:";
	PathToAdd += m_sNirvaDir;
	PathToAdd += "Bin:";
	if(bInternalJvm)
	{
		PathToAdd += m_sNirvaDir;
#ifdef SOLARIS
		PathToAdd += "Java/lib/sparc";
#endif
#ifdef LINUX
#ifdef X64
		PathToAdd += "Java/lib/amd64:";
		PathToAdd += m_sNirvaDir;
		PathToAdd += "Java/lib/amd64/server";
#else
		PathToAdd += "Java/lib/i386:";
		PathToAdd += m_sNirvaDir;
		PathToAdd += "Java/lib/i386/server";
#endif
#endif
#ifdef HPUX
		PathToAdd += "Java/lib/PA_RISC2.0:";
		PathToAdd += m_sNirvaDir;
		PathToAdd += "Java/lib/PA_RISC2.0/server";
#endif
#ifdef HPUXI
		PathToAdd += "Java/lib/IA64N:";
		PathToAdd += m_sNirvaDir;
		PathToAdd += "Java/lib/IA64N/server";
#endif
#ifdef AIX
		PathToAdd += "Java/bin:";
		PathToAdd += m_sNirvaDir;
		PathToAdd += "Java/bin/classic";
#endif
	}
	SetDirName(PathToAdd);


#ifdef AIX
	EnvVar = "LIBPATH";
#endif

	string EnvValue = "";
	envbuf = getenv(EnvVar.c_str());
	if(envbuf != NULL)
		EnvValue = envbuf;
	CleanString(EnvValue);
	bool UpdateIt = false;
	string NewEnvValue = EnvVar;
	NewEnvValue += "=";
	NewEnvValue += PathToAdd;
	string::size_type index = EnvValue.find("NIRVA_EXE_SET");
	if(index == string::npos)
		UpdateIt = true;
//	{
//		NewEnvValue += PathToAdd;
//		UpdateIt = true;
//	}

/*
	// We add also the nirva Bin directory
	PathToAdd = m_sNirvaDir;
	PathToAdd += "Bin";
	SetDirName(PathToAdd);
	index = EnvValue.find(PathToAdd);
	if(index == string::npos)
	{
		if(UpdateIt)
			NewEnvValue += ":";
		NewEnvValue += PathToAdd;
		UpdateIt = true;
	}
*/
	if(UpdateIt)
	{
		if(EnvValue.length())
			NewEnvValue += ":";
		NewEnvValue += EnvValue;
		NewEnvValue += ":NIRVA_EXE_SET";
		char *ProviBuf = (char*)malloc(NewEnvValue.length()+10);
		strcpy(ProviBuf, NewEnvValue.c_str());
		putenv(ProviBuf);
#ifdef HPUX
		// For Hp ux we must also set the SHLIB_PATH
		NewEnvValue = "SHLIB_PATH";
		NewEnvValue += "=";
		NewEnvValue += PathToAdd;
		EnvValue = "";
		envbuf = getenv("SHLIB_PATH");
		if(envbuf != NULL)
			EnvValue = envbuf;
		CleanString(EnvValue);
		if(EnvValue.length())
			NewEnvValue += ":";
		NewEnvValue += EnvValue;
		char *ProviBuf2 = (char*)malloc(NewEnvValue.length()+10);
		strcpy(ProviBuf2, NewEnvValue.c_str());
		putenv(ProviBuf2);

		// For Hp ux we must also set the LD_PRELOAD_ONCE to the path of the jvm
		// otherwise there is an error when loading the jvm (because jvm has TLS)
		// THis requires a patch to HPUX PARISC (PHSS_35381 on hp-ux 11.11)
		if(bInternalJvm)
		{
			NewEnvValue = "LD_PRELOAD_ONCE=";
			NewEnvValue += m_sNirvaDir;
			NewEnvValue += "Java/lib/PA_RISC2.0/server/libjvm.sl";
			char *ProviBuf3 = (char*)malloc(NewEnvValue.length()+10);
			strcpy(ProviBuf3, NewEnvValue.c_str());
			putenv(ProviBuf3);
		}
		else
		{
			if(ExtJvmPath.length())
			{
				NewEnvValue = "LD_PRELOAD_ONCE=";
				NewEnvValue += ExtJvmPath;
				char *ProviBuf3 = (char*)malloc(NewEnvValue.length()+10);
				strcpy(ProviBuf3, NewEnvValue.c_str());
				putenv(ProviBuf3);
			}
		}
#endif

#ifdef HPUXI
		// For Hp ux we must also set the SHLIB_PATH
		NewEnvValue = "SHLIB_PATH";
		NewEnvValue += "=";
		NewEnvValue += PathToAdd;
		EnvValue = "";
		envbuf = getenv("SHLIB_PATH");
		if(envbuf != NULL)
			EnvValue = envbuf;
		CleanString(EnvValue);
		if(EnvValue.length())
			NewEnvValue += ":";
		NewEnvValue += EnvValue;
		char *ProviBuf2 = (char*)malloc(NewEnvValue.length()+10);
		strcpy(ProviBuf2, NewEnvValue.c_str());
		putenv(ProviBuf2);

		// For Hp ux we must also set the LD_PRELOAD_ONCE to the path of the jvm
		// otherwise there is an error when loading the jvm (because jvm has TLS)
		if(bInternalJvm)
		{
			NewEnvValue = "LD_PRELOAD_ONCE=";
			NewEnvValue += m_sNirvaDir;
			NewEnvValue += "Java/lib/IA64N/server/libjvm.so";
			char *ProviBuf3 = (char*)malloc(NewEnvValue.length()+10);
			strcpy(ProviBuf3, NewEnvValue.c_str());
			putenv(ProviBuf3);
		}
		else
		{
			if(ExtJvmPath.length())
			{
				NewEnvValue = "LD_PRELOAD_ONCE=";
				NewEnvValue += ExtJvmPath;
				char *ProviBuf3 = (char*)malloc(NewEnvValue.length()+10);
				strcpy(ProviBuf3, NewEnvValue.c_str());
				putenv(ProviBuf3);
			}
		}
#endif

//		free(ProviBuf);
		string ExePath = m_sNirvaDir;
		ExePath += "Bin/nvs";
		execv(ExePath.c_str(), argv);
		// The actual process lives here
	}

#endif //ifndef WIN32



#ifdef WIN32
	// initialize Win32 sockets
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD( 2, 0 );
 
	err = WSAStartup(wVersionRequested, &wsaData);
	if ( err != 0 )
	{
		printf("Cannot initialize sockets\n");
		delete MainServer;
		return -1;
	}
#endif
	
	if(argc > 1)
	{
		int i;
		for(i=1; i<argc; i++)
		{
			char buffer[256];
			strcpy(buffer, argv[i]);
			if ((buffer[0] == '-') || (buffer[0] == '/'))
			{
				strlwr(buffer);
				if(strstr(buffer, "c") != NULL)
					ConsoleMode = true;
				if(strstr(buffer, "t") != NULL)
					WithTime = true;
				if(strstr(buffer, "v") != NULL)
					Verbose = true;
				if(strstr(buffer, "d") != NULL)
					DefaultParam = true;
				if(strstr(buffer, "j") != NULL)
					NoJava = true;
				if(strstr(buffer, "q") != NULL)
					NoMq = true;
				if(strstr(buffer, "k") != NULL)
					CertificateMode = true;
				if(strstr(buffer, "u") != NULL)
					InternalUnicode = true;
				if(strstr(buffer, "l") != NULL)
				{
					ClusterMode = true;
					if(i<(argc-1))
					{
						i++;
						ClusterServer = argv[i];
					}
				}
				if(strstr(buffer, "g") != NULL)
				{
					if(i<(argc-1))
					{
						i++;
						MainServer->m_sDebugMode = argv[i];
					}
				}
			}
		}
	}

	if(CertificateMode)
	{
		ConsoleMode = true;
		NoJava = true;
		NoMq = true;
		InternalUnicode = false;

		// In certifiacte mode the nvs command line must be only nvs -k certificatefile privatekeyfile

		if(argc < 4)
		{
			printf("Usage in certificate mode: nvs -k certificatefile privatekeyfile\n");
			delete MainServer;
			return -1;
		}
		CertificateFile = argv[2];
		KeyFile = argv[3];
	}

#ifndef WIN32
	if(!ConsoleMode)	// we run the process as a daemon
	{
	   switch (fork())
	   {
		case -1:			// can't fork
			printf("Cannot start NIRVA server\n");
			delete MainServer;
			return -1;
		 case 0:			// child, process becomes a daemon:
//			fclose(stdin);
//			fclose(stdout);
//			fclose(stderr);
			if(setsid() == -1)		// request a new session (job control)
			{
				delete MainServer;
				return -1;
			}
			break;
		 default:					// parent returns to calling process:
			delete MainServer;
			return 0;
	   }
	}
#endif

		
	// Catching system signals for trying to terminate normally on crash or kill command or control-C request

	signal(SIGINT, CatchSignal);
	signal(SIGTERM, CatchSignal);

#ifndef WIN32
	signal(SIGQUIT, CatchSignal);
	signal(SIGIOT, CatchSignal);
	signal(SIGTSTP, CatchSignal);

//PMARC Modif we don't block the SIGCHLD because it creates problems with the ExecPrg function on AIX
// The ExecPrgm returns an error when the process has correctly terminated
//	signal(SIGCHLD, SIG_IGN);	// block this signal to not allow services running process to create zombies
//PMARC Modif big files
	signal(SIGPIPE, SIG_IGN);	// block this signal to avoid problems in socket layer

	// Init thread signal handler
	InitThreadSignalHandler();

#endif

	// Init session command table
	InitCommandTable();

	if(ConsoleMode)
		printf("\nNIRVA server started\n");


#ifdef WIN32
#ifdef _DEBUG
	//_CrtMemState s1;
	//_CrtMemCheckpoint( &s1 );
#endif
#endif


	RetCode = MainServer->Run(true);

	if(ConsoleMode)
		printf("\nNIRVA server stopped\n");

	try
	{
		delete MainServer;
	}
	catch(...)
	{
	}
	CleanupCommandTable();

#ifdef WIN32
#ifdef _DEBUG
	//_CrtMemDumpAllObjectsSince(  &s1 );
#endif
#endif

}
#ifdef WIN32
#ifdef _DEBUG
  //DeInitAllocCheck();
#endif
#endif
	
	if(RetCode)
		return 0;
	else
		return -1;
}


//**********************************************************************************
//**********************************************************************************
// catch signals
// This function is called when major termination signals are sent to the process
// It simply stops the main task
// This ensure that the server is closed properly even in case of scratch or kill commands
//**********************************************************************************
//**********************************************************************************
void CatchSignal(int SigNum)
{
	// just set the AskToTerminate global variable to 1
	// it will be catched by all threads for them to terminate themselves

	if(MainServer != NULL)
		MainServer->Stop();
}


NvsServer::NvsServer()
{
	b_Initialized = false;
	tStartTime = time(NULL);
	Containers.m_bUseLocking = true;
	Queues.MqLib = &MqInterface;
	Aliases.m_bCaseSensitive = true;
	m_iHttpMaxMemBody = 1024*1024;
	m_sServerStatus = "STARTING";
	m_bStopRequest = false;
	m_bOldObjectsUnicode = false;
	m_sDebugMode = "";
	Garbage = NULL;
}

NvsServer* GetMain()
{
	return MainServer;
}



// server main thread function
int NvsServer::Main()
{
	int LoopCountForSessionListUpdate = 0;
	int LoopCountForGarbageCollectorUpdate = 0;
	int LoopCountForRegistryCacheClear = 0;
	int LoopCountForXsltCacheClear = 0;
	int LoopCountForFreePerl = 0;
	int LoopCountForCheckSema = 0;
	int LoopCountForCleanJava = 0;
	int LoopCountForHttpClientConnectionUpdate = 0;
	int NumChecks = 6;
	// The Main loop must be checked every 10 seconds + processing
	int SleepVal = 10000/NumChecks;
	int TimeForSessionListUpdate = 3;	// Every 30 seconds
	int TimeForGarbageCollectorUpdate = 3;	// Every 30 seconds
	int TimeForRegistryCacheClear = 360;	// Every 3600 seconds (1 hour)
	int TimeForXsltCacheClear = 360 * 24;	// Every day
	int TimeForFreePerl = 3;	// Every 30 seconds
	int TimeForCheckSema = 6;	// Every 60 seconds
	int TimeForCleanJava = 6;	// Every 60 seconds
	int TimeForHttpClientConnectionUpdate = 12;	// Every 120 seconds
	while(1)
	{
		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForSessionListUpdate++;
		if(LoopCountForSessionListUpdate >= TimeForSessionListUpdate)
		{
			Sessions.Update();	
			LoopCountForSessionListUpdate = 0;
		}

		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForGarbageCollectorUpdate++;
		if(LoopCountForGarbageCollectorUpdate >= TimeForGarbageCollectorUpdate)
		{
			Garbage->RemoveExpired();	
			LoopCountForGarbageCollectorUpdate = 0;
		}

		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForRegistryCacheClear++;
		if(LoopCountForRegistryCacheClear >= TimeForRegistryCacheClear)
		{
			SystemRegistry.Clear();
			Services.ClearRegistryCache();
			Applications.ClearRegistryCache();
			LoopCountForRegistryCacheClear = 0;
		}


		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForXsltCacheClear++;
		if(LoopCountForXsltCacheClear >= TimeForXsltCacheClear)
		{
			NvXml::EmptyCache();
			LoopCountForXsltCacheClear = 0;
		}


		// Not supported for now

		// we leave if requested
/*		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		// Not supported for now
		LoopCountForFreePerl++;
		if(LoopCountForFreePerl >= TimeForFreePerl)
		{
			// The Perl autofree is used to free the Perl library when Perl is not in use
			// In fact it doesn't work for now on UNIX platforms because that pollutes the
			// fork function that not works any more when the perl library has been loaded
			// and then unloaded

			PerlInterface.AutoFree();
			LoopCountForFreePerl = 0;
		}
*/
		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForCheckSema++;
		if(LoopCountForCheckSema >= TimeForCheckSema)
		{
			SemaList.CheckSessions();
			LoopCountForCheckSema = 0;
		}

		// Not supported for now

/*		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);

		LoopCountForCleanJava++;
		if(LoopCountForCleanJava >= TimeForCleanJava)
		{
			JavaInterface.CleanJavaLib(m_sNirvaDir.c_str(), GetJvmMinMem(), GetJvmMaxMem());
			LoopCountForCleanJava = 0;
		}
*/
		// we leave if requested
		if(AskToTerminate.IsSignaled())
			break;
		SLEEP(SleepVal);


		LoopCountForHttpClientConnectionUpdate++;
		if(LoopCountForHttpClientConnectionUpdate >= TimeForHttpClientConnectionUpdate)
		{
			Connections.Update();
			LoopCountForHttpClientConnectionUpdate = 0;
		}
	}
	return 0;
}


// called to stop the server
void NvsServer::Stop()
{
//	SendToLog("Request to stop the server");
	
	// we just the AskToTerminate event to signaled
	AskToTerminate.Signal();
	SLEEP(1000);
}


// called when an exception occurs (see class NvTask)
void NvsServer::OnException(string sInfo)
{
//	SendToLog("** Exception occured on the http server **");

	// We must signal the AskToTerminate event for all client threads
	AskToTerminate.Signal();
}


// called before the Main function (see class NvTask)
bool NvsServer::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	Threads.Add(THREAD_TYPE_MAIN);

	// Initialize the random generator
	srand((unsigned)time(NULL));

	// For window, increase the maximum number of stdio file from default 512
	// to 2048 (this is the max, trying over crashes the prgm under win64).
#ifdef WIN32
	if(_setmaxstdio(2048) == -1)
		_setmaxstdio(1024);
#else
	// We do the same for unix. This includes all descriptors like sockets and pipes
	struct rlimit r;
	r.rlim_cur = r.rlim_max = 8192;
	if(setrlimit(RLIMIT_NOFILE,&r) != 0)	// first try with 8192 then with 4096, 2048 and 1024
	{
		r.rlim_cur = r.rlim_max = 4096;
		if(setrlimit(RLIMIT_NOFILE,&r) != 0)
		{
			r.rlim_cur = r.rlim_max = 2048;
			if(setrlimit(RLIMIT_NOFILE,&r) != 0)
			{
				r.rlim_cur = r.rlim_max = 1024;
				setrlimit(RLIMIT_NOFILE,&r);
			}
		}
	}
#endif

	b_Initialized = false;


	// Get machine name and user
	char buffer[1024];
	memset(buffer,0,sizeof(buffer));
#ifdef WIN32
	unsigned long Size;
	Size = sizeof(buffer);
	GetUserName(buffer, &Size);
	m_sMachineUser = buffer;
	memset(buffer,0,sizeof(buffer));
	Size = sizeof(buffer);
	GetComputerNameEx((COMPUTER_NAME_FORMAT)ComputerNameDnsHostname, buffer,&Size);
//	GetComputerName(buffer, &Size);
	m_sMachineName = buffer;
#else
	struct passwd *Ident = getpwuid(getuid());
	strcpy(buffer, Ident->pw_name);
	m_sMachineUser = buffer;
	gethostname(buffer, sizeof(buffer));
	m_sMachineName = buffer;
#endif

	m_bDemoMode = true;
	m_bMqAvailable = false;
	m_bLcAvailable = false;
	m_iMaxSessions = -1;
	m_iLicenseType = LICENSE_PROD;
	m_iNumProcs = GetNumberOfProcessors();

	Garbage = NVNEW NvGarbage;

	b_ConsoleMode = ConsoleMode;
	b_WithTime = WithTime;
	b_Verbose = Verbose;
	b_Unicode = InternalUnicode;
	m_bOldObjectsUnicode = false;

	SystemRegistry.SetName("parameters");

	// Do extra code here
	AskToTerminate.Reset();		// we reset the asktoterminate event in case it is set

	
	// Get Nirva root directory
	if(b_ConsoleMode)
		printf("Get NIRVA environment variable...\n");
	// It comes from the NIRVA environment variable
	m_sNirvaDir = "";
	char *envbuf = getenv("NIRVA");	// The environment variable must exist
	if(envbuf == NULL)
	{
		if(b_ConsoleMode)
			printf("No environment variable NIRVA\n");
		return false;
	}
		
	m_sNirvaDir = envbuf;
	CleanString(m_sNirvaDir);

	RemoveDirSuffix(m_sNirvaDir);

	// The directory must not be empty...
	if(!m_sNirvaDir.length())
	{
		if(b_ConsoleMode)
			printf("Bad Nirva directory\n");
		return false;
	}

	// .. and must exist
	if(!DirExist(m_sNirvaDir.c_str()))
	{
		if(b_ConsoleMode)
			printf("Bad Nirva directory\n");
		return false;
	}

	AddDirSuffix(m_sNirvaDir);

	// Remove the system working directory
	string m_sWorDir = m_sNirvaDir;
	m_sWorDir += "Work";
	RemoveDir(m_sWorDir.c_str(), false);			

	string EnvVar = "PATH";
	string PathToAdd = m_sNirvaDir;
	PathToAdd += "Java/bin";
	SetDirName(PathToAdd);

#ifdef AIX
	EnvVar = "LIBPATH";
#endif
#ifdef LINUX
	EnvVar = "LD_LIBRARY_PATH";
#endif
#ifdef HPUX
	EnvVar = "LD_LIBRARY_PATH";
#endif
#ifdef HPUXI
	EnvVar = "LD_LIBRARY_PATH";
#endif
#ifdef SOLARIS
	EnvVar = "LD_LIBRARY_PATH";
#endif

	string EnvValue = "";
	envbuf = getenv(EnvVar.c_str());	// The environment variable must exist
	if(envbuf != NULL)
		EnvValue = envbuf;
	CleanString(EnvValue);
	string::size_type index = EnvValue.find(PathToAdd);
	if(index == string::npos)
	{
		string NewEnvValue = EnvVar;
		NewEnvValue += "=";
		NewEnvValue += PathToAdd;
		// We add also the nirva Bin directory
#ifdef WIN32
		NewEnvValue += ";";
#else
		NewEnvValue += ":";
#endif
		PathToAdd = m_sNirvaDir;
		PathToAdd += "Bin";
		SetDirName(PathToAdd);
		NewEnvValue += PathToAdd;
		if(EnvValue.length())
		{
#ifdef WIN32
			NewEnvValue += ";";
#else
			NewEnvValue += ":";
#endif
			NewEnvValue += EnvValue;
		}
		NewEnvValue += EnvValue;
		char *ProviBuf = (char*)malloc(NewEnvValue.length()+10);
		strcpy(ProviBuf, NewEnvValue.c_str());
		putenv(ProviBuf);
//		free(ProviBuf);
	}

	if(b_ConsoleMode)
		printf("Initialize registry...\n");
	string SystemRegistryRootDirectory = m_sNirvaDir;
	SystemRegistryRootDirectory += "Registry";
	// This directory must also exist
	if(!DirExist(SystemRegistryRootDirectory.c_str()))
	{
		if(b_ConsoleMode)
			printf("Bad registry directory\n");
		return false;
	}

	AddDirSuffix(SystemRegistryRootDirectory);

	// We set the system root directory for now to default one
	SystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());

	// We set the basic system root directory
	BasicSystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());
	if(!LoadBasicParameters(DefaultParam))
	{
		if(b_ConsoleMode)
			printf("Cannot load basic system parameters\n");
		return false;
	}

	m_sNewRegistryHomeDir = m_sRegistryHomeDir;
	m_sNewLogHomeDir = m_sLogHomeDir;
	m_sNewAppWorkHomeDir = m_sAppWorkHomeDir;

	// Now we can set the system root directory to the real path
	SystemRegistryRootDirectory = m_sRegistryHomeDir.c_str();
	AddDirSuffix(SystemRegistryRootDirectory);
	SystemRegistryRootDirectory += "Registry";
	if(!DirExist(SystemRegistryRootDirectory.c_str()))
	{
		if(!CreateDir(SystemRegistryRootDirectory.c_str()))
		{
			if(b_ConsoleMode)
				printf("Bad registry directory\n");
			return false;
		}
	}
	SystemRegistry.SetRootDirectory(SystemRegistryRootDirectory.c_str());


	if(b_ConsoleMode)
		printf("Test crash condition...\n");
	// Test for crash condition
	b_Crash = false;
	string RunFile = m_sNirvaDir;
	RunFile += "Bin";
	AddDirSuffix(RunFile);
	RunFile += "nirva.run";

	if(FileExist(RunFile.c_str()))
	{
		// We have found the nirva.run file
		// We must first test if another instance of Nirva is not running
		// For that, we try to connect the TCP/IP port
		
		// We first get the Port number from registry
		if(!LoadParameters(false, true))
		{
			if(b_ConsoleMode)
				printf("Cannot load system parameters\n");
			return false;
		}
		NvSocket ProviSocket;
		if(ProviSocket.Connect("127.0.0.1", m_iHttpServerPort, 5))
		{
			// We were connected so another instance is running
			ProviSocket.Close();
			if(b_ConsoleMode)
				printf("Nirva is ever running\n");
			return false;
		}

		// Try now with the SSL socket
		ProviSocket.bSslConnection = true;
		if(ProviSocket.Connect("127.0.0.1", m_iHttpsServerPort, 5))
		{
			// We were connected so another instance is running
			ProviSocket.Close();
			if(b_ConsoleMode)
				printf("Nirva is already running\n");
			return false;
		}

		b_Crash = true;		// and set crash flag to true
	}

	
	RemoveFile(RunFile.c_str()); // removes the run file when starting the server
	FILE *File = Nvfopen(RunFile.c_str(), "wt");	// and create it again
	if(File != NULL)
	{
		fwrite("Running", 1, 7, File);
		fclose(File);	       
	}



	if(CertificateMode)
	{
		printf("Installing certificate...\n");
		printf("Certificate file : %s\n", CertificateFile);
		if((!strlen(CertificateFile)) || (!FileExist(CertificateFile)))
		{
			printf("Bad certificate file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}
		printf("Private key file : %s\n", KeyFile);
		if((!strlen(KeyFile)) || (!FileExist(KeyFile)))
		{
			printf("Bad private key file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		HttpServerSSL.b_UseSSL = true;
		if(!HttpServerSSL.InitSSL(CertificateFile, KeyFile))
		{
			printf("Error checking certificates\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		NvObjBinary BinaryCertificateFile;
		if(!BinaryCertificateFile.Load(CertificateFile))
		{
			printf("Error loading certificate file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		NvObjBinary BinaryKeyFile;
		if(!BinaryKeyFile.Load(KeyFile))
		{
			printf("Error loading private key file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		// We create a hash for both files
		string HashCertificate = MD5_bytes((unsigned char*)(BinaryCertificateFile.GetBuffer()), BinaryCertificateFile.GetSize());
		string HashKey = MD5_bytes((unsigned char*)(BinaryKeyFile.GetBuffer()), BinaryKeyFile.GetSize());
		
		// We encrypt the hashes
		char Buffer[256];
		string EncryptionKey = "p";
		EncryptionKey += "nir";
		EncryptionKey += "va";
		EncryptionKey += "m";
		EncryptionKey += "cr";
		strcpy(Buffer, HashCertificate.c_str());
		EncryptString(EncryptionKey.c_str(), Buffer, 128);
		// we create a new hash from the encrypted one
		HashCertificate = MD5_string((unsigned char*)(Buffer));
		strcpy(Buffer, HashKey.c_str());
		EncryptString(EncryptionKey.c_str(), Buffer, 128);
		// we create a new hash from the encrypted one
		HashKey = MD5_string((unsigned char*)(Buffer));

		// Now we must save the calculated hash with the files
		NvObjBinary FinalBinaryCertificateFile;
		NvObjBinary FinalBinaryKeyFile;

		FinalBinaryCertificateFile.Append(HashCertificate);
		FinalBinaryCertificateFile.AppendFrom(&BinaryCertificateFile);

		FinalBinaryKeyFile.Append(HashKey);
		FinalBinaryKeyFile.AppendFrom(&BinaryKeyFile);

		string FinalCertificateFile = m_sNirvaDir;
		FinalCertificateFile += "Bin";
		AddDirSuffix(FinalCertificateFile);
		FinalCertificateFile += "nirva.nct";

		string FinalKeyFile = m_sNirvaDir;
		FinalKeyFile += "Bin";
		AddDirSuffix(FinalKeyFile);
		FinalKeyFile += "nirva.npk";
	
		if(!FinalBinaryCertificateFile.Save(FinalCertificateFile.c_str()))
		{
			printf("Error saving certificate file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		if(!FinalBinaryKeyFile.Save(FinalKeyFile.c_str()))
		{
			printf("Error saving private key file\n");
			RemoveFile(RunFile.c_str());
			return false;
		}

		printf("\nCertificate installation successful\n\n");
		RemoveFile(RunFile.c_str());
		return false;
	}

	// Do the cluster check here
	if(ClusterMode)
	{
		string sClusterPeer = "";
		if(ClusterServer != NULL)
			sClusterPeer = ClusterServer;
		CleanString(sClusterPeer);
		if(sClusterPeer.length())
		{
			if(b_ConsoleMode)
				printf("Waiting for peer server to be unreachable...\n");
			bool bInFirstTest = true;
			while(1)
			{
				NvRequest ProviRequest;
				ProviRequest.m_sServer = sClusterPeer;
				string Result;
				ProviRequest.PrivateCommand("GET_STATUS", Result);
				if(AskToTerminate.IsSignaled())
				{
					RemoveFile(RunFile.c_str());
					return false;
				}
				if(!Result.length())
				{
					if(bInFirstTest)
					{
						// Wait again 10 seconds to be sure
						printf("Peer server is not responding...\n");
						printf("Check again in 10 seconds...\n");
						bInFirstTest = false;
						int i;
						for(i=0;i<10;i++)
						{
							SLEEP(1000);
							if(AskToTerminate.IsSignaled())
							{
								RemoveFile(RunFile.c_str());
								return false;
							}
						}
						continue;
					}
					break;
				}
				if(AskToTerminate.IsSignaled())
				{
					RemoveFile(RunFile.c_str());
					return false;
				}
				if(!bInFirstTest)
				{
					printf("Peer server has been connected successfully...\n");
					printf("Waiting for peer server to be unreachable...\n");
				}
				bInFirstTest = true;
				SLEEP(1000);
				if(AskToTerminate.IsSignaled())
				{
					RemoveFile(RunFile.c_str());
					return false;
				}
			}
		}
	}
	
	// The server really starts here
	tStartTime = time(NULL);

	if(b_ConsoleMode)
		printf("Starting system log...\n");
	// Load logs
	Logs.Load();

	// Create system log if not existing
	if(!Logs.Exist("System"))
	{
		Logs.Create("System");
		Logs.SetLevel("System", 0);	// Initial system log is 0
		Logs.SetDescription("System", "Main system log");
	}
	
	// Create Webfiles log if not existing
	if(!Logs.Exist("Webfiles"))
	{
		Logs.Create("Webfiles");
		Logs.SetLevel("Webfiles", 0);	// Initial system log is 0
		Logs.SetDescription("Webfiles", "Web server pages log");
	}
	
	// Create Webservices log if not existing
	if(!Logs.Exist("Webservices"))
	{
		Logs.Create("Webservices");
		Logs.SetLevel("Webservices", 0);	// Initial system log is 0
		Logs.SetDescription("Webservices", "Web service log");
	}
	
	// Create Webservices log if not existing
	if(!Logs.Exist("Scheduler"))
	{
		Logs.Create("Scheduler");
		Logs.SetLevel("Scheduler", 0);	// Initial system log is 0
		Logs.SetDescription("Scheduler", "Scheduler log");
	}
	
	// Create Threads log if not existing
	if(!Logs.Exist("Threads"))
	{
		Logs.Create("Threads");
		Logs.SetLevel("Threads", 0);	// Initial system log is 0
		Logs.SetDescription("Threads", "Threads log");
	}
	
	// Create perl log if not existing
	if(!Logs.Exist("Perl"))
	{
		Logs.Create("Perl");
		Logs.SetLevel("Perl", 0);	// Initial system log is 0
		Logs.SetDescription("Perl", "Perl interface log");
	}
	
	// Create perl log if not existing
	if(!Logs.Exist("Sessions"))
	{
		Logs.Create("Sessions");
		Logs.SetLevel("Sessions", 0);	// Initial system log is 0
		Logs.SetDescription("Sessions", "Sessions log");
	}
	
	// Create java log if not existing
/*	if(!Logs.Exist("Java"))
	{
		Logs.Create("Java");
		Logs.SetLevel("Java", 0);	// Initial system log is 0
		Logs.SetDescription("Java", "Java interface log");
	}
*/	
	
	Logs.Write("System", "Starting Nirva server", "", NV_LOG_INFO, 1, "", "system");

	// Check if http/https server port is ever in use and leave if yes
	if(!CertificateMode)
	{
		NvSocket ProviSocketCheckHttp;
		if(ProviSocketCheckHttp.Connect("127.0.0.1", m_iHttpServerPort, 2))
		{
			// We were connected
			ProviSocketCheckHttp.Close();
			char Buf[256];
			sprintf(Buf, "Another program listens on the nirva HTTP port %d", m_iHttpServerPort);
			Logs.Write("System", Buf, "", NV_LOG_ERROR, 1, "", "system");
			if(b_ConsoleMode)
				printf("Another program listens on the nirva HTTP port %d\nNirva cannot be started", m_iHttpServerPort);
			RemoveFile(RunFile.c_str());
			return false;
		}
		if(!DefaultParam && m_bUseSSL)
		{
			NvSocket ProviSocketCheckHttps;
			if(ProviSocketCheckHttps.Connect("127.0.0.1", m_iHttpsServerPort, 2))
			{
				// We were connected
				ProviSocketCheckHttps.Close();
				char Buf[256];
				sprintf(Buf, "Another program listens on the nirva HTTPS port %d", m_iHttpsServerPort);
				Logs.Write("System", Buf, "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("Another program listens on the nirva HTTPS port %d\nNirva cannot be started", m_iHttpsServerPort);
				RemoveFile(RunFile.c_str());
				return false;
			}
		}
	}

	// Get the number od CPUs
	char Buffer[100];
	sprintf(Buffer, "Number of CPU found: %d", m_iNumProcs);
	Logs.Write("System", Buffer, "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
	{
		printf("Checking number of CPU...\n");
		printf("Number of CPU found: %d\n", m_iNumProcs);
	}

	// Load the License interface library
	Logs.Write("System", "Loading license library", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading  license library...\n");

	m_bLcAvailable = true;
	try
	{
		if(!LcInterface.LoadLcLib(m_sNirvaDir.c_str()))
		{
			Logs.Write("System", "Error loading license library", "", NV_LOG_ERROR, 1, "", "system");
			printf("Error loading license library. The license features will not be available\n");
			m_bLcAvailable = false;
		}
		if(m_bLcAvailable && !Licenses.Open())
		{
			Logs.Write("System", "Error initializing license library", "", NV_LOG_ERROR, 1, "", "system");
			printf("Error initializing license library. The license features will not be available\n");
			m_bLcAvailable = false;
			LcInterface.FreeLcLib();
		}
	}
	catch(...)
	{
		Logs.Write("System", "Error loading license library", "", NV_LOG_ERROR, 1, "", "system");
		printf("Error loading license library. The license features will not be available\n");
		m_bLcAvailable = false;
	}


	if(m_bLcAvailable)
	{
		Logs.Write("System", "Checking licenses", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Checking licenses...\n");
		string LicenseFileName = m_sNirvaDir;
		LicenseFileName += "Bin";
		AddDirSuffix(LicenseFileName);
		LicenseFileName += "nirva.lsc";
		Licenses.SetLicenseFile(LicenseFileName);


		if(FileExist(LicenseFileName.c_str()))
		{
			// We load the license file if it exists
			if(!Licenses.Load())
			{
				Logs.Write("System", "Cannot load license file", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("Cannot load license file\n");
				RemoveFile(RunFile.c_str()); // removes the run file when starting the server
				return false;
			}
		}
		else
		{
			// Otherwise, we create an empty license file
			if(!Licenses.Save())
			{
				Logs.Write("System", "Cannot create license file", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("Cannot create license file\n");
				RemoveFile(RunFile.c_str()); // removes the run file when starting the server
				return false;
			}
		}
	}

	// Load system parameters
	Logs.Write("System", "Loading system parameters", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading system parameters...\n");
	if(!LoadParameters(DefaultParam))
	{
		Logs.Write("System", "Cannot load system parameters", "", NV_LOG_ERROR, 1, "", "system");
		if(b_ConsoleMode)
			printf("Cannot load system parameters\n");
		RemoveFile(RunFile.c_str());
		return false;
	}

	m_sCtrlInstance = GetUniqString();

	// Set also deadlock checking in action if the debug mode is set
	if(GetDebug())
		NV_SETCHECKDEADLOCKS(true);

	// Enable the xslt cache if required
	if(GetEnableXslCache())
		SetEnableXslCache(true);


	// Set the deadlock file name int the nirva directory
	string DeadLockFile = m_sNirvaDir;
	DeadLockFile += "Bin";
	AddDirSuffix(DeadLockFile);
	DeadLockFile += "nvdead.log";
	NV_SETDEADLOCKFILENAME(DeadLockFile);
	

	// Load mime types table from registry
	LoadMimeTable();

	// Load alias table from registry
	LoadAliasTable();

	// Load virtual host table from registry
	VirtualHosts.Load();


	Logs.Write("System", "Initialize XML", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Initialize XML...\n");
	InitXml();

	// We must load now the description file if it exists
	LoadInfo();

	// get the Nirva version
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Info.GetObject("INFO"));
	if(IndStringList != NULL)
	{
		m_sVersion=IndStringList->GetValue("VERSION");
		m_sVersionType=IndStringList->GetValue("TYPE");
	}

	if(!NoJava)
	{
		JavaInterface.m_bAvailable = true;
		// Load the java interface library
		Logs.Write("System", "Loading java VM", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Loading java VM...\n");
		try
		{
			if(m_bInternalJvm)
			{
				if(!JavaInterface.LoadJavaLib(m_sNirvaDir.c_str(), m_JVMMinMemSize, m_JVMMaxMemSize, true, NULL, NULL, m_sJVMParam.c_str()))
				{
					Logs.Write("System", "Error loading java VM", "", NV_LOG_ERROR, 1, "", "system");
					printf("Error loading java VM. The java extension will not be available\n");
				}
			}
			else
			{
				if(!JavaInterface.LoadJavaLib(m_sNirvaDir.c_str(), m_JVMMinMemSize, m_JVMMaxMemSize, true, m_sJVMHomeDir.c_str(), m_sJVMPathName.c_str(), m_sJVMParam.c_str()))
				{
					Logs.Write("System", "Error loading java VM", "", NV_LOG_ERROR, 1, "", "system");
					printf("Error loading java VM. The java extension will not be available\n");
				}
			}
		}
		catch(...)
		{
			Logs.Write("System", "Error loading java VM", "", NV_LOG_ERROR, 1, "", "system");
			printf("Error loading java VM. The java extension will not be available\n");
		}
	}

#ifdef WIN32
	DotnetInterface.m_bAvailable = true;
	// Load the dotnet interface library
	Logs.Write("System", "Loading .net interface", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading .net interface...\n");
	try
	{
		if(!DotnetInterface.LoadDotnetLib(m_sNirvaDir.c_str(), true))
		{
			Logs.Write("System", "Error loading .net interface", "", NV_LOG_ERROR, 1, "", "system");
			printf("Error loading .net interface. The .net extension will not be available\n");
		}
	}
	catch(...)
	{
		Logs.Write("System", "Error loading .net interface", "", NV_LOG_ERROR, 1, "", "system");
		printf("Error loading java VM. The .net interface will not be available\n");
	}
#endif

//	if(b_ConsoleMode)
//		printf("Loading perl library...\n");
//	try
//	{
//		if(!PerlInterface.LoadPerlLib())
//			printf("Error loading perl library. The perl extension will not be available\n");
//	}
//	catch(...)
//	{
//		printf("Error loading perl library. The perl extension will not be available\n");
//	}


	// We load MQ queues even if not authorized
	Queues.Load();

	// Load the MQ interface library
	if(!NoMq)
	{
		Logs.Write("System", "Loading MQ connector", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Loading MQ connector...\n");

//		string Value;
//		if(Licenses.CheckKey("SYSTEM", NVS_SERVICE_ID, "MQ_CONNECTOR", Value))
			m_bMqAvailable = true;
//		else
//		{
//			printf("No license for MQ connector. The MQ connector will not be available\n");
//			Logs.Write("System", "No license for MQ connector", "", NV_LOG_ERROR, 1, "", "system");
//		}
	}
	if(m_bMqAvailable)
	{
		try
		{
			if(!MqInterface.LoadMqLib(m_sNirvaDir.c_str()))
			{
				Logs.Write("System", "Error loading MQ connector", "", NV_LOG_ERROR, 1, "", "system");
				printf("Error loading MQ connector. The MQ connector will not be available\n");
				m_bMqAvailable = false;
			}
		}
		catch(...)
		{
			Logs.Write("System", "Error loading MQ connector", "", NV_LOG_ERROR, 1, "", "system");
			printf("Error loading MQ connector. The MQ connector will not be available\n");
			m_bMqAvailable = false;
		}
	}

	Logs.Write("System", "Loading test sets", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading test sets...\n");
	TestSets.m_sDirPath = m_sNirvaDir;
	TestSets.m_sDirPath += "Files";
	AddDirSuffix(TestSets.m_sDirPath);
	TestSets.m_sDirPath += "Test";
	AddDirSuffix(TestSets.m_sDirPath);
	TestSets.m_sDirPath += "Sets";
	TestSets.Load();

	Logs.Write("System", "Loading web services", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading web services...\n");
	// Load the web service list
	Webservices.Load();


	Logs.Write("System", "Loading services", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading services...\n");
	// Load the service list
	Services.Load();


	// Load NVDEF application
	Logs.Write("System", "Loading NVDEF", "", NV_LOG_INFO, 1, "", "system");
	Applications.Load(true);

	// Create NVDEF if it doesn't exist
	if(Applications.GetApplication(NV_APPLICATION_DEFAULT_NAME, false) == NULL)
	{
		if(!Applications.CreateApplication(NV_APPLICATION_DEFAULT_NAME, "Nirva default application", m_iDefaultAppTimeOut, m_sDefaultAppSecuApp.c_str(), m_bDefaultAppUseScheduled, m_bDefaultAppAllowWebXml, true, m_bDefaultAppAllowNotOwnedSession, false, m_iDefaultAppAuthMode, m_bDefaultNirvaIsSsoUser))
		{
			Logs.Write("System", "Cannot create default application", "", NV_LOG_ERROR, 1, "", "system");
			if(b_ConsoleMode)
				printf("Cannot create default application\n");
			RemoveFile(RunFile.c_str());
			return false;
		}
	}
	NvApplication* DefaultApplication = Applications.GetApplication(NV_APPLICATION_DEFAULT_NAME, false);
	if(DefaultApplication != NULL)
	{
		DefaultApplication->m_iDefaultTimeOut = m_iDefaultAppTimeOut;
		DefaultApplication->m_iAuthMode = m_iDefaultAppAuthMode;
		DefaultApplication->m_bNirvaIsSsoUser = m_bDefaultNirvaIsSsoUser;
		DefaultApplication->m_sSecurityApplication = m_sDefaultAppSecuApp;
		DefaultApplication->bUseScheduledTransac = m_bDefaultAppUseScheduled;
		DefaultApplication->bAllowWebXml = m_bDefaultAppAllowWebXml;
		DefaultApplication->bAllowNotOwnedSession = m_bDefaultAppAllowNotOwnedSession;
	}
	Applications.EnableApplication(NV_APPLICATION_DEFAULT_NAME, true);

	// Starts the required services
	Logs.Write("System", "Start required services", "", NV_LOG_INFO, 1, "", "system");
	Services.LoadOnStart();

	// Load security
	Logs.Write("System", "Loading system security", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading system security...\n");
	Security.Load();
	
	Logs.Write("System", "Loading applications", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Loading applications...\n");
	
	// Load all the applications
	Applications.Load();

	// Create Default application if it doesn't exist
	if(Applications.GetApplication(NV_APPLICATION_DEFAULT_NAME, false) == NULL)
	{
		if(!Applications.CreateApplication(NV_APPLICATION_DEFAULT_NAME, "Nirva default application", m_iDefaultAppTimeOut, m_sDefaultAppSecuApp.c_str(), m_bDefaultAppUseScheduled, m_bDefaultAppAllowWebXml, false, m_bDefaultAppAllowNotOwnedSession, false, m_iDefaultAppAuthMode, m_bDefaultNirvaIsSsoUser))
		{
			if(b_ConsoleMode)
				printf("Cannot create default application\n");
			RemoveFile(RunFile.c_str());
			return false;
		}
	}
	DefaultApplication = Applications.GetApplication(NV_APPLICATION_DEFAULT_NAME, false);
	if(DefaultApplication != NULL)
	{
		DefaultApplication->m_iDefaultTimeOut = m_iDefaultAppTimeOut;
		DefaultApplication->m_iAuthMode = m_iDefaultAppAuthMode;
		DefaultApplication->m_bNirvaIsSsoUser = m_bDefaultNirvaIsSsoUser;
		DefaultApplication->m_sSecurityApplication = m_sDefaultAppSecuApp;
		DefaultApplication->bUseScheduledTransac = m_bDefaultAppUseScheduled;
		DefaultApplication->bAllowWebXml = m_bDefaultAppAllowWebXml;
		DefaultApplication->bAllowNotOwnedSession = m_bDefaultAppAllowNotOwnedSession;
	}
	Applications.EnableApplication(NV_APPLICATION_DEFAULT_NAME);

	// Clear the application work directories
	Applications.CleanWorkDir();


	// Save the application list immediatly for the case where the default application
	// has just been created
	Applications.Save();

	
	Logs.Write("System", "Starting applications", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Starting applications...\n");
	// Start applications
	// The applications are started without listeners
	// First start default application
	DefaultApplication = Applications.GetApplication(NV_APPLICATION_DEFAULT_NAME, false);
	DefaultApplication->Start(false, false);

	// Then others
	Applications.Start();


	sprintf(Buffer, "Starting HTTP server on port %d", m_iHttpServerPort);
	Logs.Write("System", Buffer, "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Starting HTTP server on port %d ...\n", m_iHttpServerPort);
	// starts the http server object

	HttpServer.Port = m_iHttpServerPort;
	HttpServer.ClientTimeOut = m_iHttpSessionTimeOut;
	HttpServer.sRootDir = m_sNirvaDir;
	HttpServer.sRootDir += "Wroot";
	HttpServer.sDefaultPage = GetNirvaDefaultPage();
	if(!HttpServer.sDefaultPage.length())
		HttpServer.sDefaultPage = "default.html";
	HttpServer.MaxStopWaitTime = m_iHttpStopWaitTime;
	HttpServer.MaxMemBody = m_iHttpMaxMemBody;
	HttpServer.b_LocalOnly = GetHttpLocal();
	
	// Check if somebody is ever listening on the same http port

	if(!HttpServer.Run())
	{
		Logs.Write("System", "Cannot start HTTP server", "", NV_LOG_ERROR, 1, "", "system");
		if(b_ConsoleMode)
			printf("Cannot start HTTP server\n");
		RemoveFile(RunFile.c_str());
		return false;
	}
	

	// starts the https server object

	HttpServerSSL.Port = m_iHttpsServerPort;
	HttpServerSSL.ClientTimeOut = m_iHttpSessionTimeOut;
	HttpServerSSL.sRootDir = m_sNirvaDir;
	HttpServerSSL.sRootDir += "Wroot";
	HttpServerSSL.sDefaultPage = GetNirvaDefaultPage();
	if(!HttpServerSSL.sDefaultPage.length())
		HttpServerSSL.sDefaultPage = "default.html";
	HttpServerSSL.MaxStopWaitTime = m_iHttpStopWaitTime;
	HttpServerSSL.MaxMemBody = m_iHttpMaxMemBody;
	HttpServerSSL.b_UseSSL = true;

	if(m_bUseSSL)
	{
		sprintf(Buffer, "Starting HTTPS server on port %d", m_iHttpsServerPort);
		Logs.Write("System", Buffer, "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Starting HTTPS server on port %d ...\n", m_iHttpsServerPort);

		string CertificateFile = m_sNirvaDir;
		CertificateFile += "Bin";
		AddDirSuffix(CertificateFile);
		CertificateFile += "nirva.nct1";
		RemoveFile(CertificateFile.c_str());

		string KeyFile = m_sNirvaDir;
		KeyFile += "Bin";
		AddDirSuffix(KeyFile);
		KeyFile += "nirva.npk1";
		RemoveFile(KeyFile.c_str());

		
		string FinalCertificateFile = m_sNirvaDir;
		FinalCertificateFile += "Bin";
		AddDirSuffix(FinalCertificateFile);
		FinalCertificateFile += "nirva.nct";

		string FinalKeyFile = m_sNirvaDir;
		FinalKeyFile += "Bin";
		AddDirSuffix(FinalKeyFile);
		FinalKeyFile += "nirva.npk";

		NvObjBinary FinalBinaryCertificateFile;

		bool bSuccess = true;
		if(!FinalBinaryCertificateFile.Load(FinalCertificateFile.c_str()))
		{
			Logs.Write("System", "Cannot load certificate", "", NV_LOG_ERROR, 1, "", "system");
			if(b_ConsoleMode)
				printf("Cannot load certificate...\n");
			bSuccess = false;
		}

		NvObjBinary FinalBinaryKeyFile;

		if(bSuccess)
		{
			if(!FinalBinaryKeyFile.Load(FinalKeyFile.c_str()))
			{
				Logs.Write("System", "Cannot load private key", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("Cannot load private key...\n");
				bSuccess = false;
			}
		}

		NvObjBinary BinaryCertificateFile;
		NvObjBinary BinaryKeyFile;
		if(bSuccess)
		{
			BinaryCertificateFile.AppendFrom(&FinalBinaryCertificateFile, 32);
			BinaryKeyFile.AppendFrom(&FinalBinaryKeyFile, 32);
			BinaryCertificateFile.Seek(0);
			BinaryKeyFile.Seek(0);
			FinalBinaryCertificateFile.Seek(0);
			FinalBinaryKeyFile.Seek(0);

			char Buffer[256];
			Buffer[32] = '\0';
			string HashFromFileCertificate = "";
			if(FinalBinaryCertificateFile.Read(Buffer, 32) == 32)
				HashFromFileCertificate = Buffer;
			Buffer[32] = '\0';
			string HashFromFileKey = "";
			if(FinalBinaryKeyFile.Read(Buffer, 32) == 32)
				HashFromFileKey = Buffer;

			// We create a hash for both files
			string HashCertificate = MD5_bytes((unsigned char*)(BinaryCertificateFile.GetBuffer()), BinaryCertificateFile.GetSize());
			string HashKey = MD5_bytes((unsigned char*)(BinaryKeyFile.GetBuffer()), BinaryKeyFile.GetSize());
			
			// We encrypt the hashes
			string EncryptionKey = "p";
			EncryptionKey += "nir";
			EncryptionKey += "va";
			EncryptionKey += "m";
			EncryptionKey += "cr";
			strcpy(Buffer, HashCertificate.c_str());
			EncryptString(EncryptionKey.c_str(), Buffer, 128);
			// we create a new hash from the encrypted one
			HashCertificate = MD5_string((unsigned char*)(Buffer));
			strcpy(Buffer, HashKey.c_str());
			EncryptString(EncryptionKey.c_str(), Buffer, 128);
			// we create a new hash from the encrypted one
			HashKey = MD5_string((unsigned char*)(Buffer));

			if(HashCertificate != HashFromFileCertificate)
			{
				Logs.Write("System", "bad certificate", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("bad certificate...\n");
				bSuccess = false;
			}
			if(HashKey != HashFromFileKey)
			{
				Logs.Write("System", "bad private key", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("bad private key...\n");
				bSuccess = false;
			}

			if(bSuccess)
			{
				BinaryCertificateFile.Save(CertificateFile.c_str());
				BinaryKeyFile.Save(KeyFile.c_str());
			}
		}


		if(!HttpServerSSL.InitSSL(CertificateFile.c_str(), KeyFile.c_str(), m_bSSLV3Only, m_bHttpsUseClientCertificate, m_sHttpsCAFile.c_str()))
		{
			if(b_ConsoleMode)
			{
				Logs.Write("System", "Error checking certificates", "", NV_LOG_ERROR, 1, "", "system");
				printf("Error checking certificates\n");
				printf("The HTTPS server cannot be started\n");
			}
		}
		else
		{
			if(!HttpServerSSL.Run())
			{
				Logs.Write("System", "Cannot start HTTPS server", "", NV_LOG_ERROR, 1, "", "system");
				if(b_ConsoleMode)
					printf("Cannot start HTTPS server\n");
			}
		}
		RemoveFile(CertificateFile.c_str());
		RemoveFile(KeyFile.c_str());
	}

	if(m_bMqAvailable)
	{
		Logs.Write("System", "Starting queues", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Starting queues...\n");
		if(!Queues.AutoStartQueues())
		{
			Logs.Write("System", "Cannot start queues. Please check MQSeries installation", "", NV_LOG_ERROR, 1, "", "system");
			if(b_ConsoleMode)
				printf("Cannot start queues. Please check MQSeries installation\n");
		}
	}



	if(m_bUseTransactionServer)
	{
		TransactionServer.SetMaxTransactions(m_iMaxTransactions);
		TransactionServer.SetCheckEvery(m_iTransactionCheckEvery);

		Logs.Write("System", "Starting transaction server", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Starting transaction server...\n");
		// starts the transaction server object
		if(!TransactionServer.Run())
		{
			Logs.Write("System", "Cannot start transaction server", "", NV_LOG_ERROR, 1, "", "system");
			if(b_ConsoleMode)
				printf("Cannot start transaction server\n");
			// we continue anyway
		}
	}

	SchedulerServer.SetMaxTasks(m_iMaxSchedulerTasks);
	SchedulerServer.SetCheckEvery(m_iSchedulerCheckEvery);

	Logs.Write("System", "Starting scheduler", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Starting scheduler...\n");
	// starts the scheduler server object
	if(!SchedulerServer.Run())
	{
		Logs.Write("System", "Cannot start scheduler", "", NV_LOG_ERROR, 1, "", "system");
		if(b_ConsoleMode)
			printf("Cannot start scheduler\n");
		// we continue anyway
	}

	Logs.Write("System", "Starting listeners", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("Starting listeners...\n");
	// starts the listeners for all applications
	Applications.StartListeners();

	Logs.Write("System", "NIRVA server ready", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("\nNIRVA server ready\n");

	SetStatus("RUNNING");

	b_Initialized = true;

	if(ClusterMode)
	{
		// In cluster mode we call a trigger procedure in the system Procs named failover_start.nvp
		NvSession Session;
		Session.Init();
		Session.SetAppName("NVDEF");
		Session.Application = DefaultApplication;
		Session.m_bAdministrator = true;

		NvCommand Command;
		Session.OriginalCommand = &Command;
		Command.m_pSession = &Session;
		Command.m_iCommandSource = NV_SOURCE_SERVER;
		Command.m_sService = "SYSTEM";
		Command.m_sClass = "MISC";
		Command.m_sCommand = "NOP";
		Session.ExecuteProcedure(&Command, "(system/failover_start)", false);
		Session.Application = NULL;
	}

	return true;
}



// called after the Main function (see class NvTask)
void NvsServer::OnCleanup()
{
	Logs.Write("System", "Stopping NIRVA server", "", NV_LOG_INFO, 1, "", "system");
	SetStopRequest();
	SetStatus("STOPPING");
	if(b_ConsoleMode)
		printf("\nStopping NIRVA server\n");

	if(b_Initialized)
	{
		

		// we have to wait for all clients to close themselves
		// and them to hardly close the rest


		// PM Stop the listeners, the sheduler and the transaction server before stopping http server

		// Stop listeners
		Logs.Write("System", "Stopping listeners", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Stopping listeners...\n");

		try
		{
			Applications.StopListeners();
		}
		catch(...)
		{
		}

		// Stop scheduler server
		Logs.Write("System", "Stopping scheduler", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Stopping scheduler...\n");

		try
		{
			SchedulerServer.Stop();
			while(SchedulerServer.IsThreadRunning())	// wait for http server to end
				SLEEP(200);
		}
		catch(...)
		{
		}


		if(TransactionServer.IsThreadRunning())
		{
			// Stop transaction server
			Logs.Write("System", "Stopping transaction server", "", NV_LOG_INFO, 1, "", "system");
			if(b_ConsoleMode)
				printf("Stopping transaction server...\n");
			try
			{
				TransactionServer.Stop();
				while(TransactionServer.IsThreadRunning())	// wait for http server to end
					SLEEP(200);
			}
			catch(...)
			{
			}
		}


		// Stop queues
		if(m_bMqAvailable)
		{
			Logs.Write("System", "Stopping queues", "", NV_LOG_INFO, 1, "", "system");
			if(b_ConsoleMode)
				printf("Stopping queues...\n");
			Queues.StopQueues();
		}


		// Stop HTTP server
		Logs.Write("System", "Stopping HTTP server", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Stopping HTTP server...\n");
		try
		{
			HttpServer.Stop();
			while(HttpServer.IsThreadRunning())	// wait for http server to end
				SLEEP(200);
		}
		catch(...)
		{
		}


		// Stop HTTPS server
		if(m_bUseSSL)
		{
			Logs.Write("System", "Stopping HTTPS server", "", NV_LOG_INFO, 1, "", "system");
			if(b_ConsoleMode)
				printf("Stopping HTTPS server...\n");
			try
			{
				HttpServerSSL.Stop();
				while(HttpServerSSL.IsThreadRunning())	// wait for https server to end
					SLEEP(200);
			}
			catch(...)
			{
			}
		}
		

		// Clear sessions
		Logs.Write("System", "Clearing running sessions", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Clearing running sessions...\n");
		try
		{
			Sessions.Clear();	
		}
		catch(...)
		{
		}
		

		// Stopping applications
		Logs.Write("System", "Stopping applications", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Stopping applications...\n");
		try
		{
			Applications.Stop();	
		}
		catch(...)
		{
		}

		

		// Clear all sessions by security
		Logs.Write("System", "Clearing pending sessions", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Clearing pending sessions...\n");
		try
		{
			Sessions.Clear();	
		}
		catch(...)
		{
		}


		// Clear connections to other NIRVAs
		Logs.Write("System", "Clearing connections to other NIRVA servers", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Clearing connections to other NIRVA servers...\n");
		try
		{
			Connections.Clear();
		}
		catch(...)
		{
		}


		// Unload applications except the default one
		Logs.Write("System", "Unloading applications", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unloading applications...\n");
		try
		{
			Applications.Clear(false);	
		}
		catch(...)
		{
		}


		Logs.Write("System", "Unloading services", "", NV_LOG_INFO, 1, "", "system");
		// Unload services
		if(b_ConsoleMode)
			printf("Unloading services...\n");
		try
		{
			Services.Clear();	
		}
		catch(...)
		{
		}


		// Clears applications including the default one (only the default one still exists)
		Logs.Write("System", "Unloading default application", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unloading default application...\n");
		try
		{
			Applications.Clear();	
		}
		catch(...)
		{
		}


		// Clears web services
		Logs.Write("System", "Stopping web services", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Stopping web services...\n");
		try
		{
			Webservices.Clear();	
		}
		catch(...)
		{
		}
	
		// Clears the system test sets
		Logs.Write("System", "Unload test sets", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unload test sets...\n");
		try
		{
			TestSets.Clear();
		}
		catch(...)
		{
		}

		// Clears the system registry
		Logs.Write("System", "Freeing registry cache", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Freeing registry cache...\n");
		try
		{
			SystemRegistry.Clear();
		}
		catch(...)
		{
		}
		

		// Removes all temporary files
		Logs.Write("System", "Removing temporary files", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Removing temporary files...\n");
		try
		{
			Garbage->Clear();
		}
		catch(...)
		{
		}


		// Unlock all locks
		Logs.Write("System", "Unlocking objects", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unlocking objects...\n");
		try
		{
			LockList.Clear();			// Clears the lock list at the end because it can be used
		}
		catch(...)
		{
		}


		// Unlock all semaphores
		Logs.Write("System", "Unlocking semaphores", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unlocking semaphores...\n");
		try
		{
			SemaList.Clear();			// Clears the sema list at the end because it can be used
		}
		catch(...)
		{
		}
									// by some other objects


		// Unlock all filter locks
		Logs.Write("System", "Unlocking executables", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unlocking executables...\n");
		try
		{
			FilterLockList.Clear();		// Clears the filter lock list
		}
		catch(...)
		{
		}


		// Unload the perl interface library if loaded
		Logs.Write("System", "Unloading perl interface", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unloading perl interface...\n");
		try
		{
			PerlInterface.FreePerlLib();
		}
		catch(...)
		{
		}
		

		// Unload the java interface library if loaded
		Logs.Write("System", "Unloading java virtual machine", "", NV_LOG_INFO, 1, "", "system");
		if(!NoJava)
			if(b_ConsoleMode)
				printf("Unloading java virtual machine...\n");
		try
		{
			JavaInterface.FreeJavaLib();
		}
		catch(...)
		{
		}
		
		// Unload the dotnet interface library if loaded
		Logs.Write("System", "Unloading .net interface", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unloading .net interface...\n");
		try
		{
			DotnetInterface.FreeDotnetLib();
		}
		catch(...)
		{
		}

		// Unload the mq interface library if loaded
		if(m_bMqAvailable)
		{
			Logs.Write("System", "Unloading MQ connector", "", NV_LOG_INFO, 1, "", "system");
			if(b_ConsoleMode)
				printf("Unloading MQ connector...\n");
			try
			{
				MqInterface.FreeMqLib();
			}
			catch(...)
			{
			}
		}
		
		
		// Unload the license interface library if loaded
		if(m_bLcAvailable)
		{
			Logs.Write("System", "Unloading license library", "", NV_LOG_INFO, 1, "", "system");
			if(b_ConsoleMode)
				printf("Unloading license library...\n");
			try
			{
				Licenses.Close();
				LcInterface.FreeLcLib();
			}
			catch(...)
			{
			}
		}

		// Free XML library
		Logs.Write("System", "Unloading XML engine", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Unloading XML engine...\n");
		try
		{
			CleanupXml();
		}
		catch(...)
		{
		}

		// Free Sqlite library
		try
		{
			SqLiteInterface.FreeSqLiteLib();
		}
		catch(...)
		{
		}

		// Clear system containers
		Logs.Write("System", "Clearing system containers", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Clearing system containers...\n");
		Containers.Clear();		
		
		// Clear all user registries at system level
		Logs.Write("System", "Clearing system user registries", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Clearing system user registries...\n");
		Registries.Clear();		

		// Delete Garbage collector object
		Logs.Write("System", "Freeing garbage collector", "", NV_LOG_INFO, 1, "", "system");
		if(b_ConsoleMode)
			printf("Freeing garbage collector...\n");
		try
		{
			delete Garbage;
		}
		catch(...)
		{
		}
		Garbage = NULL;

		string RunFile = GetNirvaDir();
		RunFile += "Bin";
		AddDirSuffix(RunFile);
		RunFile += "nirva.run";
		
		RemoveFile(RunFile.c_str()); // removes the run file when stopping the server
	}


	Logs.Write("System", "NIRVA successfuly stopped", "", NV_LOG_INFO, 1, "", "system");
	if(b_ConsoleMode)
		printf("\nNIRVA successfuly stopped\n");

	if(Garbage != NULL)
	{
		try
		{
			Garbage->Clear();
			delete Garbage;
		}
		catch(...)
		{
		}
	}
	Garbage = NULL;

	// to call before leaving
	Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}






// Load server information from dsc file
bool NvsServer::LoadInfo()
{
	string DescFileName = m_sNirvaDir;
	AddDirSuffix(DescFileName);
	DescFileName += "Files";
	AddDirSuffix(DescFileName);

	DescFileName += "nirva.dsc";

	if(GetNoCaseRealFileName(DescFileName))
	{
		if(FileExist(DescFileName.c_str()))
		{
			if(!Info.Load(DescFileName.c_str()))
			{
				if(b_ConsoleMode)
					printf("Cannot load Nirva description file\n");
				return false;
			}
		}
	}

	return true;
}



// Load basic server parameters from registry (paths for registry, appwork and logs)
bool NvsServer::LoadBasicParameters(bool Default)
{
	NV_LOCKOBJECT(this);
	m_sRegistryHomeDir = m_sNirvaDir;
	m_sLogHomeDir = m_sNirvaDir;
	m_sAppWorkHomeDir = m_sNirvaDir;

	if(!Default)
	{
		// load the parameters
		// The parameters are stored in an indexed string list object named "parameters" in the
		// system registry key system.parameters
		NvContainer Container;
		if(BasicSystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "basics", false, true))
		{
			NvObj *Object = Container.GetObject("basics");
			if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
			{
				NvObjIndStringList *Parameters = (NvObjIndStringList*)Object;
				string Value;
				
				
				// Get registry home directory
				Value = Parameters->GetValue("RegHome"); 
				if(Value.length())
					m_sRegistryHomeDir = Value;
				CleanString(m_sRegistryHomeDir);
				AddDirSuffix(m_sRegistryHomeDir);
				
				// Get log home directory
				Value = Parameters->GetValue("LogHome"); 
				if(Value.length())
					m_sLogHomeDir = Value;
				CleanString(m_sLogHomeDir);
				AddDirSuffix(m_sLogHomeDir);
				
				// Get application work home directory
				Value = Parameters->GetValue("AppWorkHome"); 
				if(Value.length())
					m_sAppWorkHomeDir = Value;
				CleanString(m_sAppWorkHomeDir);
				AddDirSuffix(m_sAppWorkHomeDir);
				
			}
		}
	}

	// also load the uniq counter who is in a specific object
	NvContainer Container;
	m_iUniqCounter = 0;
	if(BasicSystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "uniqcount", false, true))
	{
		NvObj *Object = Container.GetObject("uniqcount");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INTEGER))
		{
			NvObjInteger *Uniq = (NvObjInteger*)Object;
			m_iUniqCounter = (unsigned int)(Uniq->GetValue());
		}
	}

	NV_UNLOCKOBJECT(this);

	return true;
}



// Save basic server parameters to registry
// called from internal so no lock
string NvsServer::GetUniqString()
{
	NV_LOCKOBJECT(this);
	m_iUniqCounter++;

	// The uniq string is based on the uniq counter, the machine name and a random value
	char Buffer[256];
	sprintf(Buffer, "%s_%08x_%08x_%08x", m_sMachineName.c_str(), m_iUniqCounter, rand(), getpid());
	string Retval = MD5_string(Buffer);
	MakeUpper(Retval);
	
	// Save the new counter value to registry
	NvContainer Container;
	Container.CreateObject("uniqcount", NV_OBJTYPE_INTEGER);
	NvObjInteger *Uniq = (NvObjInteger*)(Container.GetObject("uniqcount"));
	if(Uniq != NULL)
	{
		Uniq->SetValue(m_iUniqCounter);
		BasicSystemRegistry.SetKey("system.parameters", &Container);
	}
	NV_UNLOCKOBJECT(this);
	return Retval;
}



// Save basic server parameters to registry
// called from internal so no lock
bool NvsServer::SaveBasicParameters()
{
	NvContainer Container;
	Container.CreateObject("basics", NV_OBJTYPE_INDSTRINGLIST);
	NvObjIndStringList *Parameters = (NvObjIndStringList*)(Container.GetObject("basics"));
	if(Parameters == NULL)
		return false;
	Parameters->SetValue("RegHome", m_sNewRegistryHomeDir);
	Parameters->SetValue("LogHome", m_sNewLogHomeDir);
	Parameters->SetValue("AppWorkHome", m_sNewAppWorkHomeDir);
	
	if(!BasicSystemRegistry.SetKey("system.parameters", &Container))
		return false;
	return true;
}


// Load server parameters from registry
bool NvsServer::LoadParameters(bool Default, bool NoSave)
{
	NV_LOCKOBJECT(this);
	m_iHttpServerPort = 1081;
	m_iHttpsServerPort = 1082;
	m_iHttpSessionTimeOut = 120;
	m_iHttpStopWaitTime = 120;
	m_iHttpMaxMemBody = 1024*1024;
	m_bUseSSL = false;
	m_bSSLV3Only = false;
	m_bHttpLocal = false;
	m_bHttpsUseClientCertificate = false;
	m_sHttpsCAFile = "";
	m_bAutoDetectEncoding = false;
	m_sNirvaDefaultPage = "default.htm";
	m_bDebug = false;
	m_bRemoteAdmin = true;
	m_bInternalXsl = true;
	m_bEnableXslCache = false;
	m_bLongSessionId = false;
	m_bOldObjectsUnicode = false;
	m_iXslMaxDepth = 5000;
	m_iXslParserMaxDepth = 1024;

	m_sXslProcessorPath = m_sNirvaDir;
#ifdef WIN32
	m_sXslProcessorPath += "";
#else
//	m_sXslProcessorPath = "java org.apache.xalan.xslt.Process -IN %INPUTFILE -XSL %XSLFILE -OUT %OUTPUTFILE";
//	m_sXslProcessorPath += "Bin/Xalan/Xalan -o %OUTPUTFILE %INPUTFILE %XSLFILE";
//	m_sXslProcessorPath = "xsltproc -o %OUTPUTFILE %XSLFILE %INPUTFILE";
	m_sXslProcessorPath = "";
#endif
	m_bLocalTime = false;
	m_bInternalJvm = true;
	m_JVMMinMemSize = 4;
	m_JVMMaxMemSize = 64;
	m_sJVMParam = "";
	m_sJVMHomeDir = "";
	m_sJVMPathName = "";
	m_bUseTransactionServer = false;
	m_iMaxTransactions = 10;
	m_iTransactionCheckEvery = 2;
	m_iMaxSchedulerTasks = 10;
	m_iSchedulerCheckEvery = 2;
	m_iDefaultAppTimeOut = 300;
	m_iDefaultAppAuthMode = NV_AUTHENTICATION_NIRVA;
	m_bDefaultNirvaIsSsoUser = false;
	m_sDefaultAppSecuApp = "";
	m_bDefaultAppUseScheduled = true;
	m_bDefaultAppAllowWebXml = true;
	m_bDefaultAppAllowNotOwnedSession = true;
	m_sAdminPassword = MD5_string("nirva");
	MakeUpper(m_sAdminPassword);
	m_sDefaultMailServer = "localhost";
	m_sDefaultMailUser = "nvadmin";
	m_sDefaultMailPassword = "nirva";
	m_sHostWebs = m_sMachineName;
	if(!m_sHostWebs.length())
		m_sHostWebs = "127.0.0.1";
	m_sHostWebs += ":1081";
	MakeLower(m_sHostWebs);

	if(!Default)
	{
		// load the parameters
		// The parameters are stored in an indexed string list object named "parameters" in the
		// system registry key system.parameters
		NvContainer Container;
		if(SystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "parameters", false, true))
		{
			NvObj *Object = Container.GetObject("parameters");
			if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
			{
				NvObjIndStringList *Parameters = (NvObjIndStringList*)Object;
				string Value;
				
				// Get HTTP server port
				Value = Parameters->GetValue("HttpServerPort"); 
				if(Value.length())
					m_iHttpServerPort = atoi(Value.c_str());
				if(m_iHttpServerPort < 1)
					m_iHttpServerPort = 1;
				
				// Get HTTPS server port
				Value = Parameters->GetValue("HttpsServerPort"); 
				if(Value.length())
					m_iHttpsServerPort = atoi(Value.c_str());
				else
					m_iHttpsServerPort = m_iHttpServerPort+1;
				if(m_iHttpsServerPort < 1)
					m_iHttpsServerPort = 1;
				
				// Get HTTP session time out
				Value = Parameters->GetValue("HttpSessionTimeOut"); 
				if(Value.length())
					m_iHttpSessionTimeOut = atoi(Value.c_str());
				if(m_iHttpSessionTimeOut < 10)
					m_iHttpSessionTimeOut = 10;
				
				// Get HTTP stop wait time
				Value = Parameters->GetValue("HttpStopWaitTime"); 
				if(Value.length())
					m_iHttpStopWaitTime = atoi(Value.c_str());
				if(m_iHttpStopWaitTime < 5)
					m_iHttpStopWaitTime = 5;
				
				// Get HTTP max mem body
				Value = Parameters->GetValue("HttpMaxMemBody"); 
				if(Value.length())
					m_iHttpMaxMemBody = atoi(Value.c_str());
				if(m_iHttpMaxMemBody < 10)
					m_iHttpMaxMemBody = 10;
				m_iHttpMaxMemBody *= 1024;	// It's in kilobytes
				
				// Get SSL mode
				Value = Parameters->GetValue("UseSSL"); 
				if(Value.length() && (Value == "1"))
					m_bUseSSL = true;
				
				// Get SSLV3Only mode
				Value = Parameters->GetValue("SSLV3Only"); 
				if(Value.length() && (Value == "1"))
					m_bSSLV3Only = true;
				
				// Get HttpLocal mode
				Value = Parameters->GetValue("HttpLocal"); 
				if(Value.length() && (Value == "1"))
					m_bHttpLocal = true;
				
				// Get HttpsUseClientCertificate mode
				Value = Parameters->GetValue("HttpsUseClientCertificate"); 
				if(Value.length() && (Value == "1"))
					m_bHttpsUseClientCertificate = true;
				
				// Get CA file
				Value = Parameters->GetValue("HttpsCAFile"); 
				if(Value.length())
					m_sHttpsCAFile = Value;
				
				// Get auto detect encoding mode
				Value = Parameters->GetValue("AutoDetectEncoding"); 
				if(Value.length() && (Value == "1"))
					m_bAutoDetectEncoding = true;
				
				// Get Nirva default page
				Value = Parameters->GetValue("NirvaDefaultPage"); 
				if(Value.length())
					m_sNirvaDefaultPage = Value;
				
				// Get Debug mode
				Value = Parameters->GetValue("Debug"); 
				if(Value.length() && (Value == "1"))
					m_bDebug = true;
				
				// Get RemoteAdmin mode
				Value = Parameters->GetValue("RemoteAdmin"); 
				if(Value.length() && (Value == "0"))
					m_bRemoteAdmin = false;
				
				// Get long session ID flag
				Value = Parameters->GetValue("LongSessionId"); 
				if(Value.length() && (Value == "1"))
					m_bLongSessionId = true;
				
				// Get old objects unicode flag
				Value = Parameters->GetValue("OldObjectsUnicode"); 
				if(Value.length() && (Value == "1"))
				{
					m_bOldObjectsUnicode = true;
					bOldObjectsUnicode = true;	// For access from intdb
				}

				// Get internal XSL mode
				Value = Parameters->GetValue("InternalXsl"); 
				if(Value.length() && (Value == "0"))
					m_bInternalXsl = false;

				// Get cache XSL mode
				Value = Parameters->GetValue("EnableXslCache"); 
				if(Value.length() && (Value == "1"))
					m_bEnableXslCache = true;
				
				// Get XSL max depth
				Value = Parameters->GetValue("XslMaxDepth"); 
				if(Value.length())
					m_iXslMaxDepth = atoi(Value.c_str());
				// Minimum is 5000
				if(m_iXslMaxDepth < 5000)
					m_iXslMaxDepth = 5000;
				
				// Get XSL parser max depth
				Value = Parameters->GetValue("XslParserMaxDepth"); 
				if(Value.length())
					m_iXslParserMaxDepth = atoi(Value.c_str());
				// Minimum is 1024
				if(m_iXslParserMaxDepth < 1024)
					m_iXslParserMaxDepth = 1024;
				
				// Get XSLT path
				Value = Parameters->GetValue("XsltProcessor"); 
				if(Value.length())
					m_sXslProcessorPath = Value;
				
				
				// Get local time mode
				Value = Parameters->GetValue("LocalTime"); 
				if(Value.length() && (Value == "1"))
					m_bLocalTime = true;
			
				// Get internal jvm flag
				Value = Parameters->GetValue("InternalJvm"); 
				if(Value.length() && (Value == "0"))
					m_bInternalJvm = false;

				// Get JVM allocation pool min mem size
				Value = Parameters->GetValue("JVMMinMemSize"); 
				if(Value.length())
					m_JVMMinMemSize = atoi(Value.c_str());
				if(m_JVMMinMemSize < 4)
					m_JVMMinMemSize = 4;
				
				// Get JVM allocation pool max mem size
				Value = Parameters->GetValue("JVMMaxMemSize"); 
				if(Value.length())
					m_JVMMaxMemSize = atoi(Value.c_str());
				if(m_JVMMaxMemSize < m_JVMMinMemSize)
					m_JVMMaxMemSize = m_JVMMinMemSize;
		
				// Get JVM extra parameters
				Value = Parameters->GetValue("JVMParam"); 
				if(Value.length())
					m_sJVMParam = Value;
		
				// Get m_sJVMHomeDir
				Value = Parameters->GetValue("JVMHomeDir"); 
				if(Value.length())
					m_sJVMHomeDir = Value;
		
				// Get m_sJVMPathName
				Value = Parameters->GetValue("JVMPathName"); 
				if(Value.length())
					m_sJVMPathName = Value;
			
				// Get use transaction server
				Value = Parameters->GetValue("UseTransactionServer"); 
				if(Value.length() && (Value == "1"))
					m_bUseTransactionServer = true;

				// Get max simultaneous transactions
				Value = Parameters->GetValue("MaxTransactions"); 
				if(Value.length())
					m_iMaxTransactions = atoi(Value.c_str());
				if(m_iMaxTransactions < 1)
					m_iMaxTransactions = 1;
				if(m_iMaxTransactions > 500)
					m_iMaxTransactions = 500;
				
				// Get transaction check every
				Value = Parameters->GetValue("TransactionCheckEvery"); 
				if(Value.length())
					m_iTransactionCheckEvery = atoi(Value.c_str());
				if(m_iTransactionCheckEvery < 1)
					m_iTransactionCheckEvery = 1;
			
				// Get max simultaneous scheduled tasks
				Value = Parameters->GetValue("MaxSchedulerTasks"); 
				if(Value.length())
					m_iMaxSchedulerTasks = atoi(Value.c_str());
				if(m_iMaxSchedulerTasks < 1)
					m_iMaxSchedulerTasks = 1;
				if(m_iMaxSchedulerTasks > 500)
					m_iMaxSchedulerTasks = 500;
				
				// Get check every
				Value = Parameters->GetValue("SchedulerCheckEvery"); 
				if(Value.length())
					m_iSchedulerCheckEvery = atoi(Value.c_str());
				if(m_iSchedulerCheckEvery < 1)
					m_iSchedulerCheckEvery = 1;
				
				// Get m_iDefaultAppTimeOut
				Value = Parameters->GetValue("DefaultAppTimeOut"); 
				if(Value.length())
					m_iDefaultAppTimeOut = atoi(Value.c_str());
				if(m_iDefaultAppTimeOut < 30)
					m_iDefaultAppTimeOut = 30;
				
				// Get m_iDefaultAppAuthMode
				Value = Parameters->GetValue("DefaultAppAuthMode"); 
				if(Value.length())
					m_iDefaultAppAuthMode = atoi(Value.c_str());
				if(m_iDefaultAppAuthMode < 0)
					m_iDefaultAppAuthMode = NV_AUTHENTICATION_NIRVA;
				if(m_iDefaultAppAuthMode > NV_AUTHENTICATION_MAXVALUE)
					m_iDefaultAppAuthMode = NV_AUTHENTICATION_NIRVA;
				
			
				// Get DefaultAppNirvaIsSsoUser
				Value = Parameters->GetValue("DefaultAppNirvaIsSsoUser"); 
				if(Value.length() && (Value == "1"))
					m_bDefaultNirvaIsSsoUser = true;
			
				// Get m_sDefaultAppSecuApp
				Value = Parameters->GetValue("DefaultAppSecuApp"); 
				if(Value.length())
					m_sDefaultAppSecuApp = Value;
				
				// Get m_bDefaultAppUseScheduled
				Value = Parameters->GetValue("DefaultAppUseScheduled"); 
				MakeUpper(Value);
				if(Value == "0")
					m_bDefaultAppUseScheduled = false;
				
				// Get m_bDefaultAppAllowWebXml
				Value = Parameters->GetValue("DefaultAppAllowWebXml"); 
				MakeUpper(Value);
				if(Value == "0")
					m_bDefaultAppAllowWebXml = false;
				
				// Get m_bDefaultAppAllowNotOwnedSession
				Value = Parameters->GetValue("DefaultAppAllowNotOwnedSession"); 
				MakeUpper(Value);
				if(Value == "0")
					m_bDefaultAppAllowNotOwnedSession = false;
			
				// Get m_sAdminPassword
				Value = Parameters->GetValue("AdminPassword"); 
				if(Value.length())
					m_sAdminPassword = Value;
			
				// Get m_sDefaultMailServer
				Value = Parameters->GetValue("DefaultMailServer"); 
				if(Value.length())
					m_sDefaultMailServer = Value;
			
				// Get m_sDefaultMailUser
				Value = Parameters->GetValue("DefaultMailUser"); 
				if(Value.length())
					m_sDefaultMailUser = Value;
			
				// Get m_sDefaultMailPassword
				Value = Parameters->GetValue("DefaultMailPassword"); 
				if(Value.length())
					m_sDefaultMailPassword = Value;
			
				// Get m_sHostWebs
				Value = Parameters->GetValue("HostWebs"); 
				m_sHostWebs = Value;
				if(!m_sHostWebs.length())
				{
					m_sHostWebs = m_sMachineName;
					if(!m_sHostWebs.length())
						m_sHostWebs = "127.0.0.1";
					if(m_iHttpServerPort != 80)
					{
						char Buf[20];
						sprintf(Buf, ":%d", m_iHttpServerPort);
						m_sHostWebs += Buf;
					}
					MakeLower(m_sHostWebs);
				}
			}
		}
	}

	NV_UNLOCKOBJECT(this);
	// Saves the list in case of changes
	if(!NoSave)
		SaveParameters();

	return true;
}


// Save server parameters to registry
bool NvsServer::SaveParameters()
{
	NvContainer Container;
	Container.CreateObject("parameters", NV_OBJTYPE_INDSTRINGLIST);
	NvObjIndStringList *Parameters = (NvObjIndStringList*)(Container.GetObject("parameters"));
	if(Parameters == NULL)
		return false;
	NV_LOCKOBJECT(this);

	char Buff[100];
	sprintf(Buff, "%d", m_iHttpServerPort);
	Parameters->SetValue("HttpServerPort", Buff);
	sprintf(Buff, "%d", m_iHttpsServerPort);
	Parameters->SetValue("HttpsServerPort", Buff);
	sprintf(Buff, "%d", m_iHttpSessionTimeOut);
	Parameters->SetValue("HttpSessionTimeOut", Buff);
	sprintf(Buff, "%d", m_iHttpStopWaitTime);
	Parameters->SetValue("HttpStopWaitTime", Buff);
	sprintf(Buff, "%d", m_iHttpMaxMemBody/1024);	// In kilobytes
	Parameters->SetValue("HttpMaxMemBody", Buff);
	if(m_bUseSSL)
		Parameters->SetValue("UseSSL", "1");
	else
		Parameters->SetValue("UseSSL", "0");
	if(m_bSSLV3Only)
		Parameters->SetValue("SSLV3Only", "1");
	else
		Parameters->SetValue("SSLV3Only", "0");
	if(m_bHttpLocal)
		Parameters->SetValue("HttpLocal", "1");
	else
		Parameters->SetValue("HttpLocal", "0");
	if(m_bHttpsUseClientCertificate)
		Parameters->SetValue("HttpsUseClientCertificate", "1");
	else
		Parameters->SetValue("HttpsUseClientCertificate", "0");
	Parameters->SetValue("HttpsCAFile", m_sHttpsCAFile);
	if(m_bAutoDetectEncoding)
		Parameters->SetValue("AutoDetectEncoding", "1");
	else
		Parameters->SetValue("AutoDetectEncoding", "0");
	Parameters->SetValue("NirvaDefaultPage", m_sNirvaDefaultPage);
	if(m_bDebug)
		Parameters->SetValue("Debug", "1");
	else
		Parameters->SetValue("Debug", "0");
	if(m_bRemoteAdmin)
		Parameters->SetValue("RemoteAdmin", "1");
	else
		Parameters->SetValue("RemoteAdmin", "0");
	if(m_bLongSessionId)
		Parameters->SetValue("LongSessionId", "1");
	else
		Parameters->SetValue("LongSessionId", "0");
	if(m_bOldObjectsUnicode)
		Parameters->SetValue("OldObjectsUnicode", "1");
	else
		Parameters->SetValue("OldObjectsUnicode", "0");
	if(m_bInternalXsl)
		Parameters->SetValue("InternalXsl", "1");
	else
		Parameters->SetValue("InternalXsl", "0");
	if(m_bEnableXslCache)
		Parameters->SetValue("EnableXslCache", "1");
	else
		Parameters->SetValue("EnableXslCache", "0");
	sprintf(Buff, "%d", m_iXslMaxDepth);
	Parameters->SetValue("XslMaxDepth", Buff);
	sprintf(Buff, "%d", m_iXslParserMaxDepth);
	Parameters->SetValue("XslParserMaxDepth", Buff);
	Parameters->SetValue("XsltProcessor", m_sXslProcessorPath);
	if(m_bLocalTime)
		Parameters->SetValue("LocalTime", "1");
	else
		Parameters->SetValue("LocalTime", "0");
	if(m_bInternalJvm)
		Parameters->SetValue("InternalJvm", "1");
	else
		Parameters->SetValue("InternalJvm", "0");
	Parameters->SetValue("JVMHomeDir", m_sJVMHomeDir);
	Parameters->SetValue("JVMPathName", m_sJVMPathName);
	sprintf(Buff, "%d", m_JVMMinMemSize);
	Parameters->SetValue("JVMMinMemSize", Buff);
	sprintf(Buff, "%d", m_JVMMaxMemSize);
	Parameters->SetValue("JVMMaxMemSize", Buff);
	Parameters->SetValue("JVMParam", m_sJVMParam);
	if(m_bUseTransactionServer)
		Parameters->SetValue("UseTransactionServer", "1");
	else
		Parameters->SetValue("UseTransactionServer", "0");
	sprintf(Buff, "%d", m_iMaxTransactions);
	Parameters->SetValue("MaxTransactions", Buff);
	sprintf(Buff, "%d", m_iTransactionCheckEvery);
	Parameters->SetValue("TransactionCheckEvery", Buff);
	sprintf(Buff, "%d", m_iMaxSchedulerTasks);
	Parameters->SetValue("MaxSchedulerTasks", Buff);
	sprintf(Buff, "%d", m_iSchedulerCheckEvery);
	Parameters->SetValue("SchedulerCheckEvery", Buff);
	sprintf(Buff, "%d", m_iDefaultAppTimeOut);
	Parameters->SetValue("DefaultAppTimeOut", Buff);
	sprintf(Buff, "%d", m_iDefaultAppAuthMode);
	Parameters->SetValue("DefaultAppAuthMode", Buff);
	if(m_bDefaultNirvaIsSsoUser)
		Parameters->SetValue("DefaultAppNirvaIsSsoUser", "1");
	else
		Parameters->SetValue("DefaultAppNirvaIsSsoUser", "0");
	Parameters->SetValue("DefaultAppSecuApp", m_sDefaultAppSecuApp);
	if(m_bDefaultAppUseScheduled)
		Parameters->SetValue("DefaultAppUseScheduled", "1");
	else
		Parameters->SetValue("DefaultAppUseScheduled", "0");
	if(m_bDefaultAppAllowWebXml)
		Parameters->SetValue("DefaultAppAllowWebXml", "1");
	else
		Parameters->SetValue("DefaultAppAllowWebXml", "0");
	if(m_bDefaultAppAllowNotOwnedSession)
		Parameters->SetValue("DefaultAppAllowNotOwnedSession", "1");
	else
		Parameters->SetValue("DefaultAppAllowNotOwnedSession", "0");

	Parameters->SetValue("AdminPassword", m_sAdminPassword);
	Parameters->SetValue("DefaultMailServer", m_sDefaultMailServer);
	Parameters->SetValue("DefaultMailUser", m_sDefaultMailUser);
	Parameters->SetValue("DefaultMailPassword", m_sDefaultMailPassword);
	Parameters->SetValue("HostWebs", m_sHostWebs);
	
	if(!SystemRegistry.SetKey("system.parameters", &Container))
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}


// Load mime table from registry
void NvsServer::LoadMimeTable()
{
	NV_LOCKOBJECT(this);
	// load the mime table parameters
	// The mime table is stored in an indexed string list object named "mimes" in the
	// system registry key system.parameters
	NvContainer Container;
	if(SystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "mimes", false, true))
	{
		NvObj *Object = Container.GetObject("mimes");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
		{
			NvObjIndStringList *Mimes = (NvObjIndStringList*)Object;
			MimeTypes.CopyFrom(Mimes);
		}
	}

	// Set usual mime types:
	if(!MimeTypes.KeyExist("afp"))
		MimeTypes.SetValue("afp", "application/afp");
	if(!MimeTypes.KeyExist("ai"))
		MimeTypes.SetValue("ai", "application/postscript");
	if(!MimeTypes.KeyExist("aif"))
		MimeTypes.SetValue("aif", "audio/x-aiff");
	if(!MimeTypes.KeyExist("aifc"))
		MimeTypes.SetValue("aifc", "audio/x-aiff");
	if(!MimeTypes.KeyExist("aiff"))
		MimeTypes.SetValue("aiff", "audio/x-aiff");
	if(!MimeTypes.KeyExist("au"))
		MimeTypes.SetValue("au", "audio/basic");
	if(!MimeTypes.KeyExist("avi"))
		MimeTypes.SetValue("avi", "video/x-msvideo");
	if(!MimeTypes.KeyExist("bmp"))
		MimeTypes.SetValue("bmp", "image/bmp");
	if(!MimeTypes.KeyExist("bz2"))
		MimeTypes.SetValue("bz2", "application/x-bzip2");
	if(!MimeTypes.KeyExist("css"))
		MimeTypes.SetValue("css", "text/css");
	if(!MimeTypes.KeyExist("dtd"))
		MimeTypes.SetValue("dtd", "application/xml-dtd");
	if(!MimeTypes.KeyExist("doc"))
		MimeTypes.SetValue("doc", "application/msword");
	if(!MimeTypes.KeyExist("eps"))
		MimeTypes.SetValue("eps", "application/postscript");
	if(!MimeTypes.KeyExist("exe"))
		MimeTypes.SetValue("exe", "application/octet-stream");
	if(!MimeTypes.KeyExist("gif"))
		MimeTypes.SetValue("gif", "image/gif");
	if(!MimeTypes.KeyExist("gz"))
		MimeTypes.SetValue("gz", "application/x-gzip");
	if(!MimeTypes.KeyExist("hqx"))
		MimeTypes.SetValue("hqx", "application/mac-binhex40");
	if(!MimeTypes.KeyExist("html"))
		MimeTypes.SetValue("html", "text/html");
	if(!MimeTypes.KeyExist("htm"))
		MimeTypes.SetValue("htm", "text/html");
	if(!MimeTypes.KeyExist("ief"))
		MimeTypes.SetValue("ief", "image/ief");
	if(!MimeTypes.KeyExist("jar"))
		MimeTypes.SetValue("jar", "application/java-archive");
	if(!MimeTypes.KeyExist("jpg"))
		MimeTypes.SetValue("jpg", "image/jpeg");
	if(!MimeTypes.KeyExist("jpeg"))
		MimeTypes.SetValue("jpeg", "image/jpeg");
	if(!MimeTypes.KeyExist("jpe"))
		MimeTypes.SetValue("jpe", "image/jpeg");
	if(!MimeTypes.KeyExist("js"))
		MimeTypes.SetValue("js", "application/x-javascript");
	if(!MimeTypes.KeyExist("midi"))
		MimeTypes.SetValue("midi", "audio/x-midi");
	if(!MimeTypes.KeyExist("mp3"))
		MimeTypes.SetValue("mp3", "audio/mpeg");
	if(!MimeTypes.KeyExist("mpe"))
		MimeTypes.SetValue("mpe", "video/mpeg");
	if(!MimeTypes.KeyExist("mpeg"))
		MimeTypes.SetValue("mpeg", "video/mpeg");
	if(!MimeTypes.KeyExist("mpg"))
		MimeTypes.SetValue("mpg", "video/mpeg");
	if(!MimeTypes.KeyExist("mov"))
		MimeTypes.SetValue("mov", "video/quicktime");
	if(!MimeTypes.KeyExist("movie"))
		MimeTypes.SetValue("movie", "video/x-sgi-movie");
	if(!MimeTypes.KeyExist("ogg"))
		MimeTypes.SetValue("ogg", "audio/vorbis");
	if(!MimeTypes.KeyExist("pcl"))
		MimeTypes.SetValue("pcl", "application/pcl");
	if(!MimeTypes.KeyExist("pdf"))
		MimeTypes.SetValue("pdf", "application/pdf");
	if(!MimeTypes.KeyExist("pl"))
		MimeTypes.SetValue("pl", "application/x-perl");
	if(!MimeTypes.KeyExist("png"))
		MimeTypes.SetValue("png", "image/png");
	if(!MimeTypes.KeyExist("ppt"))
		MimeTypes.SetValue("ppt", "application/vnd.ms-powerpoint");
	if(!MimeTypes.KeyExist("ps"))
		MimeTypes.SetValue("ps", "application/postscript");
	if(!MimeTypes.KeyExist("qt"))
		MimeTypes.SetValue("qt", "video/quicktime");
	if(!MimeTypes.KeyExist("ra"))
		MimeTypes.SetValue("ra", "audio/x-pn-realaudio");
	if(!MimeTypes.KeyExist("ram"))
		MimeTypes.SetValue("ram", "audio/x-pn-realaudio");
	if(!MimeTypes.KeyExist("rdf"))
		MimeTypes.SetValue("rdf", "application/rdf");
	if(!MimeTypes.KeyExist("roff"))
		MimeTypes.SetValue("roff", "application/x-troff");
	if(!MimeTypes.KeyExist("rtf"))
		MimeTypes.SetValue("rtf", "application/rtf");
	if(!MimeTypes.KeyExist("sgml"))
		MimeTypes.SetValue("sgml", "text/sgml");
	if(!MimeTypes.KeyExist("sit"))
		MimeTypes.SetValue("sit", "application/x-stuffit");
	if(!MimeTypes.KeyExist("snd"))
		MimeTypes.SetValue("snd", "audio/basic");
	if(!MimeTypes.KeyExist("svg"))
		MimeTypes.SetValue("svg", "image/svg+xml");
	if(!MimeTypes.KeyExist("swf"))
		MimeTypes.SetValue("swf", "application/x-shockwave-flash");
	if(!MimeTypes.KeyExist("t"))
		MimeTypes.SetValue("t", "application/x-troff");
	if(!MimeTypes.KeyExist("tar.gz"))
		MimeTypes.SetValue("tar.gz", "application/x-tar");
	if(!MimeTypes.KeyExist("tex"))
		MimeTypes.SetValue("tex", "application/x-tex");
	if(!MimeTypes.KeyExist("texi"))
		MimeTypes.SetValue("texi", "application/x-texinfo");
	if(!MimeTypes.KeyExist("texinfo"))
		MimeTypes.SetValue("texinfo", "application/x-texinfo");
	if(!MimeTypes.KeyExist("tgz"))
		MimeTypes.SetValue("tgz", "application/x-tar");
	if(!MimeTypes.KeyExist("tif"))
		MimeTypes.SetValue("tif", "image/tif");
	if(!MimeTypes.KeyExist("tiff"))
		MimeTypes.SetValue("tiff", "image/tif");
	if(!MimeTypes.KeyExist("tr"))
		MimeTypes.SetValue("tr", "application/x-troff");
	if(!MimeTypes.KeyExist("tsv"))
		MimeTypes.SetValue("tsv", "text/tab-separated-values");
	if(!MimeTypes.KeyExist("txt"))
		MimeTypes.SetValue("txt", "text/plain");
	if(!MimeTypes.KeyExist("wav"))
		MimeTypes.SetValue("wav", "audio/x-wav");
	if(!MimeTypes.KeyExist("wsdl"))
		MimeTypes.SetValue("wsdl", "text/xml");
	if(!MimeTypes.KeyExist("xls"))
		MimeTypes.SetValue("xls", "application/vnd.ms-excel");
	if(!MimeTypes.KeyExist("xml"))
		MimeTypes.SetValue("xml", "text/xml");
	if(!MimeTypes.KeyExist("xsd"))
		MimeTypes.SetValue("xsd", "text/xml");
	if(!MimeTypes.KeyExist("xsl"))
		MimeTypes.SetValue("xsl", "text/xml");
	if(!MimeTypes.KeyExist("xwd"))
		MimeTypes.SetValue("xwd", "image/x-xwindowdump");
	if(!MimeTypes.KeyExist("zip"))
		MimeTypes.SetValue("zip", "application/zip");

	SaveMimeTable();


	NV_UNLOCKOBJECT(this);

}

// Save mime table to registry
// NvsServer must be locked when entering this function
void NvsServer::SaveMimeTable()
{
	NvContainer Container;
	Container.CreateObject("mimes", NV_OBJTYPE_INDSTRINGLIST);
	NvObjIndStringList *Mimes = (NvObjIndStringList*)(Container.GetObject("mimes"));
	if(Mimes == NULL)
		return;
	Mimes->CopyFrom(&MimeTypes);
	
	SystemRegistry.SetKey("system.parameters", &Container);
}


// Set a mime type
void NvsServer::SetMimeType(string Extension, string MimeType)
{
	string::size_type index = Extension.rfind('.');
	if(index != string::npos)
		Extension = Extension.substr(index+1);
	CleanString(Extension);
	CleanString(MimeType);
	MakeLower(Extension);
	MakeLower(MimeType);
	if(!Extension.length())
		return;
	if(!MimeType.length())
		return;
	NV_LOCKOBJECT(this);
	MimeTypes.SetValue(Extension, MimeType);
	SaveMimeTable();	// We save the mime table to the registry
	NV_UNLOCKOBJECT(this);
}

// Get a mime type
string NvsServer::GetMimeType(string Extension)
{
	string::size_type index = Extension.rfind('.');
	if(index != string::npos)
		Extension = Extension.substr(index+1);
	CleanString(Extension);
	MakeLower(Extension);
	string MimeType = "";
	if(Extension.length())
	{
		NV_LOCKOBJECT(this);
		MimeType = MimeTypes.GetValue(Extension);
		NV_UNLOCKOBJECT(this);
	}
	return MimeType;
}



// Remove a mime type
void NvsServer::RemoveMimeType(string Extension)
{
	string::size_type index = Extension.rfind('.');
	if(index != string::npos)
		Extension = Extension.substr(index+1);
	CleanString(Extension);
	MakeLower(Extension);
	if(Extension.length())
	{
		NV_LOCKOBJECT(this);
		if(MimeTypes.KeyExist(Extension))
		{
			MimeTypes.Remove(Extension);
			SaveMimeTable();	// We save the mime table to the registry
		}
		NV_UNLOCKOBJECT(this);
	}
}

// Get complete mime table
void NvsServer::GetMimeTable(NvObjIndStringList *Result)
{
	if(Result == NULL)
		return;
	NV_LOCKOBJECT(this);
	Result->CopyFrom(&MimeTypes);
	NV_UNLOCKOBJECT(this);
}





// Load alias table from registry
void NvsServer::LoadAliasTable()
{
	NV_LOCKOBJECT(this);
	// load the alias table parameters
	// The alias table is stored in an indexed string list object named "aliases" in the
	// system registry key system.parameters
	NvContainer Container;
	if(SystemRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "aliases", false, true))
	{
		NvObj *Object = Container.GetObject("aliases");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_INDSTRINGLIST))
		{
			NvObjIndStringList *RegistryAliases = (NvObjIndStringList*)Object;
			Aliases.CopyFrom(RegistryAliases);
		}
	}
	else
	{
		// Saves the list in case of changes
		SaveAliasTable();
	}


	NV_UNLOCKOBJECT(this);

}


// Save alias table to registry
// NvsServer must be locked when entering this function
void NvsServer::SaveAliasTable()
{
	NvContainer Container;
	Container.CreateObject("aliases", NV_OBJTYPE_INDSTRINGLIST);
	NvObjIndStringList *RegistryAliases = (NvObjIndStringList*)(Container.GetObject("aliases"));
	if(RegistryAliases == NULL)
		return;
	RegistryAliases->CopyFrom(&Aliases);
	
	SystemRegistry.SetKey("system.parameters", &Container);
}



// Set an alias
void NvsServer::SetAlias(string Path, string Alias)
{
	CleanString(Path);
	MakeLower(Path);
	CleanString(Alias);
//	MakeLower(Alias);
	if(!Path.length())
		return;
	if(!Alias.length())
		return;
	if(Path[0] != '/')
		Path = "/" + Path;
	if(Alias[0] != '/')
		Alias = "/" + Alias;
	NV_LOCKOBJECT(this);
	Aliases.SetValue(Path, Alias);
	SaveAliasTable();	// We save the alias table to the registry
	NV_UNLOCKOBJECT(this);
}

// Get an alias
string NvsServer::GetAlias(string Path)
{
	CleanString(Path);
	MakeLower(Path);
	string Alias = "";
	if(Path.length())
	{
		NV_LOCKOBJECT(this);
		Alias = Aliases.GetValue(Path);
		NV_UNLOCKOBJECT(this);
	}
	return Alias;
}



// Remove an alias
void NvsServer::RemoveAlias(string Path)
{
	CleanString(Path);
	MakeLower(Path);
	if(Path.length())
	{
		NV_LOCKOBJECT(this);
		if(Aliases.KeyExist(Path))
		{
			Aliases.Remove(Path);
			SaveAliasTable();	// We save the alias table to the registry
		}
		NV_UNLOCKOBJECT(this);
	}
}

// Get complete alias table
void NvsServer::GetAliasTable(NvObjIndStringList *Result)
{
	if(Result == NULL)
		return;
	NV_LOCKOBJECT(this);
	Result->CopyFrom(&Aliases);
	NV_UNLOCKOBJECT(this);
}



#ifdef WIN32
#ifdef _DEBUG
void NvsServer::DebugState()
{
	//_CrtMemCheckpoint( &s0 );
}
void NvsServer::DebugDump()
{
	//_CrtMemDumpAllObjectsSince(  &s0 );
}
#endif
#endif

// Get nirva dir
string NvsServer::GetNirvaDir()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sNirvaDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Get machine name
string NvsServer::GetMachineName()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sMachineName;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Get machine user
string NvsServer::GetMachineUser()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sMachineUser;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Get debug mode
bool NvsServer::GetDebug()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bDebug;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set debug mode
void NvsServer::SetDebug(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bDebug = Value;
	NV_SETCHECKDEADLOCKS(m_bDebug);	// Set also the deadlock debug in action or not
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get remote admin mode
bool NvsServer::GetRemoteAdmin()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bRemoteAdmin;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set remote admin mode
void NvsServer::SetRemoteAdmin(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bRemoteAdmin = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get debug mode
bool NvsServer::GetLongSessionId()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bLongSessionId;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set debug mode
void NvsServer::SetLongSessionId(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bLongSessionId = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get debug mode
bool NvsServer::GetOldObjectsUnicode()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bOldObjectsUnicode;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set debug mode
void NvsServer::SetOldObjectsUnicode(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bOldObjectsUnicode = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get internal XSL mode
bool NvsServer::GetInternalXsl()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bInternalXsl;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


void NvsServer::SetInternalXsl(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bInternalXsl = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}



// Get XSL cache mode
bool NvsServer::GetEnableXslCache()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bEnableXslCache;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


void NvsServer::SetEnableXslCache(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bEnableXslCache = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
	NvXml::EnableCache(Value);
}



// Set SSL mode
void NvsServer::SetSSL(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bUseSSL = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get SSL mode
bool NvsServer::GetSSL()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bUseSSL;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set SSLV3Only mode
void NvsServer::SetSSLV3Only(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bSSLV3Only = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get SSLV3Only mode
bool NvsServer::GetSSLV3Only()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bSSLV3Only;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set HttpLocal mode
void NvsServer::SetHttpLocal(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bHttpLocal = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get HttpLocal mode
bool NvsServer::GetHttpLocal()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bHttpLocal;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



// Set HttpsUseClientCertificate mode
void NvsServer::SetHttpsUseClientCertificate(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bHttpsUseClientCertificate = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get HttpsUseClientCertificate mode
bool NvsServer::GetHttpsUseClientCertificate()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bHttpsUseClientCertificate;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



// Get CA file
string NvsServer::GetHttpsCAFile()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sHttpsCAFile;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set CA file
bool NvsServer::SetHttpsCAFile(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sHttpsCAFile = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


// Set auto detect encoding
void NvsServer::SetAutoDetectEncoding(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bAutoDetectEncoding = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get auto detect encoding mode
bool NvsServer::GetAutoDetectEncoding()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bAutoDetectEncoding;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



// Get default page
string NvsServer::GetNirvaDefaultPage()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sNirvaDefaultPage;
	if(!RetVal.length())
		RetVal = "default.htm";
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set default page
bool NvsServer::SetNirvaDefaultPage(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sNirvaDefaultPage = Value;
	if(!m_sNirvaDefaultPage.length())
		m_sNirvaDefaultPage = "default.htm";
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

// Get local time mode
bool NvsServer::GetLocalTime()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bLocalTime;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set local time mode
void NvsServer::SetLocalTime(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bLocalTime = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get xsl command line
string NvsServer::GetXslPath()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sXslProcessorPath;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set xsl command line
bool NvsServer::SetXslPath(const char *Value)
{
	string NewPath = Value;
	string CommandLineForSearch = NewPath;
	MakeUpper(CommandLineForSearch);
	string::size_type index = CommandLineForSearch.find("%INPUTFILE");
	if(index == string::npos)
		return false;
	index = CommandLineForSearch.find("%OUTPUTFILE");
	if(index == string::npos)
		return false;
	index = CommandLineForSearch.find("%XSLFILE");
	if(index == string::npos)
		return false;
	NV_LOCKOBJECT(this);
	m_sXslProcessorPath = NewPath;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

// Get xsl max depth
int NvsServer::GetXslMaxDepth()
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	RetVal = m_iXslMaxDepth;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set xsl max depth
void NvsServer::SetXslMaxDepth(int Value)
{
	NV_LOCKOBJECT(this);
	m_iXslMaxDepth = Value;
	if(m_iXslMaxDepth < 5000)
		m_iXslMaxDepth = 5000;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}

// Get xsl parser max depth
int NvsServer::GetXslParserMaxDepth()
{
	int RetVal = 0;
	NV_LOCKOBJECT(this);
	RetVal = m_iXslParserMaxDepth;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set xsl parser max depth
void NvsServer::SetXslParserMaxDepth(int Value)
{
	NV_LOCKOBJECT(this);
	m_iXslParserMaxDepth = Value;
	if(m_iXslParserMaxDepth < 1024)
		m_iXslParserMaxDepth = 1024;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


// Get xsl command line
string NvsServer::GetStatus()
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	RetVal = m_sServerStatus;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// Set xsl command line
void NvsServer::SetStatus(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sServerStatus = Value;
	NV_UNLOCKOBJECT(this);
}

int NvsServer::GetHttpPort()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iHttpServerPort;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetHttpPort(int Port)
{
	NV_LOCKOBJECT(this);
	m_iHttpServerPort = Port;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

int NvsServer::GetHttpsPort()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iHttpsServerPort;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetHttpsPort(int Port)
{
	NV_LOCKOBJECT(this);
	m_iHttpsServerPort = Port;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

int NvsServer::GetHttpSessionTimeOut()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iHttpSessionTimeOut;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetHttpSessionTimeOut(int SessionTimeOut)
{
	NV_LOCKOBJECT(this);
	m_iHttpSessionTimeOut = SessionTimeOut;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


// Get internal JVM mode
bool NvsServer::GetInternalJvm()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bInternalJvm;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set internal JVM mode
void NvsServer::SetInternalJvm(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bInternalJvm = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}

int NvsServer::GetJvmMinMem()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_JVMMinMemSize;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetJvmMinMem(int MemSize)
{
	NV_LOCKOBJECT(this);
	m_JVMMinMemSize = MemSize;
	if(m_JVMMinMemSize < 4)
		m_JVMMinMemSize = 4;
	if(m_JVMMaxMemSize < m_JVMMinMemSize)
		m_JVMMaxMemSize = m_JVMMinMemSize;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

int NvsServer::GetJvmMaxMem()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_JVMMaxMemSize;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetJvmMaxMem(int MemSize)
{
	NV_LOCKOBJECT(this);
	m_JVMMaxMemSize = MemSize;
	if(m_JVMMaxMemSize < m_JVMMinMemSize)
		m_JVMMaxMemSize = m_JVMMinMemSize;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

string NvsServer::GetJvmHomeDir()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sJVMHomeDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetJvmHomeDir(string HomeDir)
{
	NV_LOCKOBJECT(this);
	m_sJVMHomeDir = HomeDir;
	CleanString(m_sJVMHomeDir);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

string NvsServer::GetJvmParam()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sJVMParam;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetJvmParam(string Value)
{
	NV_LOCKOBJECT(this);
	m_sJVMParam = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

string NvsServer::GetJvmPath()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sJVMPathName;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetJvmPath(string Path)
{
	NV_LOCKOBJECT(this);
	m_sJVMPathName = Path;
	CleanString(m_sJVMPathName);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



// Get use transaction server
bool NvsServer::GetUseTransactionServer()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bUseTransactionServer;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Set use transaction server
void NvsServer::SetUseTransactionServer(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bUseTransactionServer = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


int NvsServer::GetMaxTransactions()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iMaxTransactions;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetMaxTransactions(int MaxTransactions)
{
	NV_LOCKOBJECT(this);
	m_iMaxTransactions = MaxTransactions;
	if(m_iMaxTransactions < 1)
		m_iMaxTransactions = 1;
	if(m_iMaxTransactions > 500)
		m_iMaxTransactions = 500;
	TransactionServer.SetMaxTransactions(m_iMaxTransactions);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


int NvsServer::GetTransactionCheckEvery()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iTransactionCheckEvery;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetTransactionCheckEvery(int CheckEvery)
{
	NV_LOCKOBJECT(this);
	m_iTransactionCheckEvery = CheckEvery;
	if(m_iTransactionCheckEvery < 1)
		m_iTransactionCheckEvery = 1;
	TransactionServer.SetCheckEvery(m_iTransactionCheckEvery);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



int NvsServer::GetMaxSchedulerTasks()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iMaxSchedulerTasks;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetMaxSchedulerTasks(int MaxSchedulerTasks)
{
	NV_LOCKOBJECT(this);
	m_iMaxSchedulerTasks = MaxSchedulerTasks;
	if(m_iMaxSchedulerTasks < 1)
		m_iMaxSchedulerTasks = 1;
	if(m_iMaxSchedulerTasks > 500)
		m_iMaxSchedulerTasks = 500;
	SchedulerServer.SetMaxTasks(m_iMaxSchedulerTasks);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


int NvsServer::GetSchedulerCheckEvery()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iSchedulerCheckEvery;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetSchedulerCheckEvery(int CheckEvery)
{
	NV_LOCKOBJECT(this);
	m_iSchedulerCheckEvery = CheckEvery;
	if(m_iSchedulerCheckEvery < 1)
		m_iSchedulerCheckEvery = 1;
	SchedulerServer.SetCheckEvery(m_iSchedulerCheckEvery);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


int NvsServer::GetDefaultAppTimeOut()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iDefaultAppTimeOut;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultAppTimeOut(int TimeOut)
{
	NV_LOCKOBJECT(this);
	m_iDefaultAppTimeOut = TimeOut;
	if(m_iDefaultAppTimeOut < 30)
		m_iDefaultAppTimeOut = 30;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


int NvsServer::GetDefaultAppAuthMode()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iDefaultAppAuthMode;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultAppAuthMode(int AuhtMode)
{
	NV_LOCKOBJECT(this);
	m_iDefaultAppAuthMode = AuhtMode;
	if(m_iDefaultAppAuthMode < 0)
		m_iDefaultAppAuthMode = NV_AUTHENTICATION_NIRVA;
	if(m_iDefaultAppAuthMode > NV_AUTHENTICATION_MAXVALUE)
		m_iDefaultAppAuthMode = NV_AUTHENTICATION_NIRVA;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



bool NvsServer::GetDefaultAppNirvaIsSsoUser()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bDefaultNirvaIsSsoUser;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultAppNirvaIsSsoUser(bool NirvaIsSsoUser)
{
	NV_LOCKOBJECT(this);
	m_bDefaultNirvaIsSsoUser = NirvaIsSsoUser;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

string NvsServer::GetDefaultAppSecuApp()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sDefaultAppSecuApp;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultAppSecuApp(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sDefaultAppSecuApp = Value;
	CleanString(m_sDefaultAppSecuApp);
	MakeUpper(m_sDefaultAppSecuApp);
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


bool NvsServer::GetDefaultAppUseScheduled()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bDefaultAppUseScheduled;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


void NvsServer::SetDefaultAppUseScheduled(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bDefaultAppUseScheduled = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


bool NvsServer::GetDefaultAppAllowWebXml()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bDefaultAppAllowWebXml;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


void NvsServer::SetDefaultAppAllowWebXml(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bDefaultAppAllowWebXml = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


bool NvsServer::GetDefaultAppAllowNotOwnedSession()
{
	bool RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_bDefaultAppAllowNotOwnedSession;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


void NvsServer::SetDefaultAppAllowNotOwnedSession(bool Value)
{
	NV_LOCKOBJECT(this);
	m_bDefaultAppAllowNotOwnedSession = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();
}


string NvsServer::GetAdminPassword()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sAdminPassword;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetAdminPassword(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sAdminPassword = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

int NvsServer::GetHttpStopWaitTime()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iHttpStopWaitTime;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetHttpStopWaitTime(int StopWaitTime)
{
	NV_LOCKOBJECT(this);
	m_iHttpStopWaitTime = StopWaitTime;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}

int NvsServer::GetHttpMaxMemBody()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iHttpMaxMemBody;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetHttpMaxMemBody(int MaxMemBody)
{
	NV_LOCKOBJECT(this);
	m_iHttpMaxMemBody = MaxMemBody;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



string NvsServer::GetRegistryHomeDir(bool NewVal)
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	if(NewVal)
		RetVal = m_sNewRegistryHomeDir;
	else
		RetVal = m_sRegistryHomeDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetRegistryHomeDir(const char *Value)
{
	NV_LOCKOBJECT(this);
	string m_sOldRegistryHomeDir = m_sNewRegistryHomeDir;
	m_sNewRegistryHomeDir = Value;
	CleanString(m_sNewRegistryHomeDir);
	if(!m_sNewRegistryHomeDir.length())
		m_sNewRegistryHomeDir = m_sNirvaDir;
	AddDirSuffix(m_sNewRegistryHomeDir);

	// If the value has changed we must save it)
	if(m_sNewRegistryHomeDir != m_sOldRegistryHomeDir)
		SaveBasicParameters();		
	
	NV_UNLOCKOBJECT(this);

	return true;
}


string NvsServer::GetLogHomeDir(bool NewVal)
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	if(NewVal)
		RetVal = m_sNewLogHomeDir;
	else
		RetVal = m_sLogHomeDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetLogHomeDir(const char *Value)
{
	NV_LOCKOBJECT(this);
	string m_sOldLogHomeDir = m_sNewLogHomeDir;
	m_sNewLogHomeDir = Value;
	CleanString(m_sNewLogHomeDir);
	if(!m_sNewLogHomeDir.length())
		m_sNewLogHomeDir = m_sNirvaDir;
	AddDirSuffix(m_sNewLogHomeDir);


	// If the value has changed we must save it)
	if(m_sNewLogHomeDir != m_sOldLogHomeDir)
		SaveBasicParameters();		
	
	NV_UNLOCKOBJECT(this);

	return true;
}


string NvsServer::GetAppWorkHomeDir(bool NewVal)
{
	string RetVal = "";
	NV_LOCKOBJECT(this);
	if(NewVal)
		RetVal = m_sNewAppWorkHomeDir;
	else
		RetVal = m_sAppWorkHomeDir;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvsServer::SetAppWorkHomeDir(const char *Value)
{
	NV_LOCKOBJECT(this);
	string m_sOldAppWorkHomeDir = m_sNewAppWorkHomeDir;
	m_sNewAppWorkHomeDir = Value;
	CleanString(m_sNewAppWorkHomeDir);
	if(!m_sNewAppWorkHomeDir.length())
		m_sNewAppWorkHomeDir = m_sNirvaDir;
	AddDirSuffix(m_sNewAppWorkHomeDir);

	// If the value has changed we must save it)
	if(m_sNewAppWorkHomeDir != m_sOldAppWorkHomeDir)
		SaveBasicParameters();		
	
	NV_UNLOCKOBJECT(this);

	return true;
}



string NvsServer::GetDefaultMailServer()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sDefaultMailServer;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultMailServer(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sDefaultMailServer = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



string NvsServer::GetDefaultMailUser()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sDefaultMailUser;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultMailUser(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sDefaultMailUser = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}



string NvsServer::GetDefaultMailPassword()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sDefaultMailPassword;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetDefaultMailPassword(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sDefaultMailPassword = Value;
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}


string NvsServer::GetHostWebs()
{
	string RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_sHostWebs;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::SetHostWebs(const char *Value)
{
	NV_LOCKOBJECT(this);
	m_sHostWebs = Value;
	if(!m_sHostWebs.length())
	{
		m_sHostWebs = m_sMachineName;
		if(!m_sHostWebs.length())
			m_sHostWebs = "127.0.0.1";
		if(m_iHttpServerPort != 80)
		{
			char Buf[20];
			sprintf(Buf, ":%d", m_iHttpServerPort);
			m_sHostWebs += Buf;
		}
		MakeLower(m_sHostWebs);
	}
	NV_UNLOCKOBJECT(this);
	SaveParameters();

	return true;
}




bool NvsServer::IsUnicode()
{
	return b_Unicode;
}


bool NvsServer::GetInfoEx(NvObjIndStringList *SystemInfo)
{
	if(SystemInfo == NULL)
		return false;

	SystemInfo->RemoveAll();

	NV_LOCKOBJECT(this);

	// We must load now the description file if it exists
	LoadInfo();

	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Info.GetObject("INFO"));

	if(IndStringList == NULL)
	{
		NV_UNLOCKOBJECT(this);		// Nothing to return
		return true;
	}

	SystemInfo->CopyFrom(IndStringList);
	NV_UNLOCKOBJECT(this);

	return true;
}




int NvsServer::GetMaxSessions()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iMaxSessions;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}



int NvsServer::GetLicenseType()
{
	int RetVal;
	NV_LOCKOBJECT(this);
	RetVal = m_iLicenseType;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

// return true if the license doesn't authorise to run at the actual time (stays for compatibility reason, forced to false)
bool NvsServer::CheckOverTime()
{
	return false;
}



bool NvsServer::GetPermissions(NvObjTable *SystemPermissions)
{
	NV_LOCKOBJECT(this);

	// We must load now the description file if it exists
	// This is necessary at system level in order to reflect changes in the info file
	LoadInfo();

	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Info.GetObject("PERMISSIONS"));
	if(IndStringList != NULL)
	{
		map<string,string>::iterator pos;
		for(pos = IndStringList->m_mList.begin(); pos != IndStringList->m_mList.end(); ++pos)
		{
			string KeyFromList = pos->first.c_str();
			CleanString(KeyFromList);
			MakeUpper(KeyFromList);
			if(!KeyFromList.length())
				continue;
			string ValueFromList = pos->second.c_str();
			CleanString(ValueFromList);
			if(!ValueFromList.length())
				continue;
			string KeyToAdd = "SYSTEM:";
			KeyToAdd += KeyFromList;
			SystemPermissions->InsertRow();
			SystemPermissions->InsertCellLine(-1, 0, KeyToAdd);
			SystemPermissions->InsertCellLine(-1, 1, ValueFromList);
		}
	}

	NV_UNLOCKOBJECT(this);
	return true;
}



bool NvsServer::IsMqAvailable()
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = m_bMqAvailable;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvsServer::IsLcAvailable()
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = m_bLcAvailable;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

void NvsServer::SetStopRequest()
{
	NV_LOCKOBJECT(this);
	m_bStopRequest = true;
	NV_UNLOCKOBJECT(this);
}

bool NvsServer::GetStopRequest()
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = m_bStopRequest;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}






NvThread::NvThread()
{
	m_iTid = NvGetCurrentThreadId();
	m_tCreationTime = time(NULL);
	m_iType = THREAD_TYPE_UNKNOWN;
}


NvThread::~NvThread()
{
}


NvThreadList::NvThreadList()
{
}


NvThreadList::~NvThreadList()
{
	Clear();
}


void NvThreadList::Clear()
{
	NV_LOCKLOCK(&ProtectList);
	map<int,NvThread *>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvThread *Thread = (NvThread *)(pos->second);
		delete Thread;
	}
	m_mList.clear();
	NV_UNLOCKLOCK(&ProtectList);
}


void NvThreadList::Add(int Type)
{
	NvThread *Thread = new NvThread;
	NV_LOCKLOCK(&ProtectList);
	if(m_mList.find(Thread->m_iTid) != m_mList.end())
	{
		NV_UNLOCKLOCK(&ProtectList);
		delete Thread;
		return;
	}
	Thread->m_iType = Type;
	m_mList[Thread->m_iTid] = Thread;
	NV_UNLOCKLOCK(&ProtectList);
}


void NvThreadList::Remove()
{
	int CurrentThread = NvGetCurrentThreadId();
	NV_LOCKLOCK(&ProtectList);
	if(m_mList.find(CurrentThread) != m_mList.end())
	{
		NvThread *Thread = m_mList[CurrentThread];
		delete Thread;
		m_mList.erase(CurrentThread);
	}
	NV_UNLOCKLOCK(&ProtectList);
}


bool NvThreadList::List(NvObjTable *ThreadList)
{
	ThreadList->Clear();
	
	ThreadList->InsertColumn("THREAD");
	ThreadList->InsertColumn("TYPE");
	ThreadList->InsertColumn("CREATED");

	NV_LOCKLOCK(&ProtectList);
	int j;
	char Buffer[100];
	map<int,NvThread *>::iterator pos;
	for(pos = m_mList.begin(), j=0; pos != m_mList.end(); ++pos)
	{
		NvThread *Thread = (NvThread *)(pos->second);
		ThreadList->InsertRow();
		sprintf(Buffer,"%d",Thread->m_iTid);
		ThreadList->InsertCellLine(j, 0, Buffer);
		if(Thread->m_iType == THREAD_TYPE_HTTP_SERVER)
			ThreadList->InsertCellLine(j, 1, "HTTP SERVER");
		else if(Thread->m_iType == THREAD_TYPE_HTTP_CLIENT)
			ThreadList->InsertCellLine(j, 1, "HTTP CLIENT");
		else if(Thread->m_iType == THREAD_TYPE_HTTP_CLIENT_BUILDER)
			ThreadList->InsertCellLine(j, 1, "HTTP CLIENT BUILDER");
		else if(Thread->m_iType == THREAD_TYPE_LISTENER)
			ThreadList->InsertCellLine(j, 1, "LISTENER");
		else if(Thread->m_iType == THREAD_TYPE_MQ)
			ThreadList->InsertCellLine(j, 1, "MQ");
		else if(Thread->m_iType == THREAD_TYPE_MAIN)
			ThreadList->InsertCellLine(j, 1, "MAIN");
		else if(Thread->m_iType == THREAD_TYPE_SCHEDULER_SERVER)
			ThreadList->InsertCellLine(j, 1, "SCHEDULER SERVER");
		else if(Thread->m_iType == THREAD_TYPE_SCHEDULER_TASK)
			ThreadList->InsertCellLine(j, 1, "TASK");
		else if(Thread->m_iType == THREAD_TYPE_TRANSACTION_SERVER)
			ThreadList->InsertCellLine(j, 1, "TRANSACTION SERVER");
		else if(Thread->m_iType == THREAD_TYPE_TRANSACTION_CLIENT)
			ThreadList->InsertCellLine(j, 1, "TRANSACTION");
		else if(Thread->m_iType == THREAD_TYPE_JAVA)
			ThreadList->InsertCellLine(j, 1, "JAVA");
		else if(Thread->m_iType == THREAD_TYPE_PERL)
			ThreadList->InsertCellLine(j, 1, "PERL");
		else if(Thread->m_iType == THREAD_TYPE_DOTNET)
			ThreadList->InsertCellLine(j, 1, "DOTNET");
		else if(Thread->m_iType == THREAD_TYPE_SESSION)
			ThreadList->InsertCellLine(j, 1, "SESSION");
		else
			ThreadList->InsertCellLine(j, 1, "UNKNOWN");

		struct tm *tCurrentTime = localtime(&(Thread->m_tCreationTime));
		sprintf(Buffer, "%4d/%02d/%02d %02d:%02d:%02d",
						tCurrentTime->tm_year + 1900,
						tCurrentTime->tm_mon + 1,
						tCurrentTime->tm_mday,
						tCurrentTime->tm_hour,
						tCurrentTime->tm_min,
						tCurrentTime->tm_sec); 

		ThreadList->InsertCellLine(j, 2, Buffer);
		j++;
	}
	NV_UNLOCKLOCK(&ProtectList);

	ThreadList->Sort("TYPE");

	return true;
}

NvAttachRequest::NvAttachRequest()
{
	m_sPipeFile = "";
	hServerWritePipehandle = 0;
	m_iAttachedConsolePid = 0;
	m_itype = ATTACH_REQUEST_TYPE_ANY;
	m_sApplication = "";
	m_bUpdateToRemove = false;
	m_iLevel = 1;
}


NvAttachRequest::~NvAttachRequest()
{
	if(hServerWritePipehandle > 0)
	{
#ifdef WIN32
		CloseHandle(hServerWritePipehandle);
#else
		close(hServerWritePipehandle);
#endif
	}
#ifndef WIN32
	if(m_sPipeFile.length())
		unlink(m_sPipeFile.c_str());
#endif
	hServerWritePipehandle = 0;
	m_sPipeFile = "";
	m_iAttachedConsolePid = 0;
	m_iLevel = 1;
}
NvAttachRequestList::NvAttachRequestList()
{
}


NvAttachRequestList::~NvAttachRequestList()
{
	Clear();
}

void NvAttachRequestList::Clear()
{
	NV_LOCKLOCK(&ProtectList);
	int i;
	NvAttachRequest *AttachRequest;
	for(i=0; i<(int)m_mList.size(); i++)
	{
		AttachRequest = m_mList[i];
		if(AttachRequest != NULL)
			delete AttachRequest;
	}
	m_mList.clear();
	NV_UNLOCKLOCK(&ProtectList);
}

// Create a new request for attaching a console
// Directly called from http.cpp when the command is SYSTEM:SESSION:ATTACH_CONSOLE
// and TYPE parameter is not empty
bool NvAttachRequestList::Add(NvSession *Session, NvCommand *CommandToExecute, int AttachedConsolePid, int Type, string Application, int Level)
{
	if(CommandToExecute->m_sClientAddress != "127.0.0.1")
	{
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Console can be attached localy only");
		return false;
	}
	if(CommandToExecute->m_iCommandSource != NV_SOURCE_CLIENT)
	{
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Console can be attached only from nvc client");
		return false;
	}
	if(AttachedConsolePid == 0)
	{
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "No process ID given");
		return false;
	}

#ifdef WIN32
	HANDLE hClientProcess = OpenProcess(PROCESS_DUP_HANDLE, false, AttachedConsolePid);
	if(hClientProcess == NULL)
	{
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Cannot open client process");
		return false;
	}

	HANDLE hClientPipehandle = NULL;
	SECURITY_ATTRIBUTES saAttr;
	
	// Set the bInheritHandle flag so pipe handles are not inherited. 
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
	saAttr.bInheritHandle = FALSE; 
	saAttr.lpSecurityDescriptor = NULL; 

	HANDLE hServerWritePipehandle;
	if(!CreatePipe(&hClientPipehandle, &hServerWritePipehandle, &saAttr, 0))
	{
		CloseHandle(hClientProcess);
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Cannot create pipe");
		return false;
	}

	// Ensure the read handle to the client is not inherited.
	if(!SetHandleInformation(hClientPipehandle, HANDLE_FLAG_INHERIT, 0))
	{
		CloseHandle(hServerWritePipehandle);
		CloseHandle(hClientPipehandle);
		CloseHandle(hClientProcess);
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Cannot set handle information");
		return false;
	}

	// Duplicate handle for the client process so we can close our side
	HANDLE hClientTempPipehandle = NULL;
	if(!DuplicateHandle(GetCurrentProcess(),
					hClientPipehandle,
					hClientProcess,
					&hClientTempPipehandle,
					0,
					FALSE,
					DUPLICATE_SAME_ACCESS))
	{
		CloseHandle(hServerWritePipehandle);
		CloseHandle(hClientPipehandle);
		CloseHandle(hClientProcess);
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Cannot duplicate handle");
		return false;
	}

	// Close process handle
	CloseHandle(hClientProcess);
	// Close our client handle
	CloseHandle(hClientPipehandle);

	// the client duplicated handle doesn(t have to be closed because it has been created for the
	// client process so it's under its responsability
	
	// We send back the client pipe handle to the client (in the output buffer)
	char sClientPipehandle[50];
	sprintf(sClientPipehandle, "%llu", (NvInt64)hClientTempPipehandle);
	Session->GetStringValue(CommandToExecute, sClientPipehandle);

#else
	// Under unix we use fifo (naed pipes)
	// We create a fifo in the Nirva/Work directory having the session and cleint PID in its name
	string m_sPipeFile = GetMain()->GetNirvaDir();
	m_sPipeFile += "Work";
	AddDirSuffix(m_sPipeFile);
	m_sPipeFile += GetMain()->GetUniqString();
	m_sPipeFile += ".fifo";
	if(mkfifo(m_sPipeFile.c_str(), 0777) != 0)
	{
		m_sPipeFile = "";
		CommandToExecute->Error.SetError("SYSTEM", "SESSION", NV_SESSION_CANNOT_ATTACH_CONSOLE, "Cannot open fifo file");
		return false;
	}
	Session->GetStringValue(CommandToExecute, m_sPipeFile.c_str());
#endif

	NvAttachRequest *AttachRequest = new NvAttachRequest;
	AttachRequest->m_itype = Type;
	MakeUpper(Application);
	CleanString(Application);
	AttachRequest->m_sApplication = Application;
	AttachRequest->m_iAttachedConsolePid = AttachedConsolePid;
	AttachRequest->m_iLevel = Level;
#ifdef WIN32
	AttachRequest->hServerWritePipehandle = hServerWritePipehandle;
#else
	AttachRequest->m_sPipeFile = m_sPipeFile;
#endif
	NV_LOCKLOCK(&ProtectList);
	Update();
	m_mList.push_back(AttachRequest);
	NV_UNLOCKLOCK(&ProtectList);

	return true;
}

void NvAttachRequestList::Get(NvSession *Session, int Type, string Application)
{
	NV_LOCKLOCK(&ProtectList);
	if(!m_mList.size())
	{
		NV_UNLOCKLOCK(&ProtectList);
		return;
	}
	MakeUpper(Application);
	CleanString(Application);
	if(!Application.length())
		Application = NV_APPLICATION_DEFAULT_NAME;
	NvAttachRequest *AttachRequest = NULL;
	bool bFound = false;
	int i;
	// First search for specific request types
	for(i=0; i<(int)m_mList.size(); i++)
	{
		AttachRequest = m_mList[i];
		// If we have a pending request to a specific application we check if this is the correct app
		if(AttachRequest->m_sApplication.length() && (AttachRequest->m_sApplication != Application))
			continue;
		if(Type == AttachRequest->m_itype)
		{
			// Verify that the console pid still exists
#ifdef WIN32
			HANDLE hClientProcess = OpenProcess(PROCESS_DUP_HANDLE, false, AttachRequest->m_iAttachedConsolePid);
			if(hClientProcess != NULL)
			{
				// Yes it exists
				CloseHandle(hClientProcess);
				bFound = true;
				break;
			}
#else
			int killresult = kill(AttachRequest->m_iAttachedConsolePid, 0);
			if((killresult == 0) || (errno != ESRCH))
			{
				// Yes it exists
				bFound = true;
				break;
			}
#endif
			// Doesn't exist so we prepare the request for removing
			AttachRequest->m_bUpdateToRemove = true;
		}
	}

	// Then search for any request types
	if(!bFound)
	{
		for(i=0; i<(int)m_mList.size(); i++)
		{
			AttachRequest = m_mList[i];
			if(AttachRequest->m_bUpdateToRemove)
				continue;
			// If we have a pending request to a specific application we check if this is the correct app
			if(AttachRequest->m_sApplication.length() && (AttachRequest->m_sApplication != Application))
				continue;
			if(AttachRequest->m_itype == ATTACH_REQUEST_TYPE_ANY)
			{
#ifdef WIN32
				HANDLE hClientProcess = OpenProcess(PROCESS_DUP_HANDLE, false, AttachRequest->m_iAttachedConsolePid);
				if(hClientProcess != NULL)
				{
					// Yes it exists
					CloseHandle(hClientProcess);
					bFound = true;
					break;
				}
#else
				int killresult = kill(AttachRequest->m_iAttachedConsolePid, 0);
				if((killresult == 0) || (errno != ESRCH))
				{
					// Yes it exists
					bFound = true;
					break;
				}
#endif
				// Doesn't exist so we prepare the request for removing
				AttachRequest->m_bUpdateToRemove = true;
			}
		}
	}

	if(!bFound)
	{
		Update();
		NV_UNLOCKLOCK(&ProtectList);
		return;
	}

	// We have found a request. We must transfert the ownership of pipe to session
	NV_LOCKLOCK(&(Session->ProtectPipe));
	Session->m_sPipeFile = AttachRequest->m_sPipeFile;
	AttachRequest->m_sPipeFile = "";
	Session->hServerWritePipehandle = AttachRequest->hServerWritePipehandle;
	AttachRequest->hServerWritePipehandle = 0;
	Session->m_iAttachedConsolePid = AttachRequest->m_iAttachedConsolePid;
	AttachRequest->m_iAttachedConsolePid = 0;
	Session->m_bHasPipe = true;
	Session->bFirstAttachedCommand = true;
	Session->m_iAttachedLevel = AttachRequest->m_iLevel;
	NV_UNLOCKLOCK(&(Session->ProtectPipe));

	AttachRequest->m_bUpdateToRemove = true;
	Update();
	NV_UNLOCKLOCK(&ProtectList);
}

// Not locked because called internally
void NvAttachRequestList::Update()
{
	if(!m_mList.size())
		return;

	vector<NvAttachRequest *> NewList;
	NvAttachRequest *AttachRequest = NULL;
	int i;
	// First search for specific request types
	for(i=0; i<(int)m_mList.size(); i++)
	{
		AttachRequest = m_mList[i];
		if(AttachRequest->m_bUpdateToRemove)
		{
			delete AttachRequest;
			continue;
		}
		// Verify that the console pid still exists
#ifdef WIN32
		HANDLE hClientProcess = OpenProcess(PROCESS_DUP_HANDLE, false, AttachRequest->m_iAttachedConsolePid);
		if(hClientProcess != NULL)
		{
			// Yes it exists
			CloseHandle(hClientProcess);
			NewList.push_back(AttachRequest);
			continue;
		}
#else
		int killresult = kill(AttachRequest->m_iAttachedConsolePid, 0);
		if((killresult == 0) || (errno != ESRCH))
		{
			// Yes it exists
			NewList.push_back(AttachRequest);
			continue;
		}
#endif
		// No it doesn't exist any more so we cleanup
		delete AttachRequest;
	}

	// Update the mist from new list
	m_mList.clear();
	for(i=0;i<(int)NewList.size();i++)
		m_mList.push_back(NewList[i]);
}



int GetNumberOfProcessors()
{
	// Get the number od CPUs
#ifdef WIN32
	int NumProcessors = GetNumberOfProcessorsWin32();
#endif
#ifdef LINUX
	int NumProcessors  = sysconf(_SC_NPROCESSORS_ONLN);
#endif
#ifdef HPUX
	int NumProcessors = pthread_num_processors_np();
#endif
#ifdef HPUXI
	int NumProcessors = pthread_num_processors_np();
#endif
#ifdef SOLARIS
	int NumProcessors = sysconf(_SC_NPROCESSORS_ONLN);
#endif
#ifdef AIX
	int NumProcessors = sysconf(_SC_NPROCESSORS_ONLN);
#endif
	return NumProcessors;
}


// UnHandled ExceptionFilter under windows
#ifdef WIN32
LONG __stdcall ExceptCallBack(EXCEPTION_POINTERS *pExPtrs)
{
	printf ("Unhandeled exception occured\n");
	return (EXCEPTION_EXECUTE_HANDLER);
}
#endif

