// ***************************************************************************************
// File: obj.h
// Description: Declarations of Nirva objects that can go into a container.
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifndef _OBJ_H
#define _OBJ_H



#include "task.h"
#include "intdb.h"
#include "garbage.h"
#include <vector>
#include <string>
#include <map>


// Classes declared in this file
class NvObj;
class NvObjBoolean;
class NvObjInteger;
class NvObjString;
class NvObjStringList;
class NvObjIndStringList;
class NvObjTable;
class NvObjFile;
class NvObjBinary;

// Object types
#define NV_OBJTYPE_UNKNOWN				0		// Unknown type
#define NV_OBJTYPE_BOOLEAN				1		// boolean (true or false)
#define NV_OBJTYPE_INTEGER				2		// integer (32 bits)
#define NV_OBJTYPE_STRING				3		// string of any size
#define NV_OBJTYPE_STRINGLIST			4		// list of string indexed by a numeric index
#define NV_OBJTYPE_INDSTRINGLIST		5		// list of string indexed by a string
#define NV_OBJTYPE_TABLE				6		// table (column * row)
#define NV_OBJTYPE_FILE					7		// file
#define NV_OBJTYPE_BINARY				8		// binary information

// Object encoding
#define NV_OBJ_ENCODING_ISO859_1		0
#define NV_OBJ_ENCODING_UTF8			1

// String standard separtors
#define NV_SEPARATOR_LINE				"NVLINE"
#define NV_SEPARATOR_TAB				"NVTAB"
#define NV_SEPARATOR_SPACE				"NVSPACE"


#define NV_OBJTYPE_MAXTYPEVALUE			NV_OBJTYPE_BINARY

#define NV_OBJ_DEFAULT_NAME				"NV_OBJ_DEFAULT_NAME"







/////////////////////////////////////////////////////////////////////////////
// NvObj class
class NvObj
{
public:

	std::string m_sName;
	int m_iType;
	
	NvObj();
	NvObj(int Type);
	NvObj(std::string Name);
	NvObj(const char *Name);
	NvObj(int Type, std::string Name);
	NvObj(int Type, const char *Name);
	virtual ~NvObj();

	int GetType();
	bool SetType(int Type);
	std::string GetName();
	bool SetName(std::string Name);
	bool SetName(const char *Name);
	static bool FormatObjectName(std::string& Name);
#ifdef NV_SERVER
	void EncodeName(bool bSrcUnicode, bool bDestUnicode);
#endif
};




/////////////////////////////////////////////////////////////////////////////
// NvObjBoolean class
class NvObjBoolean : public NvObj
{
public:
	bool m_bValue;
	
public:
	NvObjBoolean();
	NvObjBoolean(std::string Name);
	NvObjBoolean(const char *Name);
	virtual ~NvObjBoolean();
	void CopyFrom(NvObjBoolean *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);

	bool GetValue();
	void SetValue(bool Value);
};


/////////////////////////////////////////////////////////////////////////////
// NvObjInteger class
class NvObjInteger : public NvObj
{
public:
	int m_iValue;
	
public:
	NvObjInteger();
	NvObjInteger(std::string Name);
	NvObjInteger(const char *Name);
	virtual ~NvObjInteger();
	void CopyFrom(NvObjInteger *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);

	int GetValue();
	void SetValue(int Value);
};



/////////////////////////////////////////////////////////////////////////////
// NvObjString class
class NvObjString : public NvObj
{
public:
	std::string m_sValue;
	
public:
	NvObjString();
	NvObjString(std::string Name);
	NvObjString(const char *Name);
	virtual ~NvObjString();
	void CopyFrom(NvObjString *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);

	std::string GetValue();
	void GetValue(char *Buffer, int BufferSize);
	void SetValue(std::string Value);
	void SetValue(const char* Value);
};



/////////////////////////////////////////////////////////////////////////////
// NvObjStringList class
class NvObjStringList : public NvObj
{
public:
	std::vector<std::string> m_vList;
	
public:
	NvObjStringList();
	NvObjStringList(std::string Name);
	NvObjStringList(const char *Name);
	virtual ~NvObjStringList();
	void CopyFrom(NvObjStringList *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);
	bool CheckIndex(int index);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);

	int GetSize();
	void RemoveAll();
	void Remove(int index);
	void RemoveLast();
	void Insert(int index, std::string Value);
	void Insert(int index, const char *Value);
	void Add(std::string Value, std::string Separator = "");
	void Add(const char *Value, const char *Separator = NULL);
	void Add(NvObjStringList *Source);
	void SetValue(int index, std::string Value);
	void SetValue(int index, const char *Value);
	void Swap(int index1, int index2);
	void Sort(bool Numeric);
	std::string GetValue(int index);
	void GetValue(int index, char *Buffer, int BufferSize);
	std::string GetValues(int From = 0, int To = -1, std::string Separator = ";");
	std::string Search(std::string Value, bool Case = false, bool FirstOnly = false);
	std::string GetLastValue();
	void GetLastValue(char *Buffer, int BufferSize);
	std::string GetFirstValue();
	void GetFirstValue(char *Buffer, int BufferSize);
};



/////////////////////////////////////////////////////////////////////////////
// NvObjIndStringList class
class NvObjIndStringList : public NvObj
{
public:
	std::map<std::string, std::string> m_mList;
	bool m_bCaseSensitive;
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);
	
public:
	NvObjIndStringList();
	NvObjIndStringList(std::string Name);
	NvObjIndStringList(const char *Name);
	virtual ~NvObjIndStringList();
	void CopyFrom(NvObjIndStringList *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);

	int GetSize();
	void RemoveAll();
	void Remove(std::string key);
	void Remove(const char *key);
	void Add(NvObjIndStringList *Source);
	void SetValue(std::string key, std::string Value, std::string KeySep = ";", std::string ValSep = "=");
	void SetValue(const char *key, const char *Value);
	std::string GetValue(std::string key);
	void GetValue(const char *key, char *Buffer, int BufferSize);
	bool KeyExist(std::string key);
	bool KeyExist(const char *key);
	std::string GetKey(int index);
	bool GetKey(int index, char *Buffer, int BufferSize);
};




/////////////////////////////////////////////////////////////////////////////
// NvObjTable class
class NvObjTable : public NvObj
{
private:
	Intdbtable m_tTable;
	
public:
	NvObjTable();
	NvObjTable(std::string Name);
	NvObjTable(const char *Name);
	virtual ~NvObjTable();
	bool SetName(std::string Name);
	bool SetName(const char *Name);
	void CopyFrom(NvObjTable *Source);
	bool Populate(char *Data, unsigned long DataSize, bool bUseDecodeFromFile = false);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName, bool bUseDecodeFromFile = false);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length);

	
	bool CheckColumnIndex(int col_index);
	bool CheckRowIndex(int row_index);


	// Table functions
	void SetDescription(std::string Description);
	void SetDescription(const char *Description);
	std::string GetDescription();
	void GetDescription(char *Buffer, int BufferSize);
	void Clear();
	void ClearData();
	int Search(std::string Query, NvObjTable *Result, std::string SortColumn = "", int MaxDocs = -1, std::string Columns = "", bool RowIndexCol = false);
	int Search(const char* Query, NvObjTable *Result, const char* SortColumn = NULL, int MaxDocs = -1, const char* Columns = NULL, bool RowIndexCol = false);
	void Sort(std::string SortColumn);
	void Sort(const char* SortColumn);
	void CreateFrom(NvObjTable *Source);
	bool Save(std::string FileName);
	bool Save(const char* FileName);
	bool Load(std::string FileName);
	bool Load(const char* FileName);
	bool Import(std::string ImportFile, bool Append = false, const char* ColumnSeparator = ";", const char* LineSeparator = "|", int MaxLineLength = 1024, int Offset = 0, int SkipLines = 0, bool WithLf = false);
	bool Import(const char* ImportFile, bool Append = false, const char* ColumnSeparator = ";", const char* LineSeparator = "|", int MaxLineLength = 1024, int Offset = 0, int SkipLines = 0, bool WithLf = false);
	bool Export(std::string ExportFile, const char* ColumnSeparator = ";", const char* LineSeparator = "|", int MaxLineLength = 1024, bool WithLf = false);
	bool Export(const char* ExportFile, const char* ColumnSeparator = ";", const char* LineSeparator = "|", int MaxLineLength = 1024, bool WithLf = false);
	bool SelectFrom(NvObjTable* Source, int source_col_index, int dest_col_index, int selected_col_index, const char* SelectedYes = NULL, const char* SelectedNo = NULL, bool CaseSensitive = true);
	bool Join(NvObjTable* Source, const char* Columns, const char* Foreign, bool Replace = false);
	bool Add(NvObjTable *Source, int Mode = ADD_MODE_NORMAL);

	
	// Cell functions
	void ClearCell(int row_index, int col_index);
	void SortCell(int row_index, int col_index, bool Ascending = true);
	int GetNumCellLines(int row_index, int col_index);
	std::string GetCellLine(int row_index, int col_index, int line_index = 0, std::string Default = "");
	void GetCellLine(int row_index, int col_index, char *Buffer, int BufferSize, int line_index = 0, const char *Default = NULL);
	bool SetCellLine(int row_index, int col_index, std::string Line, int line_index = 0);
	bool SetCellLine(int row_index, int col_index, const char* Line, int line_index = 0);
	bool InsertCellLine(int row_index, int col_index, std::string Line, int line_index = -1);
	bool InsertCellLine(int row_index, int col_index, const char* Line, int line_index = -1);
	void RemoveCellLine(int row_index, int col_index, int line_index = -1);
	void SwapCellLines(int row_index, int col_index, int line_index1, int line_index2);

	
	// Row functions
	int GetNumRow();
	bool RemoveRows(std::string Query);
	bool RemoveRows(const char* Query);
	bool ClearRows(std::string Query);
	bool ClearRows(const char* Query);
	void InsertRows(int num_rows = 1, int row_index = -1);
	void ClearRow(int row_index);
	void RemoveRow(int row_index);
	void InsertRow(int row_index = -1);
	void SwapRows(int row_index1, int row_index2);
	int GetRows(int first, int last, NvObjTable *Result, const char* Columns = NULL);
	int GetRowIndex(std::string Primary);
	bool SetRowData(int row_index, const char* RecordData = NULL, const char* FieldSeparator = ";", const char* EntrySeparator = "|");
	bool AddRecords(const char* RecordData = NULL, const char* RecordSeparator = "\n", const char* FieldSeparator = ";", const char* EntrySeparator = "|");
	

	// Column functions
	int GetNumColumn();
	bool SetColumnName(int col_index, std::string column_name);
	std::string GetColumnName(int col_index);
	void GetColumnName(int col_index, char *Buffer, int BufferSize);
	void SetColumnDesc(int col_index, std::string column_desc);
	std::string GetColumnDesc(int col_index);
	void GetColumnDesc(int col_index, char *Buffer, int BufferSize);
	void SetColumnNumeric(int col_index, bool numeric);
	bool IsNumericColumn(int col_index);
	int GetColumnIndex(std::string ColumnName);
	int GetColumnIndex(const char* ColumnName);
	void ClearColumn(int col_index);
	void RemoveColumn(int col_index);
	void InsertColumn(std::string column_name, std::string column_desc = "", bool numeric = false, int col_index = -1);
	void InsertColumn(const char* column_name, const char* column_desc = NULL, bool numeric = false, int col_index = -1);
	bool ModifyColumn(std::string Query, int col_index, std::string Value = "");
	bool ModifyColumn(const char* Query, int col_index, const char* Value = NULL);
	void SwapColumns(int col_index1, int col_index2);
	bool FilterColumns(const char* Columns = NULL);
	bool SetPrimaryColumn(std::string ColumnName, bool bCaseSensitive, bool bRemoveDoubleEntries = false);
	bool GetPrimaryInfo(std::string& ColumnName, bool& bCaseSensitive);
};



/////////////////////////////////////////////////////////////////////////////
// NvObjFile class
class NvObjFile : public NvObj
{
public:
	std::string m_sFileName;
	std::string m_sOrigin;
	int m_iPersistTime;
	FILE *m_fdesc;
	NvGarbage *Garbage;		// Optional garbage collector
	
public:
	NvObjFile();
	NvObjFile(std::string Name);
	NvObjFile(const char *Name);
	virtual ~NvObjFile();
	bool CopyFrom(NvObjFile *Source, NvInt64 Offset = 0, NvInt64 BytesToCopy = -1);
	bool Populate(char *Data, unsigned long DataSize);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName);
	bool Export(FILE *stream);
	bool Import(FILE *stream, NvInt64 length, NvInt64 Offset = 0, NvInt64 BytesToCopy = -1);

	void SetAscii();
	void SetPersistTime(int NumSeconds = -1);
	void SetFileName(std::string FileName);
	void SetFileName(const char *FileName);
	void SetOrigin(std::string Origin);
	void SetOrigin(const char *Origin);
	bool SetAutoFileName(std::string Dir, std::string Extension = "", std::string Prefix = "", std::string Suffix = "");
	bool SetAutoFileName(const char *Dir, const char *Extension = NULL, const char *Prefix = NULL, const char *Suffix = NULL);
	std::string GetFilePathName();
	void GetFilePathName(char *Buffer, int BufferSize);
	std::string GetFileDir();
	void GetFileDir(char *Buffer, int BufferSize);
	std::string GetFileName(bool WithExt = true);
	void GetFileName(char *Buffer, int BufferSize, bool WithExt = true);
	std::string GetFileExtension();
	void GetFileExtension(char *Buffer, int BufferSize);
	std::string GetOrigin();
	void GetOrigin(char *Buffer, int BufferSize);
	NvInt64 GetSize();
	void Remove(bool Force = true);
	bool Create();
	bool Clear();
	bool Open(std::string Mode);
	bool Open(const char *Mode);
	void Close();
	bool AppendFrom(NvObjFile *Source, NvInt64 Offset = 0, NvInt64 BytesToCopy = -1);
	int Read(char *Data, int DataSize);
	int ReadLine(std::string& Line, int MaxLineLength = 2048);
	int ReadLine(char *Line, int LineLength);
	bool Seek(NvInt64 SeekValue, int Origin = SEEK_SET);
	NvInt64 Tell();
	int Write(std::string Data);
	int Write(const char *Data, int DataSize);
	bool Flush();
	bool Exist();
	FILE *GetFileDescriptor();
	bool Compress(NvObjFile *Destination);
	bool Decompress(NvObjFile *Destination);
};
	


/////////////////////////////////////////////////////////////////////////////
// NvObjBinary class
class NvObjBinary : public NvObj
{
public:
	char *m_cdata;
	int m_isize;
	int m_iposition;
	
public:
	NvObjBinary();
	NvObjBinary(std::string Name);
	NvObjBinary(const char *Name);
	virtual ~NvObjBinary();
	bool CopyFrom(NvObjBinary *Source, int Offset = 0, int BytesToCopy = -1);
	bool Populate(char *Data, unsigned long DataSize);
	bool GetData(char **Data, unsigned long *DataSize);
#ifdef NV_SERVER
	bool GetXmlDataToStream(FILE *stream, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	bool GetXmlDataToString(std::string& Output, int NumTabs = 0, bool WithData = true, std::string NsPrefix = "", int XmlCoding = XML_MODEL_NORMAL, void *XmlPtr = NULL, bool bUnicode = false, bool bWebs = false);
	int GetXmlDataSize();
	bool SetSoapSample(std::string& Output, int NumTabs = 0, std::string NsPrefix = "", void *XmlPtr = NULL, bool bUnicode = false);
	void Encode(bool bSrcUnicode, bool bDestUnicode);
#endif
	bool SaveToFile(const char *FileName);
	bool LoadFromFile(const char *FileName);
	bool Export(FILE *stream);
	bool Import(FILE *stream, unsigned long length, int Offset = 0, int BytesToCopy = -1);

	char* GetBuffer();
	int GetSize();
	void Clear();
	bool Load(const char *FileName, int Offset = 0, int NumBytes = -1);
	bool Save(const char *FileName, int Offset = 0, int NumBytes = -1);
	bool Append(const char *Data, int DataSize);
	bool Append(std::string Data);
	bool AppendFrom(NvObjBinary *Source, int Offset = 0, int BytesToCopy = -1);
	int Read(char *Data, int DataSize);
	int ReadLine(char *Data, int DataSize);
	int ReadLine(std::string& Line, int MaxLineLength = 2048);
	bool Seek(int SeekValue, int Origin = SEEK_SET);
	int Tell();
	bool Write(std::string Data);
	bool Write(const char *Data, int DataSize);
};



#endif //_OBJ_H
