// ***************************************************************************************
// File: package.cpp
// Description: delivery packages
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifdef WIN32
#pragma warning( disable : 4996 )
#endif

#ifdef NV_SERVER
#include "nvs.h"
#else
#include "task.h"
#include "util.h"
#include "compress.h"
#include "package.h"
#endif

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER


using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvPackage

NvPackage::NvPackage()
{
	Info = "";
}

NvPackage::~NvPackage()
{
}


// Checks that a description file contains valid file pointers
bool NvPackage::Check(const char *DescriptionFile, const char *BaseDir)
{
	Info = "";

	if(DescriptionFile == NULL)
		return false;

	// The real check is made in the Create function in order to avoid
	// code duplication
	return Create(DescriptionFile, BaseDir);
}

// Creates a package file from a description file
// If the PackageFile parameter is NULL, the function only checks the
// description file
bool NvPackage::Create(const char *DescriptionFile, const char *BaseDir, const char *PackageFile)
{
	Info = "";

	if(DescriptionFile == NULL)
	{
		Info = "No description file";
		return false;
	}

	string BaseDirHere = "";
	if(BaseDir != NULL)
		BaseDirHere = BaseDir;
	CleanString(BaseDirHere);
	AddDirSuffix(BaseDirHere);

	bool CreateMode = true;
	if(PackageFile == NULL)
		CreateMode = false;

	FILE* PkgFile = NULL;
	FILE* DescFile = Nvfopen(DescriptionFile, "r");

	if(DescFile == NULL)
	{
		Info = "Cannot open description file ";
		Info += DescriptionFile;
		return false;
	}
	
	string TemporaryPackageFile = "";
	string Line;

	if(CreateMode)
	{
		TemporaryPackageFile = PackageFile;
		TemporaryPackageFile += ".nv_temp";
		PkgFile = Nvfopen(TemporaryPackageFile.c_str(), "w");
		if(PkgFile == NULL)
		{
			Info = "Cannot open temporary package file ";
			Info += TemporaryPackageFile;
			fclose(DescFile);
			return false;
		}
		Line = PACKAGE_FILE_IDENTIFIER;
		Line += "\n";
		Line += PACKAGE_FILE_VERSION;
		Line += "\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			fclose(DescFile);
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
			return false;
		}
	}

	string::size_type index;
	bool InHeader = false;
	bool InFileSection = false;
	bool HeaderWritten = false;
	bool PerlEncode = false;
	string DirName = "";
	string SectionName = "";
	vector<string> filelines;
	vector<string> perlfilelines;
	vector<string> dirlines;
	vector<string> dirsublines;
	vector<string> removedirlines;
	vector<string> removedirsublines;
	vector<string> execlines;
	vector<string> textfilelines;
	vector<string> textdirlines;
	vector<string> excludefilelines;
	vector<string> excludedirlines;

	char Buffer[2048];

	// Main loop
	// For each directory section, the function separates file lines, dir lines and dir with subdir lines
	// It then processes first the file lines, then the dir lines and then the dir with subdir lines. This is to
	// avoid to put some files in bad directories.
	while(freadline(Buffer, sizeof(Buffer), DescFile))
	{
		Line = Buffer;
		CleanString(Line);
		if(!Line.length())
			continue;
		if(Line.find("\\\\") == 0)	// comment
			continue;
		if(Line.find("//") == 0)	// comment
			continue;
		if(Line.find(";") == 0)		// comment
			continue;

		if((Line.find('[') == 0) && ((index = Line.find(']')) != string::npos))
		{
			// Process remove dir lines
			if(!ProcessRemoveDirLines(&removedirlines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process remove dir with subdir lines
			if(!ProcessRemoveDirSubLines(&removedirsublines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process file lines
			if(!ProcessFileLines(&filelines, PkgFile, PerlEncode))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process perl file lines
			if(!ProcessPerlFileLines(&perlfilelines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process dir lines
			if(!ProcessDirLines(&dirlines, &excludefilelines, PkgFile, SectionName, PerlEncode))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process dir with subdir lines
			if(!ProcessDirSubLines(&dirsublines, &excludefilelines, &excludedirlines, PkgFile, SectionName, PerlEncode))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process exec lines
			if(!ProcessExecLines(&execlines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process text file lines
			if(!ProcessTextFileLines(&textfilelines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			// Process text dir lines
			if(!ProcessTextDirLines(&textdirlines, PkgFile))
			{
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}

			filelines.clear();
			perlfilelines.clear();
			dirlines.clear();
			dirsublines.clear();
			removedirlines.clear();
			removedirsublines.clear();
			execlines.clear();
			textfilelines.clear();
			textdirlines.clear();
			excludefilelines.clear();
			excludedirlines.clear();

			InHeader = false;		// We are not any more in header at this step
			InFileSection = false;	// We are not any more in directory section at this step
			PerlEncode = false;		// No perl encoding by default in the section

			// This a new section
			// We get its name
			SectionName = "";
			if(index > 1)
				SectionName = Line.substr(1, index-1);
			CleanString(SectionName);
			string SectionNameUpper = SectionName;
			MakeUpper(SectionNameUpper);
			if(SectionNameUpper == "HEADER")
			{
				// This is the header
				if(HeaderWritten)	// If header ever written, we do nothing with this line
					continue;

				InHeader = true;
				if(CreateMode)
				{
					// We write header section line
					Line += "\n";
					if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
					{
						Info = "Cannot write package file";
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}
				}
				continue;
			}

			// Here we are in a new section but it's not a header
			// So its a directory

			while(SectionName.length() && (SectionName[0] == '\\'))
				SectionName.erase(0, 1);
			while(SectionName.length() && (SectionName[0] == '/'))
				SectionName.erase(0, 1);

			SetDirName(SectionName);
			RemoveDirSuffix(SectionName);

			// We are in a file section now
			InFileSection = true;

			DirName = BaseDir;
			DirName += SectionName;
			RemoveDirSuffix(DirName);
			SetDirName(DirName);

			// The next part is commented because we must be able to get files
			// also if the directory doesn't exist
/*			if(!DirExist(DirName.c_str()))
			{
				Info = "Bad directory: ";
				Info += DirName;
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
*/
			// We forbid to use .. in a directory name for security reason
			if(DirName.find("..") != string::npos)
			{
				Info = "Bad directory: ";
				Info += DirName;
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}

			if(!CreateMode)
				continue;

			Line = "[Dir]\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}

			Line = "Name = ";
			Line += SectionName;
			Line += "\n\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				fclose(DescFile);
				if(PkgFile != NULL)
				{
					fclose(PkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
				}
				return false;
			}
			continue;
		}

		if(InHeader)
		{
			if(CreateMode)
			{
				// We write header section line
				Line += "\n";
				if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
				{
					Info = "Cannot write package file";
					fclose(DescFile);
					if(PkgFile != NULL)
					{
						fclose(PkgFile);
						RemoveFile(TemporaryPackageFile.c_str());
					}
					return false;
				}
			}
		}

		if(InFileSection)
		{
			// We are in a file section
			index = Line.find(':');
			bool bSourceIsDir = false;
			bool bWithSubDir = false;
			bool bRemoveDir = false;
			bool bExec = false;
			bool bTextFile = false;
			bool bTextDir = false;
			bool bExcludeFile = false;
			bool bExcludeDir = false;
			string Action = "";
			if(index != string::npos)
			{
				Action = Line.substr(0, index);
				CleanString(Action);
				MakeLower(Action);
				bool RealAction = false;
				if(Action == "copyfile")
					RealAction = true;
				if(Action == "copyperlfile")
					RealAction = true;
				if(Action == "perlencode")
				{
					if(CreateMode)
						PerlEncode = true;
					continue;
				}
				if(Action == "copydir")
				{
					RealAction = true;
					bSourceIsDir = true;
				}
				if(Action == "copydirsub")
				{
					RealAction = true;
					bSourceIsDir = true;
					bWithSubDir = true;
				}
				if(Action == "removedir")
				{
					RealAction = true;
					bSourceIsDir = true;
					bRemoveDir = true;
				}
				if(Action == "removedirsub")
				{
					RealAction = true;
					bSourceIsDir = true;
					bWithSubDir = true;
					bRemoveDir = true;
				}
				if(Action == "excludefile")
				{
					RealAction = true;
					bSourceIsDir = true;
					bExcludeFile = true;
				}
				if(Action == "excludedir")
				{
					RealAction = true;
					bSourceIsDir = true;
					bExcludeDir = true;
				}
				if(Action == "exec")
				{
					RealAction = true;
					bExec = true;
				}
				if(Action == "textfile")
				{
					RealAction = true;
					bTextFile = true;
				}
				if(Action == "textdir")
				{
					RealAction = true;
					bSourceIsDir = true;
					bTextDir = true;
				}

				if(RealAction)
				{
					Line = Line.substr(index+1);
					CleanString(Line);
				}
			}

			if(bSourceIsDir)
			{
				if(bRemoveDir)
				{
					if(!CreateMode)
						continue;
					string DirSource = SectionName;
					RemoveDirSuffix(DirSource);
					if(bWithSubDir)
						removedirsublines.push_back(DirSource);
					else
						removedirlines.push_back(DirSource);
				}
				else if(bTextDir)
				{
					if(!CreateMode)
						continue;
					string DirSource = SectionName;
					RemoveDirSuffix(DirSource);
					textdirlines.push_back(DirSource);
				}
				else if(bExcludeFile)
				{
					if(!CreateMode)
						continue;
					string ExcludeFile = Line;
					if(!ExcludeFile.length())
						continue;
					excludefilelines.push_back(ExcludeFile);
				}
				else if(bExcludeDir)
				{
					if(!CreateMode)
						continue;
					string ExcludeDir = Line;
					if(!ExcludeDir.length())
						continue;
					excludedirlines.push_back(ExcludeDir);
				}
				else
				{
					string DirSource = Line;
					if(!DirSource.length())
						DirSource = DirName;
					SetDirName(DirSource);
					RemoveDirSuffix(DirSource);
					if(DirSource.find("..") != string::npos)
					{
						Info = "Bad directory: ";
						Info += DirSource;
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}
					if(!DirExist(DirSource.c_str()))
					{
						Info = "Bad directory: ";
						Info += DirSource;
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}

					if(!CreateMode)
						continue;
					
					if(bWithSubDir)
						dirsublines.push_back(DirSource);
					else
						dirlines.push_back(DirSource);
				}
			}
			else
			{
				if(bExec)
				{
					string FileName = Line;
					CleanString(FileName);

					if(!CreateMode)
						continue;

					if(!FileName.length())
						continue;
			
					Line = FileName;
					execlines.push_back(Line);
				}
				else if(bTextFile)
				{
					string FileName = Line;
					CleanString(FileName);

					if(!CreateMode)
						continue;

					if(!FileName.length())
						continue;
			
					Line = FileName;
					textfilelines.push_back(Line);
				}
				else
				{
					string FileName = Line;
					string FileSource = DirName;
					AddDirSuffix(FileSource);
					FileSource += FileName;
					index = Line.find('=');
					if(index != string::npos)
					{
						FileName = Line.substr(0, index);
						FileSource = Line.substr(index+1);
					}
					CleanString(FileName);
					CleanString(FileSource);
					SetDirName(FileSource);
					if(!FileSource.length())
					{
						Info = "No source file";
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}
					if(!FileName.length())
					{
						Info = "No target file";
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}
					if(!FileExist(FileSource.c_str()))
					{
						Info = "Bad file: ";
						Info += FileSource;
						fclose(DescFile);
						if(PkgFile != NULL)
						{
							fclose(PkgFile);
							RemoveFile(TemporaryPackageFile.c_str());
						}
						return false;
					}
					if(!CreateMode)
						continue;
			
					Line = FileName;
					Line += "=";
					Line += FileSource;
					if(Action == "copyperlfile")
						perlfilelines.push_back(Line);
					else
						filelines.push_back(Line);
				}
			}
		}
	}

	fclose(DescFile);

	// Process last remove dir lines
	if(!ProcessRemoveDirLines(&removedirlines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last remove dir with subdir lines
	if(!ProcessRemoveDirSubLines(&removedirsublines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last file lines
	if(!ProcessFileLines(&filelines, PkgFile, PerlEncode))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last perl file lines
	if(!ProcessPerlFileLines(&perlfilelines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last dir lines
	if(!ProcessDirLines(&dirlines, &excludefilelines, PkgFile, SectionName, PerlEncode))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last dir with subdir lines
	if(!ProcessDirSubLines(&dirsublines, &excludefilelines, &excludedirlines, PkgFile, SectionName, PerlEncode))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last exec lines
	if(!ProcessExecLines(&execlines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last text file lines
	if(!ProcessTextFileLines(&textfilelines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}
	// Process last text file lines
	if(!ProcessTextDirLines(&textdirlines, PkgFile))
	{
		if(PkgFile != NULL)
		{
			fclose(PkgFile);
			RemoveFile(TemporaryPackageFile.c_str());
		}
	}

	if(PkgFile != NULL)
		fclose(PkgFile);

	if(CreateMode)
	{
		// The package file is huffman compressed
		CHufComp HufComp;
		if(HufComp.make_lzh_compre(TemporaryPackageFile.c_str(), PackageFile) != 0)
		{
			Info = "Cannot compress package file";
			RemoveFile(TemporaryPackageFile.c_str());
			return false;
		}
		RemoveFile(TemporaryPackageFile.c_str());
	}

	return true;
}

bool NvPackage::ProcessFileLines(vector<string> *Lines, FILE* PkgFile, bool bPerlEncode)
{
	int i;
	string::size_type index;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		index = Line.find('=');
		if(index == string::npos)
			continue;	// should not occur
		string FileName = Line.substr(0, index);
		string FileSource = Line.substr(index+1);
		if(!AddFile(FileName, FileSource, PkgFile, false, false, false, false, "", NULL, NULL, bPerlEncode))
			return false;
	}
	return true;
}

bool NvPackage::ProcessPerlFileLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	string::size_type index;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		index = Line.find('=');
		if(index == string::npos)
			continue;	// should not occur
		string FileName = Line.substr(0, index);
		string FileSource = Line.substr(index+1);
		if(!AddFile(FileName, FileSource, PkgFile, false, false, false, false, "", NULL, NULL, true))
			return false;
	}
	return true;
}

bool NvPackage::ProcessDirLines(vector<string> *Lines, vector<string> *ExcludeFiles, FILE* PkgFile, string RelativeDir, bool bPerlEncode)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		if(!AddFile("", Line, PkgFile, true, false, false, false, RelativeDir, ExcludeFiles, NULL, bPerlEncode))
			return false;
	}
	return true;
}


bool NvPackage::ProcessDirSubLines(vector<string> *Lines, vector<string> *ExcludeFiles, vector<string> *ExcludeDirs, FILE* PkgFile, string RelativeDir, bool bPerlEncode)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		if(!AddFile("", Line, PkgFile, true, true, false, false, RelativeDir, ExcludeFiles, ExcludeDirs, bPerlEncode))
			return false;
	}
	return true;
}
	
bool NvPackage::ProcessRemoveDirLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		if(!AddFile("", Line, PkgFile, true, false, true, false, ""))
			return false;
	}
	return true;
}

bool NvPackage::ProcessRemoveDirSubLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		if(!AddFile("", Line, PkgFile, true, true, true, false, ""))
			return false;
	}
	return true;
}
	

bool NvPackage::ProcessExecLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string FileName = (*Lines)[i];

		string Line = "[Exec]\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}

		Line = "Name = ";
		Line += FileName;
		Line += "\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}

		Line = "\n\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}
	}
	return true;
}
	

bool NvPackage::ProcessTextFileLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string FileName = (*Lines)[i];

		string Line = "[TextFile]\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}

		Line = "Name = ";
		Line += FileName;
		Line += "\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}

		Line = "\n\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			return false;
		}
	}
	return true;
}

	
bool NvPackage::ProcessTextDirLines(vector<string> *Lines, FILE* PkgFile)
{
	int i;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		if(!AddFile("", Line, PkgFile, true, false, false, true, ""))
			return false;
	}
	return true;
}

// Used internally for storing a single file in the package file
// The function is recursive because the source may be a directory (SourceisDir set to true)
bool NvPackage::AddFile(string FileName, string FileSource, FILE* PkgFile, bool bSourceIsDir, bool bWithSubDir, bool bRemoveDir, bool bTextDir, string RelativeDir, vector<string> *ExcludeFiles, vector<string> *ExcludeDirs, bool bPerlEncode)
{
	string Line;
	if(bSourceIsDir)
	{
		if(bRemoveDir)
		{
			Line = "[RmDir]\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}

			Line = "Name = ";
			Line += FileSource;
			Line += "\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}
			
			Line = "Withsubdir = ";
			if(bWithSubDir)
				Line += "YES";
			else
				Line += "NO";
			Line += "\n\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}

		}
		else if(bTextDir)
		{
			Line = "[TextDir]\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}

			Line = "Name = ";
			Line += FileSource;
			Line += "\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}
			
			Line += "\n\n";
			if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
			{
				Info = "Cannot write package file";
				return false;
			}

		}
		else
		{
			// We must now list the source files
			vector<string> files;
			FileList(FileSource.c_str(), NULL, &files);

			// Process each file of the directory
			int i;
			for(i=0; i<(int)(files.size()); i++)
			{
				// If there is an exclude list, verify if the file is not in it
				bool bGetIt = true;
				if(ExcludeFiles != NULL)
				{
					int j;
					for(j=0; j<(int)(ExcludeFiles->size()); j++)
					{
						string ExcludeFile = (*ExcludeFiles)[j];
						if(!ExcludeFile.length())
							continue;
						if(ExcludeFile == "*")
						{
							// exclude this file
							bGetIt = false;
							break;
						}
						if(ExcludeFile[0] == '*')
						{
							// exclude all files terminating with given value
							ExcludeFile = ExcludeFile.substr(1);
							string::size_type index = files[i].rfind(ExcludeFile);
							if(index == string::npos)
								continue;
							if((index + ExcludeFile.length()) == files[i].length())
							{
								// exclude this file
								bGetIt = false;
								break;
							}
						}
						else if(ExcludeFile[ExcludeFile.length()-1] == '*')
						{
							// exclude all files starting with given value
							ExcludeFile = ExcludeFile.substr(0, ExcludeFile.length()-1);
							string::size_type index = files[i].find(ExcludeFile);
							if(index == 0)
							{
								// exclude this directory
								bGetIt = false;
								break;
							}
						}
						else
						{
							if(files[i] == ExcludeFile)
							{
								// exclude this file
								bGetIt = false;
								break;
							}
						}
					}
				}
				if(!bGetIt)
					continue;
				string RealFile = FileSource;
				AddDirSuffix(RealFile);
				RealFile += files[i];
				if(!AddFile(files[i], RealFile, PkgFile, false, false, false, false, RelativeDir, NULL, NULL, bPerlEncode))
					return false;
			}

			// Now eventually process the subdirectories
			if(bWithSubDir)
			{
				vector<string> subdirs;
				DirList(FileSource.c_str(), NULL, &subdirs);

				
				// Process each subdirectory
				for(i=0; i<(int)(subdirs.size()); i++)
				{
					// If there is an exclude list, verify if the file is not in it
					bool bGetIt = true;
					if(ExcludeDirs != NULL)
					{
						int j;
						for(j=0; j<(int)(ExcludeDirs->size()); j++)
						{
							string ExcludeDir = (*ExcludeDirs)[j];
							if(!ExcludeDir.length())
								continue;
							if(ExcludeDir == "*")
							{
								// exclude this directory
								bGetIt = false;
								break;
							}
							if(ExcludeDir[0] == '*')
							{
								// exclude all dirs terminating with given value
								ExcludeDir = ExcludeDir.substr(1);
								string::size_type index = subdirs[i].rfind(ExcludeDir);
								if(index == string::npos)
									continue;
								if((index + ExcludeDir.length()) == subdirs[i].length())
								{
									// exclude this directory
									bGetIt = false;
									break;
								}
							}
							else if(ExcludeDir[ExcludeDir.length()-1] == '*')
							{
								// exclude all dirs starting with given value
								ExcludeDir = ExcludeDir.substr(0, ExcludeDir.length()-1);
								string::size_type index = subdirs[i].find(ExcludeDir);
								if(index == 0)
								{
									// exclude this directory
									bGetIt = false;
									break;
								}
							}
							else
							{
								if(subdirs[i] == ExcludeDir)
								{
									// exclude this directory
									bGetIt = false;
									break;
								}
							}
						}
					}
					if(!bGetIt)
						continue;

					string SubDir = FileSource;
					AddDirSuffix(SubDir);
					SubDir += subdirs[i];
					AddDirSuffix(SubDir);

					string NewRelativeDir = RelativeDir;
					AddDirSuffix(NewRelativeDir);
					NewRelativeDir += subdirs[i];

					Line = "[Dir]\n";
					if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
					{
						Info = "Cannot write package file";
						return false;
					}

					Line = "Name = ";
					Line += NewRelativeDir;
					Line += "\n\n";
					if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
					{
						Info = "Cannot write package file";
						return false;
					}

					if(!AddFile("", SubDir, PkgFile, true, true, false, false, NewRelativeDir, ExcludeFiles, ExcludeDirs, bPerlEncode))
						return false;
				}
			}
		}
		return true;
	}
#ifdef WIN32
	string FileNameLower = FileName;
	MakeLower(FileNameLower);
	if(FileNameLower == "thumbs.db")
		return true;
#endif

	FILE* ObjSourceFile = Nvfopen(FileSource.c_str(), "r");
	if(ObjSourceFile == NULL)
	{
		Info = "Cannot read file: ";
		Info += FileSource;
		return false;
	}
	fseek(ObjSourceFile, 0, SEEK_END);
	int iSize = ftell(ObjSourceFile);
	fseek(ObjSourceFile, 0, SEEK_SET);

	Line = "[File]\n";
	if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
	{
		Info = "Cannot write package file";
		fclose(ObjSourceFile);
		return false;
	}

	// Check if we have to encrypt a perl file
	bool bDoPerlEncode = false;
	if(bPerlEncode)
	{
		string::size_type index = FileSource.rfind('.');
		string Extension = "";
		if(index != string::npos)
			Extension = FileSource.substr(index);
		MakeLower(Extension);
		if(Extension == ".pl")
		{
			bDoPerlEncode = true;
			// we must verify if the file is not ever encoded
			char Header[50];
			Header[0] = '\0';
			freadline(Header, 49, ObjSourceFile);
			fseek(ObjSourceFile, 0, SEEK_SET);
			string HeaderHere = Header;
			if(HeaderHere == "NIRVA PERL ENCODED FILE")
				bDoPerlEncode = false;	// this is ever a crypted file
		}
	}

	if(bDoPerlEncode)
	{
		if(!AppendPerlFile(ObjSourceFile, PkgFile, FileName))
		{
			Info = "Cannot add perl file: ";
			Info += FileSource;
			fclose(ObjSourceFile);
			return false;
		}
	}
	else
	{
		Line = "Name = ";
		Line += FileName;
		Line += "\n";
		Line += "Size = ";
		char Buf[50];
		sprintf(Buf, "%lu", iSize);
		Line += Buf;
		Line += "\n";
		if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
		{
			Info = "Cannot write package file";
			fclose(ObjSourceFile);
			return false;
		}

		if(!AppendFileEx(ObjSourceFile, PkgFile))
		{
			Info = "Cannot read file: ";
			Info += FileSource;
			fclose(ObjSourceFile);
			return false;
		}
	}
	fclose(ObjSourceFile);

	Line = "\n\n";
	if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
	{
		Info = "Cannot write package file";
		return false;
	}

	return true;
}



// Adds a perl crypted file to the package
bool NvPackage::AppendPerlFile(FILE* ObjSourceFile, FILE* PkgFile, string& FileName)
{
#ifdef NV_SERVER
	string Header = "NIRVA PERL ENCODED FILE";
	string Ctrl = GetMain()->GetUniqString().substr(0, 8);
	string EncryptionKey = "p";
	EncryptionKey += "nri";
	EncryptionKey += "Aav";
	EncryptionKey += "m";
	EncryptionKey += "lerp";
	EncryptionKey += Ctrl;
	EncryptionKey = MD5_string(EncryptionKey).substr(0,10);

	int BuffSize = 0;

	fseek(ObjSourceFile, 0, SEEK_END);
		BuffSize = ftell(ObjSourceFile); 

	fseek(ObjSourceFile, 0, SEEK_SET);

	char *ScriptData = (char *)malloc(BuffSize + 1);
	if(ScriptData == NULL)
		return false;

	ScriptData[BuffSize] = '\0';
	if(fread(ScriptData, 1, BuffSize, ObjSourceFile) != BuffSize)
	{	
		free(ScriptData);
		return false;
	}


	XOR(ScriptData, BuffSize, EncryptionKey.c_str());

	int iSize  = (int)(Header.length()) + 1 + (int)(Ctrl.length()) +1 + BuffSize;
	string Line = "Name = ";
	Line += FileName;
	Line += "\n";
	Line += "Size = ";
	char Buf[50];
	sprintf(Buf, "%lu", iSize);
	Line += Buf;
	Line += "\n";
	Line += Header;
	Line += "\n";
	Line += Ctrl;
	Line += "\n";
	// Write package header
	if(fwrite(Line.c_str(), 1, Line.length(), PkgFile) != Line.length())
	{
		free(ScriptData);
		return false;
	}
	// Write file data
	if(fwrite(ScriptData, 1, BuffSize, PkgFile) != BuffSize)
	{
		free(ScriptData);
		return false;
	}
	free(ScriptData);
#endif
	return true;
}



// Install a package file
bool NvPackage::Install(const char *PackageFile, const char *BaseDir, bool WithInfo)
{
	Info = "";

	if(PackageFile == NULL)
	{
		Info = "No package file";
		return false;
	}

	string BaseDirHere = "";
	if(BaseDir != NULL)
		BaseDirHere = BaseDir;
	CleanString(BaseDirHere);
	RemoveDirSuffix(BaseDirHere);
	if(!DirExist(BaseDirHere.c_str()))
	{
		// Creates the base directory if it doesn't exist
		if(!CreateDir(BaseDirHere.c_str()))
		{
			Info = "cannot create target directory: ";
			Info += BaseDirHere;
			return false;
		}
	}

	if(WithInfo)
		printf("\nInstalling package %s", PackageFile);

	AddDirSuffix(BaseDirHere);

	string TemporaryPackageFile = BaseDirHere;
	string PackageFileName = PackageFile;

	string::size_type index;
	index = PackageFileName.rfind('/');

	if(index == string::npos)
		index = PackageFileName.rfind('\\');

	if(index != string::npos)
		PackageFileName = PackageFileName.substr(index+1);

	TemporaryPackageFile += PackageFileName;
	TemporaryPackageFile += ".nv_temp";
	
	if(WithInfo)
		printf("\nUncompressing package file");
	// The package file is huffman compressed
	CHufComp HufComp;
	if(HufComp.make_lzh_decomp(PackageFile, TemporaryPackageFile.c_str()) != 0)
	{
		Info = "Cannot decompress package file";
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}

	FILE* TemporaryPkgFile = Nvfopen(TemporaryPackageFile.c_str(), "r");

	if(TemporaryPkgFile == NULL)
	{
		Info = "Cannot open temporary package file ";
		Info += TemporaryPackageFile;
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}

#ifdef AIX
	// Under AIX, we must try to run the slibclean command in order to free
	// unused shared libraries from memory
	int RetCode;
	ExecPrgm("slibclean", &RetCode, -1);
#endif

	char Buffer[2048];

	if(WithInfo)
		printf("\nVerifying package");
	// Read and check file identifier and version
	string IdentifierInFile = "";
	string VersionInFile = "";
	if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	IdentifierInFile = Buffer;
	CleanString(IdentifierInFile);
	if(IdentifierInFile != PACKAGE_FILE_IDENTIFIER)
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	VersionInFile = Buffer;
	CleanString(VersionInFile);

	string Line;
	string DirName = "";
	while(freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Line = Buffer;
		CleanString(Line);

		if((Line.find('[') == 0) && ((index = Line.find(']')) != string::npos))
		{

			// This a new section
			// We get its name
			string SectionName = "";
			if(index > 1)
				SectionName = Line.substr(1, index-1);
			CleanString(SectionName);
			string SectionNameUpper = SectionName;
			MakeUpper(SectionNameUpper);
			if(SectionNameUpper == "DIR")
			{
				// This is a directory section
				// We must create the directory
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string OriginDirName = Line.substr(index+7);
				CleanString(OriginDirName);
#ifdef WIN32
				while((index = OriginDirName.find('/')) != string::npos)
					OriginDirName.replace(index, 1, 1, '\\');
#else
				while((index = OriginDirName.find('\\')) != string::npos)
					OriginDirName.replace(index, 1, 1, '/');
#endif
				DirName = OriginDirName;
				string RealDirName = BaseDir;
				RemoveDirSuffix(RealDirName);
				if(WithInfo)
					printf("\nDirectory %s", DirName.c_str());

				if(!CreateDir(RealDirName.c_str()))
				{
					Info = "cannot create directory: ";
					Info += RealDirName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				AddDirSuffix(RealDirName);

#ifdef WIN32
				while((index = DirName.find('\\')) != string::npos)
#else
				while((index = DirName.find('/')) != string::npos)
#endif
				{
					string SubDir = DirName.substr(0, index);
					DirName = DirName.substr(index+1);
					string DirToCreate = RealDirName;
					DirToCreate += SubDir;
					if(!CreateDir(DirToCreate.c_str()))
					{
						Info = "cannot create directory: ";
						Info += DirToCreate;
						fclose(TemporaryPkgFile);
						RemoveFile(TemporaryPackageFile.c_str());
						return false;
					}
					RealDirName = DirToCreate;
					AddDirSuffix(RealDirName);
				}

				// Create last directory
				if(DirName.length())
				{
					string DirToCreate = RealDirName;
					DirToCreate += DirName;
					if(!CreateDir(DirToCreate.c_str()))
					{
						Info = "cannot create directory: ";
						Info += DirToCreate;
						fclose(TemporaryPkgFile);
						RemoveFile(TemporaryPackageFile.c_str());
						return false;
					}
				}

				DirName = BaseDir;
				AddDirSuffix(DirName);
				DirName += OriginDirName;
				AddDirSuffix(DirName);
			}
			if(SectionNameUpper == "RMDIR")
			{
				// This is a remove directory section
				// We must remove the directory
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string OriginDirName = Line.substr(index+7);
				CleanString(OriginDirName);
#ifdef WIN32
				while((index = OriginDirName.find('/')) != string::npos)
					OriginDirName.replace(index, 1, 1, '\\');
#else
				while((index = OriginDirName.find('\\')) != string::npos)
					OriginDirName.replace(index, 1, 1, '/');
#endif
				string RealDirName = BaseDir;
				AddDirSuffix(RealDirName);
				RealDirName += OriginDirName;
				RemoveDirSuffix(RealDirName);

				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Withsubdir = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}

				string sWithSubDir = Line.substr(index+13);
				MakeUpper(sWithSubDir);
				CleanString(sWithSubDir);
				if(sWithSubDir == "YES")
					sWithSubDir = "*";
				else
					sWithSubDir = "";

				if(!RemoveDir(RealDirName.c_str(), false, "*", sWithSubDir.c_str()))
				{
					Info = "cannot remove directory: ";
					Info += RealDirName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
			}
			if(SectionNameUpper == "FILE")
			{
				// This is a file section
				// We must create the file
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string FileName = Line.substr(index+7);
				CleanString(FileName);
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Size = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string FileSize = Line.substr(index+7);
				CleanString(FileSize);
				unsigned long iFileSize = strtoul(FileSize.c_str(), NULL, 10);

				int Position = ftell(TemporaryPkgFile);
				if(fseek(TemporaryPkgFile, iFileSize, SEEK_CUR) != 0)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
#ifdef WIN32
				string FileNameLower = FileName;
				MakeLower(FileNameLower);
				if(FileNameLower == "thumbs.db")
					continue;
#endif
				string CompleteFileName = DirName;
				CompleteFileName += FileName;
#ifdef WIN32
				// Change file attribute to normal in case the file has been turned to hidden or read only
				if(FileExist(CompleteFileName.c_str()))
					SetFileAttributes(CompleteFileName.c_str(), FILE_ATTRIBUTE_NORMAL);
#endif
				FILE* DestFile = Nvfopen(CompleteFileName.c_str(), "wb");
				if(DestFile == NULL)
				{
					Info = "Cannot create file ";
					Info += CompleteFileName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				if(!CopyFileEx(TemporaryPkgFile, DestFile, (NvInt64)Position, (NvInt64)iFileSize))
				{
					Info = "Cannot create file ";
					Info += CompleteFileName;
					fclose(TemporaryPkgFile);
					fclose(DestFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				fclose(DestFile);
			}
			if(SectionNameUpper == "EXEC")
			{
				// This is a exec section
				// We must do a exec for unix
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string FileName = Line.substr(index+7);
				CleanString(FileName);

				string CompleteFileName = DirName;
				CompleteFileName += FileName;
#ifndef WIN32
				// Do the job only for UNIX

				mode_t iParam = S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
				if(chmod(CompleteFileName.c_str(), iParam) != 0)
				{
					Info = "Cannot change file access rigths ";
					Info += CompleteFileName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
#endif
			}
			if(SectionNameUpper == "TEXTFILE")
			{
				// This is a text file section
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string FileName = Line.substr(index+7);
				CleanString(FileName);

				string CompleteFileName = DirName;
				CompleteFileName += FileName;

				if(!SetTextFile(CompleteFileName.c_str()))
				{
					Info = "Cannot set text file ";
					Info += CompleteFileName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
			}
			if(SectionNameUpper == "TEXTDIR")
			{
				// This is a text directory section
				// We must process all directory files
				if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				Line = Buffer;
				index = Line.find("Name = ");
				if(index == string::npos)
				{
					Info = "Bad package file ";
					Info += PackageFile;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					return false;
				}
				string OriginDirName = Line.substr(index+7);
				CleanString(OriginDirName);
#ifdef WIN32
				while((index = OriginDirName.find('/')) != string::npos)
					OriginDirName.replace(index, 1, 1, '\\');
#else
				while((index = OriginDirName.find('\\')) != string::npos)
					OriginDirName.replace(index, 1, 1, '/');
#endif
				string RealDirName = BaseDir;
				AddDirSuffix(RealDirName);
				RealDirName += OriginDirName;
				RemoveDirSuffix(RealDirName);

				vector<string> List;

				if(!FileList(RealDirName.c_str(), NULL, &List))
				{
					Info = "cannot list directory: ";
					Info += RealDirName;
					fclose(TemporaryPkgFile);
					RemoveFile(TemporaryPackageFile.c_str());
					List.clear();
					return false;
				}

				int i;
				for(i=0; i< (int)(List.size()); i++)
				{
					string CompleteFileName = RealDirName;
					AddDirSuffix(CompleteFileName);
					CompleteFileName += List[i];

					if(!SetTextFile(CompleteFileName.c_str()))
					{
						Info = "Cannot set text file ";
						Info += CompleteFileName;
						fclose(TemporaryPkgFile);
						RemoveFile(TemporaryPackageFile.c_str());
						List.clear();
						return false;
					}
				}
				List.clear();

			}
		}
	}

	fclose(TemporaryPkgFile);
	RemoveFile(TemporaryPackageFile.c_str());
	return true;
}







// Get header information from a package file
bool NvPackage::GetHeaderLines(const char *PackageFile, vector<string> *Lines)
{
	Info = "";
	Lines->clear();

	if(PackageFile == NULL)
	{
		Info = "No package file";
		return false;
	}


	string TemporaryPackageFile = PackageFile;
	TemporaryPackageFile += ".nv_temp";

	
	// The package file is huffman compressed
	CHufComp HufComp;
	int MaxBytes = 100000;
	if(HufComp.make_lzh_decomp(PackageFile, TemporaryPackageFile.c_str(), MaxBytes) != 0)
	{
		Info = "Cannot decompress package file";
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}

	FILE* TemporaryPkgFile = Nvfopen(TemporaryPackageFile.c_str(), "r");

	if(TemporaryPkgFile == NULL)
	{
		Info = "Cannot open temporary package file ";
		Info += TemporaryPackageFile;
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}

	char Buffer[2048];

	// Read and check file identifier and version
	string IdentifierInFile = "";
	string VersionInFile = "";
	if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	IdentifierInFile = Buffer;
	CleanString(IdentifierInFile);
	if(IdentifierInFile != PACKAGE_FILE_IDENTIFIER)
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	if(!freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Info = "Bad package file ";
		Info += PackageFile;
		fclose(TemporaryPkgFile);
		RemoveFile(TemporaryPackageFile.c_str());
		return false;
	}
	VersionInFile = Buffer;
	CleanString(VersionInFile);

	string Line;
	bool InHeader = false;
	string::size_type index;
	while(freadline(Buffer, sizeof(Buffer), TemporaryPkgFile))
	{
		Line = Buffer;
		CleanString(Line);

		if((Line.find('[') == 0) && ((index = Line.find(']')) != string::npos))
		{

			// This a new section
			// We get its name
			string SectionName = "";
			if(index > 1)
				SectionName = Line.substr(1, index-1);
			CleanString(SectionName);
			string SectionNameUpper = SectionName;
			MakeUpper(SectionNameUpper);
			if(SectionNameUpper != "HEADER")
				break;
			InHeader = true;
			continue;
		}
		if(InHeader)
		{
			// Get Header line here
			index = Line.find('=');
			if(index == string::npos)
				continue;	// Bad header entry
			Lines->push_back(Line);
		}
	}

#ifdef NV_SERVER
	// We must set the header lines to the proper encoding
	
	// We first search for the ENCODING entry if there is one
	int i;
	bool bUnicode = false;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		index = Line.find('=');
		if(index == string::npos)
			continue;	// Bad header entry
		string EntryName = Line.substr(0, index);
		string EntryValue = Line.substr(index+1);
		CleanString(EntryName);
		CleanString(EntryValue);
		MakeUpper(EntryName);
		MakeUpper(EntryValue);
		if(EntryName == "ENCODING")
		{
			// We do not need this entry
			Lines->erase(Lines->begin()+i);
			if(EntryValue == "UTF-8")
				bUnicode = true;
			break;
		}
	}
	vector<string> CopyLines;
	for(i=0; i<(int)(Lines->size()); i++)
	{
		string Line = (*Lines)[i];
		DecodeUrlString(Line, bUnicode, false);
		CopyLines.push_back(Line);
	}
	Lines->clear();
	*Lines = CopyLines;
#endif

	fclose(TemporaryPkgFile);
	RemoveFile(TemporaryPackageFile.c_str());
	return true;
}



bool NvPackage::SetTextFile(const char *FileName)
{
	string TemporaryFile = FileName;
	TemporaryFile += ".nv_temp";

	
	FILE* TmpFile = Nvfopen(TemporaryFile.c_str(), "w");
	if(TmpFile == NULL)
	{
		Info = "Cannot open temporary file ";
		Info += TemporaryFile;
		RemoveFile(TemporaryFile.c_str());
		return false;
	}

	FILE* File = Nvfopen(FileName, "r");
	if(File == NULL)
	{
		Info = "Cannot open file ";
		Info += FileName;
		fclose(TmpFile);
		RemoveFile(TemporaryFile.c_str());
		return false;
	}

	int BufferSize = 65536;
	char *Buffer = (char*)malloc(65536);	// Max line length
	if(Buffer == NULL)
	{
		Info = "Not enough memory";
		fclose(TmpFile);
		fclose(File);
		RemoveFile(TemporaryFile.c_str());
		return false;
	}


	string Line;
	while(freadline(Buffer, BufferSize, File))
	{
		Line = Buffer;
#ifdef WIN32
		Line += "\r";
#endif
		Line += "\n";

		if(fwrite(Line.c_str(), 1, Line.length(), TmpFile) != Line.length())
		{
			free(Buffer);
			Info = "Cannot write file";
			Info += TemporaryFile;
			fclose(TmpFile);
			fclose(File);
			RemoveFile(TemporaryFile.c_str());
			return false;
		}
	}

	fclose(TmpFile);
	fclose(File);
	free(Buffer);

	if(!CopyFileEx(TemporaryFile.c_str(), FileName))
	{
		Info = "Cannot wricopyte file";
		Info += TemporaryFile;
		Info += " -> ";
		Info += FileName;
		RemoveFile(TemporaryFile.c_str());
		return false;
	}

	RemoveFile(TemporaryFile.c_str());
	return true;
}

