// ***************************************************************************************
// File: protocol.cpp
// Description: layer used by http for networking
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: This file defines the protocol classes.
//           There is one abstract class named NvProtocol that defines the
//           interface of the protocol to external system.
//           Then, the classes derived from this base class implement a
//           specific protocol. They only have to overload the Read and Write
//           functions of the base class.
//           The NvData class is used only by the NvProtocol class. It
//           encapsulate binary data information associated with a message. The
//           NvProtocol interface defines a collection of NvData objects.
// ***************************************************************************************


#ifdef NV_SERVER
#include "nvs.h"
#else
#include "../nvc/nvcc.h"
#endif
#include "protocol.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;

// Uncomment next line if log socket needed (used to catch read exception)
//#define LOG_SOCKET 1

// NvData CLASS
// Manages binary data
NvData::NvData()
{
	DataSize = 0;
	Data = NULL;
	sInfo = "";
	DataFile = "";
	bDeleteDataFile = false;
}


// constructor by copy
NvData::NvData(const NvData& DataSrc)
{
	Data = NULL;
	sInfo = DataSrc.sInfo;
	DataSize = DataSrc.DataSize;
	if(DataSize)
		Data = (char *)malloc(DataSize);
	memcpy(Data, DataSrc.Data, DataSize);
	mHeaders = DataSrc.mHeaders;
	DataFile = DataSrc.DataFile;
	bDeleteDataFile = DataSrc.bDeleteDataFile;
}


NvData::~NvData()
{
	Clear();
}


// resets the class to its original state
void NvData::Clear()
{
	sInfo = "";
	mHeaders.clear();	// we also remove all headers
	FreeData();
}


// just free the associated data
void NvData::FreeData()
{
	if(Data != NULL)	// we delete eventual associated data
		free(Data);
	DataSize = 0;
	Data = NULL;
	if(bDeleteDataFile)
		if(DataFile.length())
			RemoveFile(DataFile.c_str());
	DataFile = "";
}


// copy operator
const NvData& NvData::operator = (const NvData& DataSrc)
{
	Clear();
	sInfo = DataSrc.sInfo;
	DataSize = DataSrc.DataSize;
	if(DataSize)
		Data = (char *)malloc(DataSize);
	memcpy(Data, DataSrc.Data, DataSize);
	mHeaders = DataSrc.mHeaders;
	DataFile = DataSrc.DataFile;
	bDeleteDataFile = DataSrc.bDeleteDataFile;

	return *this;	
}



// NvProtocol CLASS
NvProtocol::NvProtocol()
{
	sSocket = NULL;
	sPath="";
	sService="";
	sCommand="";
}

NvProtocol::NvProtocol(NvSocket *Socket)
{
	sPath="";
	sService="";
	sCommand="";
	sSocket = Socket;
}

NvProtocol::~NvProtocol()
{
	Clear();
}


// resets the class to its original state
void NvProtocol::Clear()
{
	sPath="";
	sService="";
	sCommand="";
	mParameters.clear();		// we also remove all parameters
	int i;
	for(i=0; i<(int)vDatas.size(); i++)
		if( vDatas[i] != NULL)
			delete vDatas[i];
	vDatas.clear();				// we also remove all datas
}



// Add a header value to the given data index
void NvProtocol::AddHeader(string HeaderName, string HeaderValue, int index)
{
	// first check if the index is correct
	if(index+1 > (int)vDatas.size())
		return;

	NvData *Data = vDatas[index];

	// erase the previous value if it exists
	if(Data->mHeaders.find(HeaderName) != Data->mHeaders.end())
		Data->mHeaders.erase(HeaderName);

	// set the new value
	Data->mHeaders[HeaderName] = HeaderValue;
}


// test if a parameter exist
bool NvProtocol::ParameterExist(string ParameterName)
{
	// make the test
	if(mParameters.find(ParameterName) != mParameters.end())
		return true;
	
	return false;
}


// get a parameter value
string NvProtocol::GetParameter(string ParameterName)
{
	// by default the return value is empty string
	string result = "";
	

	// we get the value if it exists
	if(ParameterExist(ParameterName))
		result = mParameters[ParameterName];
	
	return result;
}


// test if a header exist in the given index data
bool NvProtocol::HeaderExist(string HeaderName, int index)
{
	if(index+1 > (int)vDatas.size())
		return false;
	
	NvData *Data = vDatas[index];

	if(Data->mHeaders.find(HeaderName) != Data->mHeaders.end())
		return true;
	return false;
}


// get a header value from the given index data
string NvProtocol::GetHeader(string HeaderName, int index)
{
	// by default the return value is empty string
	string result = "";


	// we get the value if it exists
	if(HeaderExist(HeaderName, index))
		result = vDatas[index]->mHeaders[HeaderName];
	
	return result;
}


// Sets the data bytes from a file for the given data index 
bool NvProtocol::SetDataFromFile(const char *FileName, int index)
{
	// check if the data index is valid
	if(index+1 > (int)vDatas.size())
		return false;

	// the filename must not be null
	if(FileName == NULL)
		return false;

	// the filename must not be empty
	if(!strlen(FileName))
		return false;

	NvData *Data = vDatas[index];


	// we delete eventual associated data
	if(Data->Data != NULL)
		free(Data->Data);
	Data->DataSize = 0;
	Data->Data = NULL;

	if(Data->DataFile.length())
	{
		// There is an attached file
		// So its only a copy
		string FileNameHere = FileName;
		if(FileNameHere != Data->DataFile)
			if(!CopyFileEx(FileNameHere.c_str(), FileName))
				return false;
		return true;
	}

	// now we open the file
	FILE *File;
	File = Nvfopen(FileName, "rb");
	if(File == NULL)
		return false;

	// we get the file size
	fseek(File, 0, SEEK_END);
	int Length = ftell(File);
	fseek(File, 0, SEEK_SET);

	if(Length == 0)	// file is empty so there is nothing to do
	{
		fclose(File);
		return true;
	}

	// we allocate memory
	Data->Data = (char *)malloc(Length);
	if(Data->Data == NULL)
	{
		fclose(File);
		return false;
	}
	Data->DataSize = Length;

	// we then copy the file bytes to data
	if(fread(Data->Data, 1, Length, File) != Length)
	{
		fclose(File);
		return false;
	}

	fclose(File);
	return true;
}



// NvHttpProtocol CLASS
NvHttpProtocol::NvHttpProtocol()
{
	iMethod = 0;
	sHttpVersion = HTTP_VERSION_STRING;
	bMultipart = false;
	NumCr = 0;
	ReadBuf = NULL;
	ReadBufSize = 0;
	ReadBufStart = 0;
	ReadBufDataSize = 0;
}

NvHttpProtocol::NvHttpProtocol(NvSocket *Socket) : NvProtocol(Socket)
{
	iMethod = 0;
	sHttpVersion = HTTP_VERSION_STRING;
	bMultipart = false;
	NumCr = 0;
}

NvHttpProtocol::~NvHttpProtocol()
{
	Clear();
}


// resets the class to its original state
void NvHttpProtocol::Clear()
{
	iMethod = 0;
	sHttpVersion = HTTP_VERSION_STRING;
	bMultipart = false;
	NumCr = 0;
	sSessionId = "";
	NvProtocol::Clear();	// we reset also the base class

}


//**********************************************************************************
//**********************************************************************************
// reads a complete http message
// returns 0 if time out, 1 if OK and less than 0 if error
// -1 -> socket doesn't exist
// -2 -> data is null
// -3 -> nothing received
// -4 -> no boundary in multipart message
// -5 -> get temporary file error
// negative value is socket error - 10000 if error comes from WaitRead function
// negative value is socket error - 20000 if error comes from ReadData function
//**********************************************************************************
//**********************************************************************************
int NvHttpProtocol::Read(unsigned short TimeOut, bool UseDataFile, int DataFileMinSize, const char *WorkDir)
{
	bMultipart = false;
	NumCr = 0;

	if(sSocket == NULL)
		return -1;

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::1");
#endif
	int Ret = sSocket->WaitRead(TimeOut);
	if(Ret != 1)
		return Ret;


#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::2");
#endif
	Clear();		// resets the message


	// first read initial and header lines
	string line;
	int LineCounter = 0;
	string ParameterName = "";
	string ParameterValue = "";
	string::size_type index;
	bool CountCr = true;
	ReadBuf = (char *)malloc(1024);
	ReadBufSize = 1024;
	ReadBufStart = 0;
	ReadBufDataSize = 0;

	// First data part
	NvData *Data = NVNEW NvData;

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::3");
#endif
	// now, we can add the first data to the message
	// this first data contains only headers
	vDatas.push_back(Data);
	while(1)
	{
		// read the line
		// PM Modif after 2.5.004
		// we must at least read the first line
		// otherwise, it's an error (socket closed by peer)
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::4");
#endif
		if((Ret = ReadLine(line, -1, CountCr)) <= 0)
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::5");
#endif
			free(ReadBuf);
			Clear();
			if(Ret == 0)
				Ret = -3;
			return Ret;		// we leave if error
		}
/*		if((Ret = ReadLine(line, -1, CountCr)) < 0)
		{
			free(ReadBuf);
			Clear();
			return Ret;		// we leave if error
		}
*/
		CleanString(line);

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::6");
#endif
		if(!line.length())		// The line is empty, we have reach the end of headers
		{
			if(ParameterName.length())	// we save the last header
			{
				if(Data->mHeaders.find(ParameterName) ==  Data->mHeaders.end())
					Data->mHeaders[ParameterName] = ParameterValue;
				else
				{
					// The parameter ever exists
					// We must check for the authentication parameter and proceed with priority
					// The priority is the following: Negotiate (WINDOWS only), Kerberos, NTLM, DIGEST, Basic
					if((ParameterName == "WWW-AUTHENTICATE") || (ParameterName == "PROXY-AUTHENTICATE"))
					{
						string ValueHere = ParameterValue;
						MakeUpper(ValueHere);
						CleanString(ValueHere);
						// Get old value
						string OldValue = Data->mHeaders[ParameterName];
						MakeUpper(OldValue);
						CleanString(OldValue);
						int iOldValue = 0;
						if(OldValue.find("BASIC") == 0)
							iOldValue = 1;
						else if(OldValue.find("DIGEST") == 0)
							iOldValue = 2;
						else if(OldValue.find("NTLM") == 0)
							iOldValue = 3;
						else if(OldValue.find("KERBEROS") == 0)
							iOldValue = 4;
						else if(OldValue.find("NEGOTIATE") == 0)
							iOldValue = 5;

						int iNewValue = 0;
						if(ValueHere.find("BASIC") == 0)
							iNewValue = 1;
						else if(ValueHere.find("DIGEST") == 0)
							iNewValue = 2;
						else if(ValueHere.find("NTLM") == 0)
							iNewValue = 3;
						else if(ValueHere.find("KERBEROS") == 0)
							iNewValue = 4;
						else if(ValueHere.find("NEGOTIATE") == 0)
							iNewValue = 5;

						if(iNewValue >= iOldValue)
							Data->mHeaders[ParameterName] = ParameterValue;
					}
				}
			}
			break;			// so we stop reading lines
		}

		LineCounter++;

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::7");
#endif
		// we process the first line
		if(LineCounter == 1)
		{
			SetParameters(line);
			Data->sInfo = line;
			CountCr = false;
			continue;
		}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::8");
#endif
		// Search for header
		if((index = line.find(':')) != string::npos)
		{
			// We have found a new parameter
			// we first save the previous one
			if(ParameterName.length())
			{
				if(Data->mHeaders.find(ParameterName) ==  Data->mHeaders.end())
					Data->mHeaders[ParameterName] = ParameterValue;
				else
				{
					// The parameter ever exists
					// We must check for the authentication parameter and proceed with priority
					// The priority is the following: Negotiate (WINDOWS only), Kerberos, NTLM, DIGEST, Basic
					if((ParameterName == "WWW-AUTHENTICATE") || (ParameterName == "PROXY-AUTHENTICATE"))
					{
						string ValueHere = ParameterValue;
						MakeUpper(ValueHere);
						CleanString(ValueHere);
						// Get old value
						string OldValue = Data->mHeaders[ParameterName];
						MakeUpper(OldValue);
						CleanString(OldValue);
						int iOldValue = 0;
						if(OldValue.find("BASIC") == 0)
							iOldValue = 1;
						else if(OldValue.find("DIGEST") == 0)
							iOldValue = 2;
						else if(OldValue.find("NTLM") == 0)
							iOldValue = 3;
						else if(OldValue.find("KERBEROS") == 0)
							iOldValue = 4;
						else if(OldValue.find("NEGOTIATE") == 0)
							iOldValue = 5;

						int iNewValue = 0;
						if(ValueHere.find("BASIC") == 0)
							iNewValue = 1;
						else if(ValueHere.find("DIGEST") == 0)
							iNewValue = 2;
						else if(ValueHere.find("NTLM") == 0)
							iNewValue = 3;
						else if(ValueHere.find("KERBEROS") == 0)
							iNewValue = 4;
						else if(ValueHere.find("NEGOTIATE") == 0)
							iNewValue = 5;

						if(iNewValue >= iOldValue)
							Data->mHeaders[ParameterName] = ParameterValue;
					}
				}
			}

			// then we get its name and value
			ParameterName = line;
			ParameterName.erase(index);
			ParameterValue = line.substr(index+1);
			CleanString(ParameterName);
			CleanString(ParameterValue);
			MakeUpper(ParameterName);	// the header name is case insensitive
		}
		else if(ParameterName.length())	// if the line contains something else
			ParameterValue += line;		// it is probably the suite of the previous header

	}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::9");
#endif

	// we first try to get the session id from the parameters. This is used only on server side
#ifdef NV_SERVER
	sSessionId = "";
	if(mParameters.find("NV_SESSION_ID") != mParameters.end())
		sSessionId = mParameters["NV_SESSION_ID"];

	// then we try to get the session id from the headers
	if(!sSessionId.length())
		if(Data->mHeaders.find("NV_SESSION_ID") != Data->mHeaders.end())
			sSessionId = Data->mHeaders["NV_SESSION_ID"];
#endif

	NvInt64 bodysize = 0;

	// we first try to get the body size from the parameters
	if(mParameters.find("CONTENT-LENGTH") != mParameters.end())
		bodysize = NvAtoi64(mParameters["CONTENT-LENGTH"].c_str());

	// then we try to get the body size from the headers
	if(Data->mHeaders.find("CONTENT-LENGTH") != Data->mHeaders.end())
		bodysize = NvAtoi64(Data->mHeaders["CONTENT-LENGTH"].c_str());

	
	// now we try to see if the order contains a multipart data
	string boundary;
	if(Data->mHeaders.find("CONTENT-TYPE") != Data->mHeaders.end())
	{
		string multipart = Data->mHeaders["CONTENT-TYPE"];
		if(multipart.find("multipart/form-data") != string::npos)
			bMultipart = true;
	}


#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::10");
#endif
	// If no multipart data, we just have to read the body, push the data and to leave with success
	if(!bMultipart)
	{
		if(UseDataFile && (bodysize >= DataFileMinSize))
		{
			string DirHere = "";
			if(WorkDir != NULL)
				DirHere = WorkDir;
			string FileNameHere;
			if(!GetTemporaryFile(FileNameHere, DirHere, "tmp", "", ""))
			{
				Clear();
				free(ReadBuf);
				return -5;
			}
			if(Data->DataFile.length())
				RemoveFile(Data->DataFile.c_str());
			Data->DataFile = FileNameHere;
			Data->bDeleteDataFile = true;
		}
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::11");
#endif
		Ret = ReadBody(Data, bodysize);
		free(ReadBuf);
		if(Ret != 1)
			Clear();
		return Ret;
	}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::12");
#endif
	
	// we isolate the Content-Type header
	string sContentType = Data->mHeaders["CONTENT-TYPE"];
	

	string sBoundary;
	if((index = sContentType.find("boundary=")) == string::npos)
	{
		Clear();		// this is considered as an error. We must have a boundary
		free(ReadBuf);
		return -4;
	}
	sBoundary = sContentType.substr(index+9);
	
	// we eliminate other parameters
	if((index = sBoundary.find(';')) != string::npos)
		sBoundary.erase(index);

	// we must eliminate double quotes if there are here
	if(sBoundary.length())
	{
		if(sBoundary[0] == '\"')
		{
			sBoundary = sBoundary.substr(1);
			if((index = sBoundary.find('\"')) != string::npos)
				sBoundary.erase(index);
		}
		else
			CleanString(sBoundary);
	}

	if(!sBoundary.length())
	{
		Clear();		// this is considered as an error. We must have a boundary
		free(ReadBuf);
		return -4;
	}

	// the real boundary starts with "--" characters
	sBoundary.insert((string::size_type)0, "--");

	// now we must read all parts and all datas from the socket
	NvInt64 TotalBytesToRead = bodysize;
	NvInt64 TotalBytesRead = 0;
	int BytesRead;
	bool Prolog = true;
	bool Epilog = false;
	bool Middlelog = false;
	bool BoudaryRead = false;

	// We create a new Data
	Data = NVNEW NvData;
	vDatas.push_back(Data);
	while((TotalBytesRead < TotalBytesToRead) || BoudaryRead)
	{
		if(Epilog)
		{
			// if we are in the epilog, we read data character by character
			char c;
			if((Ret = ReadData(&c, 1)) != 1)
			{
				free(ReadBuf);
				Clear();
				return Ret;		// error reading data
			}
			TotalBytesRead += 1;
			continue;
		}

		// first read lines until the next boundary or end of header
		if(!BoudaryRead)
		{
			BytesRead = ReadLine(line);
			if(BytesRead < 0)
			{
				Clear();
				free(ReadBuf);
				return BytesRead;		// we leave if error
			}
			TotalBytesRead += BytesRead;
		}

		CleanString(line);
		
		if(BoudaryRead || (index = line.find(sBoundary)) == 0)
		{
			BoudaryRead = false;
			// we found a boundary
			if(!Prolog)
			{
				// end of a part
				// we create a new one
				Data = NVNEW NvData;
				vDatas.push_back(Data);
			}
			// now we are not any more in the prolog
			Prolog = false;
			// now we are not any more in the middlelog
			Middlelog = false;

			// we must reset these variables for preparing for a new header reading
			ParameterName = "";
			ParameterValue = "";

			// We must test now for the epilog
			string sEpilogBoundary = sBoundary;
			sEpilogBoundary += "--";
			if((index = line.find(sEpilogBoundary)) == 0)
				Epilog = true;				// we are now in the epilog
			else
				Data->sInfo = line;			// otherwise we put the boundary in Data

			continue;	// we must continue now to the end of headers
		}

		// we do nothing while in the prolog
		if(Prolog)
			continue;
		// we do nothing while in the middlelog
		if(Middlelog)
			continue;

		if(line.length())		// The line is not empty, we are always in the headers
		{
			// Search for header
			if((index = line.find(':')) != string::npos)
			{
				// We have found a new parameter
				// we first save the previous one
				if(ParameterName.length())
					if(Data->mHeaders.find(ParameterName) ==  Data->mHeaders.end())
						Data->mHeaders[ParameterName] = ParameterValue;

				// then we get its name and value
				ParameterName = line;
				ParameterName.erase(index);
				ParameterValue = line.substr(index+1);
				CleanString(ParameterName);
				CleanString(ParameterValue);
				MakeUpper(ParameterName);	// the header name is case insensitive
			}
			else if(ParameterName.length())	// if the line contains something else
				ParameterValue += line;		// it is probably the suite of the previous header

			continue;	// we must continue until end of headers
		}

		if(ParameterName.length())	// we save the last header
			if(Data->mHeaders.find(ParameterName) ==  Data->mHeaders.end())
				Data->mHeaders[ParameterName] = ParameterValue;

		// we must reset these variables for preparing for a new header reading
		ParameterName = "";
		ParameterValue = "";

		// now we are not more in the headers

		// We try to get the part body size from the headers
		bodysize = -1;
		if(Data->mHeaders.find("CONTENT-LENGTH") != Data->mHeaders.end())
			bodysize = NvAtoi64(Data->mHeaders["CONTENT-LENGTH"].c_str());

		// if there is a body, we read it
		if(bodysize > 0)
		{
			if(UseDataFile && (bodysize >= (NvInt64)DataFileMinSize))
			{
				string DirHere = "";
				if(WorkDir != NULL)
					DirHere = WorkDir;
				string FileNameHere;
				if(!GetTemporaryFile(FileNameHere, DirHere, "tmp", "", ""))
				{
					Clear();
					free(ReadBuf);
					return -5;
				}
				if(Data->DataFile.length())
					RemoveFile(Data->DataFile.c_str());
				Data->DataFile = FileNameHere;
				Data->bDeleteDataFile = true;
			}
			Ret = ReadBody(Data, bodysize);
			if(Ret != 1)
			{
				free(ReadBuf);
				Clear();
				return Ret;
			}
			TotalBytesRead += bodysize;
		}
		else if(bodysize == -1)
		{
#ifdef NV_SERVER
			bodysize = 0;
			string sBoundaryToFind = "";
			int j;
			for(j=0; j<NumCr; j++)
				sBoundaryToFind += "\r";
			sBoundaryToFind += "\n";
			sBoundaryToFind += sBoundary;
			int SizeBlock = (int)(sBoundaryToFind.length());
			// We don't know if there is a body size
			// So we must read data until the next boundary
			string DirHere = "";
			if(WorkDir != NULL)
				DirHere = WorkDir;
			string FileName;
			char Buf[100];
			sprintf(Buf, "%x_", this);
			if(!GetTemporaryFile(FileName, DirHere, "tmp", Buf, ""))
			{
				Clear();
				free(ReadBuf);
				return -5;
			}
			FILE *File = Nvfopen(FileName.c_str(), "wb");
			if(File == NULL)
			{
				free(ReadBuf);
				RemoveFile(FileName.c_str());
				Clear();
				return -2;
			}

			char *DataToRead = (char *)malloc(SizeBlock+1);
			if(DataToRead == NULL)
			{
				fclose(File);
				RemoveFile(FileName.c_str());
				free(ReadBuf);
				Clear();
				return -2;
			}
			DataToRead[SizeBlock] = '\0';

			char *DataToReadBis = (char *)malloc(SizeBlock+1);
			if(DataToReadBis == NULL)
			{
				fclose(File);
				RemoveFile(FileName.c_str());
				free(DataToRead);
				free(ReadBuf);
				Clear();
				return -2;
			}
			DataToReadBis[SizeBlock] = '\0';
			
			int Offset = 0;
			int SizeDataToRead = SizeBlock;
			while(1)
			{
				SizeDataToRead = SizeBlock-Offset;
				if((Ret = ReadData(DataToRead+Offset, SizeDataToRead)) != SizeDataToRead)
				{
					fclose(File);
					RemoveFile(FileName.c_str());
					free(DataToRead);
					free(DataToReadBis);
					free(ReadBuf);
					Clear();
					return Ret;		// error reading data
				}
				Offset = 0;
				TotalBytesRead += SizeDataToRead;
				// Check here the read data to see if we have found the start of a boundary
				char *Buf = (char *)memchr(DataToRead, '\r', SizeBlock);
				if(Buf == NULL)
					Buf = (char *)memchr(DataToRead, '\n', SizeBlock);
				if(Buf == NULL)
				{
					// The starting character has not been found
					// So we write all what we have read
					if(fwrite(DataToRead, 1, SizeBlock, File) != SizeBlock)
					{
						fclose(File);
						RemoveFile(FileName.c_str());
						free(DataToRead);
						free(DataToReadBis);
						free(ReadBuf);
						Clear();
						return -2;
					}
					continue;
				}
				
				// We have found the starting character of the boundary
				Offset = (int)(Buf-DataToRead);
				if(Offset)
				{
					// We write all until the starting character
					if(fwrite(DataToRead, 1, Offset, File) != Offset)
					{
						fclose(File);
						RemoveFile(FileName.c_str());
						free(DataToRead);
						free(DataToReadBis);
						free(ReadBuf);
						Clear();
						return -2;
					}
				}

				memcpy(DataToReadBis, DataToRead + Offset, SizeBlock-Offset);
				if(Offset)
				{
					if((Ret = ReadData(DataToReadBis+SizeBlock-Offset, Offset)) != Offset)
					{
						fclose(File);
						RemoveFile(FileName.c_str());
						free(DataToRead);
						free(DataToReadBis);
						free(ReadBuf);
						Clear();
						return Ret;		// error reading data
					}
					TotalBytesRead += Offset;
				}

				// Now in DataToReadBis, we may have a complete boundary
				// lets check it
				DataToReadBis[SizeBlock] = '\0';
				if(strcmp(DataToReadBis, sBoundaryToFind.c_str()) != 0)
				{
					// It's not a boundary
					// We must check if find the start boundary character
					Buf = (char *)memchr(DataToReadBis+1, '\r', SizeBlock-1);
					if(Buf == NULL)
						Buf = (char *)memchr(DataToReadBis+1, '\n', SizeBlock-1);
					if(Buf == NULL)
					{
						// We can write all
						if(fwrite(DataToReadBis, 1, SizeBlock, File) != SizeBlock)
						{
							fclose(File);
							RemoveFile(FileName.c_str());
							free(DataToRead);
							free(DataToReadBis);
							free(ReadBuf);
							Clear();
							return -2;
						}
						Offset = 0;
					}
					else
					{
						// The starting character has been found again
						// We write all until it
						Offset = (int)(Buf-DataToReadBis);
						if(Offset)
						{
							// We write all until the starting character
							if(fwrite(DataToReadBis, 1, Offset, File) != Offset)
							{
								fclose(File);
								RemoveFile(FileName.c_str());
								free(DataToRead);
								free(DataToReadBis);
								free(ReadBuf);
								Clear();
								return -2;
							}
						}
						// Then we recopy from the start character to DataToRead
						memcpy(DataToRead, DataToReadBis + Offset, SizeBlock-Offset);
						// The new offset is now the following:
						Offset = SizeBlock-Offset;


					}
					// We continue reading
					if(TotalBytesRead >= TotalBytesToRead)
						break;	// Except if we have read too much (should not occur)
					continue;
				}

				// Yes it's really a boundary
				line = DataToReadBis;
				BoudaryRead = true;


				break;
			}
			
			fclose(File);
			free(DataToRead);
			free(DataToReadBis);

			if(BoudaryRead)
			{
				NvInt64 FileLength = GetFileLength(FileName.c_str());
				bool bAttachedFile = false;
				if(UseDataFile && (FileLength >= DataFileMinSize))
					bAttachedFile = true;
				if(bAttachedFile)
				{
					if(Data->DataFile.length())
						RemoveFile(Data->DataFile.c_str());
					Data->DataFile = FileName;
					Data->bDeleteDataFile = true;
				}
				else
				{
					File = Nvfopen(FileName.c_str(), "rb");
					if(File == NULL)
					{
						RemoveFile(FileName.c_str());
						free(ReadBuf);
						Clear();
						return -2;
					}

					// we get the file size
					// THis is a small file here so we don't need to use large file support
					fseek(File, 0, SEEK_END);
					int Length = ftell(File);
					fseek(File, 0, SEEK_SET);

					if(Length != 0)	// if file is empty, there is nothing to do
					{
						// we allocate memory
						Data->Data = (char *)malloc(Length);
						if(Data->Data == NULL)
						{
							fclose(File);
							RemoveFile(FileName.c_str());
							free(ReadBuf);
							Clear();
							return -2;
						}
						Data->DataSize = Length;

						// we then copy the file bytes to data
						if(fread(Data->Data, 1, Length, File) != Length)
						{
							fclose(File);
							RemoveFile(FileName.c_str());
							free(ReadBuf);
							Clear();
							return -2;
						}
					}

					fclose(File);
				}

				// We must read until next line feed
				int i = 0;
				strcpy(Buf, "");
				for(i=0; i<(sizeof(Buf)-1); i++)
				{
					if((Ret = ReadData(Buf+i, 1)) != 1)
					{
						if(!bAttachedFile)
							RemoveFile(FileName.c_str());
						free(ReadBuf);
						return Ret;		// error reading data
					}
					TotalBytesRead += 1;
					if(TotalBytesRead >= TotalBytesToRead)
					{
						Buf[i+1] = '\0';
						break;
					}
					if(Buf[i] == '\n')
					{
						Buf[i+1] = '\0';
						break;
					}
				}
				line += Buf;
				CleanString(line);
		
				if(!bAttachedFile)
					RemoveFile(FileName.c_str());
			}
#endif
		}

		// we are somewhere where we have to wait for next boundary
		Middlelog = true;
		
	}

	free(ReadBuf);

	// Get the last Data
	// And remove it if there is nothing in sInfo
	Data = vDatas[vDatas.size()-1];
	if(!Data->sInfo.length())
	{
		vDatas.pop_back();
		delete Data;
	}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::Read::13");
#endif
	// now everything has been read
	return 1;

}


// Reads the body part and write it in data memory or file if there is an attached file to the NvData
// Return 1 if OK
int NvHttpProtocol::ReadBody(NvData *Data, NvInt64 bodysize)
{
	int Ret = 1;
	if(bodysize)
	{
		if(Data->DataFile.length())
		{
			FILE *File = Nvfopen(Data->DataFile.c_str(), "wb");
			if(File == NULL)
				return -2;

			NvInt64 TotalBytesRead = 0;
			int BufSize = 4096;
			int SizePacketToRead;
			char *DataBuffer = (char *)malloc(BufSize);
			if(DataBuffer == NULL)
			{
				fclose(File);
				return -2;
			}
			while(1)
			{
				SizePacketToRead = BufSize;
				if((bodysize-TotalBytesRead) < (NvInt64)SizePacketToRead)
					SizePacketToRead = (int)(bodysize-TotalBytesRead);
				if((Ret = ReadData(DataBuffer, SizePacketToRead)) != SizePacketToRead)
				{
					free(DataBuffer);
					fclose(File);
					return Ret;		// error reading data
				}

				if(fwrite(DataBuffer, 1, SizePacketToRead, File) != SizePacketToRead)
				{
					free(DataBuffer);
					fclose(File);
					return -2;		// error reading data
				}

				// we must read all bytes
				TotalBytesRead += (NvInt64)SizePacketToRead;

				if(TotalBytesRead >= bodysize)
					break;
			}
			free(DataBuffer);
			fclose(File);
		}
		else
		{
			Data->Data = (char *)malloc((size_t)bodysize);
			if(Data->Data == NULL)
				return -2;
			Data->DataSize = (unsigned long)bodysize;
			if((Ret = ReadData(Data->Data, Data->DataSize)) != Data->DataSize)
				return Ret;		// error reading data
		}
	}
	else
	{
	
		// No body size given
		// We must check if we have a Transfer-Encoding: chunked header
		bool bChunked = false;
		if(Data->mHeaders.find("TRANSFER-ENCODING") != Data->mHeaders.end())
		{
			string sChunked = Data->mHeaders["TRANSFER-ENCODING"];
			MakeUpper(sChunked);
			if(sChunked.find("CHUNKED") != string::npos)
				bChunked = true;
		}
		// We must check if we have a Connection: close header
		bool bClosed = false;
		if(Data->mHeaders.find("CONNECTION") != Data->mHeaders.end())
		{
			string sClosed = Data->mHeaders["CONNECTION"];
			MakeUpper(sClosed);
			if(sClosed.find("CLOSE") != string::npos)
				bClosed = true;
		}

		if(bChunked)
		{
			// PM Not tested
			string line;

			// Read chunked body
			if(Data->DataFile.length())
			{
				// In fact the data file should never be used since the body size is not know by advance
				FILE *File = Nvfopen(Data->DataFile.c_str(), "wb");
				if(File == NULL)
					return -2;

				// The first line should contain the first block size
				if(ReadLine(line) < 0)
				{
					fclose(File);
					return -2;
				}
				
				// Keep only the digits
				string::size_type index;
				if((index = line.find(';')) != string::npos)
					line = line.substr(0,index);
				int NumBytesToRead = axtoi(line.c_str());
				if(NumBytesToRead <= 0)
				{
					fclose(File);
					return 1;
				}
				NvInt64 TotalBytesRead = 0;
				char *DataBuffer = NULL;
				int DataBufferSize = 0;
				DataBuffer = (char *)malloc(NumBytesToRead);
				if(DataBuffer == NULL)
				{
					fclose(File);
					return -2;
				}
				DataBufferSize = NumBytesToRead;
				
				
				while(NumBytesToRead > 0)
				{
					if(NumBytesToRead > DataBufferSize)
					{
						// realloc needed
						char *NewDataBuffer = (char *)realloc(DataBuffer, NumBytesToRead);
						if(NewDataBuffer == NULL)
						{
							free(DataBuffer);
							fclose(File);
							return -2;
						}
						DataBuffer = NewDataBuffer;
						DataBufferSize = NumBytesToRead;
					}
					// Read data
					if((Ret = ReadData(DataBuffer, NumBytesToRead)) != NumBytesToRead)
					{
						free(DataBuffer);
						fclose(File);
						return Ret;		// error reading data
					}

					if(fwrite(DataBuffer, 1, NumBytesToRead, File) != NumBytesToRead)
					{
						free(DataBuffer);
						fclose(File);
						return -2;		// error reading data
					}
					
					// Append it to body
					TotalBytesRead += (NvInt64)NumBytesToRead;

					// Read the CRLF
					if(ReadLine(line) < 0)
					{
						free(DataBuffer);
						fclose(File);
						return -2;		// error reading data
					}

					// Read nex block size
					if(ReadLine(line) < 0)
					{
						free(DataBuffer);
						fclose(File);
						return -2;		// error reading data
					}
					NumBytesToRead = axtoi(line.c_str());
				}
				free(DataBuffer);
				fclose(File);
			}
			else
			{
				// Read data directly to data buffer

				// The first line should contain the first block size
				if(ReadLine(line) < 0)
					return -2;
				
				// Keep only the digits
				string::size_type index;
				if((index = line.find(';')) != string::npos)
					line = line.substr(0,index);
				int NumBytesToRead = axtoi(line.c_str());
				if(NumBytesToRead <= 0)
					return 1;
				int TotalBytesRead = 0;
				Data->Data = (char *)malloc(NumBytesToRead);
				if(Data->Data == NULL)
					return -2;
				Data->DataSize = NumBytesToRead;
				
				
				while(NumBytesToRead > 0)
				{
					char *NewDataBuffer = (char *)realloc(Data->Data, TotalBytesRead+NumBytesToRead);
					if(NewDataBuffer == NULL)
						return -2;
					Data->Data = NewDataBuffer;
					Data->DataSize = TotalBytesRead+NumBytesToRead;

					// Read data
					if((Ret = ReadData(Data->Data + TotalBytesRead, NumBytesToRead)) != NumBytesToRead)
						return Ret;		// error reading data

					// Append it to body
					TotalBytesRead += NumBytesToRead;

					// Read the CRLF
					if(ReadLine(line) < 0)
						return -2;		// error reading data

					// Read nex block size
					if(ReadLine(line) < 0)
						return -2;		// error reading data
					NumBytesToRead = axtoi(line.c_str());
				}
			}

			// Now we must read until a blank line
			while((ReadLine(line) >= 0) && line.length());
		}	// End chunked

		if(bClosed)
		{
			// We must read until nothing more to read
			if(Data->DataFile.length())
			{
				// In fact the data file should never be used since the body size is not know by advance
				FILE *File = Nvfopen(Data->DataFile.c_str(), "wb");
				if(File == NULL)
					return -2;

				NvInt64 TotalBytesRead = 0;
				int BufSize = 4096;
				char *DataBuffer = (char *)malloc(BufSize);
				if(DataBuffer == NULL)
				{
					fclose(File);
					return -2;
				}
				
				int BytesRead = 0;

				while(BytesRead >= 0)
				{
					BytesRead = ReadData(DataBuffer, BufSize, true);
					if(BytesRead < 0)
					{
						free(DataBuffer);
						fclose(File);
						return BytesRead;		// error reading data
					}
					if(BytesRead == 0)
					{
						// Nothing more to read
						free(DataBuffer);
						fclose(File);
						return 1;
					}

					if(fwrite(DataBuffer, 1, BytesRead, File) != BytesRead)
					{
						free(DataBuffer);
						fclose(File);
						return -2;		// error reading data
					}
					TotalBytesRead += (NvInt64)BytesRead;
				}
				free(DataBuffer);
				fclose(File);
			}
			else
			{

				int TotalBytesRead = 0;
				int BufSize = 500;
				char *DataBuffer = (char *)malloc(BufSize);
				if(DataBuffer == NULL)
					return -2;
				
				int BytesRead = 0;

				while(BytesRead >= 0)
				{
					BytesRead = ReadData(DataBuffer, BufSize, true);
					if(BytesRead < 0)
					{
						free(DataBuffer);
						return BytesRead;		// error reading data
					}
					if(BytesRead == 0)
					{
						// Nothing more to read
						free(DataBuffer);
						return 1;
					}
					if(Data->Data == NULL)
					{
						Data->Data = (char *)malloc(BytesRead);
						if(Data->Data == NULL)
						{
							free(DataBuffer);
							return -2;
						}
						Data->DataSize = BytesRead;
					}
					else
					{
						char *NewBuffer = (char *)realloc(Data->Data, TotalBytesRead+BytesRead);
						if(NewBuffer == NULL)
						{
							free(DataBuffer);
							return -2;
						}
						Data->Data = NewBuffer;
						Data->DataSize = TotalBytesRead+BytesRead;
					}
					memcpy((void *)(Data->Data + TotalBytesRead), DataBuffer, BytesRead);
					TotalBytesRead += BytesRead;
				}
				free(DataBuffer);
			}
		}
	}
	return 1;
}


//**********************************************************************************
//**********************************************************************************
// writes a complete http message
// returns 0 if time out, 1 if OK and less than 0 if error
// -1 -> socket doesn't exist
// -2 -> data is null
// negative value is socket error - 10000 if error comes from WaitWrite function
// negative value is socket error - 20000 if error comes from WriteData function
//**********************************************************************************
//**********************************************************************************
int NvHttpProtocol::Write()
{
	int Ret = 1;

	Ret = sSocket->WaitWrite();
	if(Ret != 1)
		return Ret;

	int i=0;
	char EndOfLine[3];
	strcpy(EndOfLine, "\r\n");
	char HeaderSeparator[3];
	strcpy(HeaderSeparator, ": ");

	// First calculate the length of the message
	NvInt64 MessageSize = 0;
	for(i=0; i<(int)vDatas.size(); i++)
	{
		NvData *Data = vDatas[i];
		MessageSize += (int)(Data->sInfo.length());
		MessageSize += (int)strlen(EndOfLine);
		map<string,string>::iterator pos;
		for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
		{
			// write header name
			MessageSize += (int)(pos->first.length());
			MessageSize += (int)strlen(HeaderSeparator);
			MessageSize += (int)(pos->second.length());
			MessageSize += (int)strlen(EndOfLine);
		}
		MessageSize += (int)strlen(EndOfLine);
		if(Data->DataFile.length())
			MessageSize += GetFileLength(Data->DataFile.c_str());
		else
			MessageSize += (int)(Data->DataSize);
	}

	if(MessageSize < SOCKET_PACKET_SIZE)
	{
		// We send all the message in one time
		char *OutBuffer;
		int OutBufferSize = (int)MessageSize + 10;
		OutBuffer = (char *)malloc(OutBufferSize);
		if(OutBuffer == NULL)
			return -2;

		int RealSize = 0;

		char *buff = OutBuffer;
		int Size = 0;

		for(i=0; i<(int)vDatas.size(); i++)
		{
			NvData *Data = vDatas[i];

			// first write the Info line
			Size = (int)(Data->sInfo.length());
			if(Size)
				memcpy(buff, Data->sInfo.data(), Size);
			buff += Size;
			RealSize += Size;
			
			Size = (int)strlen(EndOfLine);
			if(Size)
				memcpy(buff, EndOfLine, Size);
			buff += Size;
			RealSize += Size;
			
			map<string,string>::iterator pos;
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				// write header name
				Size = (int)(pos->first.length());
				if(Size)
					memcpy(buff, pos->first.data(), Size);
				buff += Size;
				RealSize += Size;
			
				// write header separator
				Size = (int)strlen(HeaderSeparator);
				if(Size)
					memcpy(buff, HeaderSeparator, Size);
				buff += Size;
				RealSize += Size;
			
				// write header value
				Size = (int)(pos->second.length());
				if(Size)
					memcpy(buff, pos->second.data(), Size);
				buff += Size;
				RealSize += Size;
			
				// end line
				Size = (int)strlen(EndOfLine);
				if(Size)
					memcpy(buff, EndOfLine, Size);
				buff += Size;
				RealSize += Size;
			}
			
			
			// end line
			Size = (int)strlen(EndOfLine);
			if(Size)
				memcpy(buff, EndOfLine, Size);
			buff += Size;
			RealSize += Size;

			if(Data->DataFile.length())
			{
				// we open the file
				FILE *File;
				File = Nvfopen(Data->DataFile.c_str(), "rb");
				Size = 0;
				if(File != NULL)
				{
					// we get the file size
					fseek(File, 0, SEEK_END);
					Size = ftell(File);
					fseek(File, 0, SEEK_SET);

					if((NvInt64)(Size + RealSize) > MessageSize)
					{
						// Should not occur but we must realloc
						char *NewBuffer = (char *)realloc(OutBuffer, Size + RealSize + 10);
						if(NewBuffer == NULL)
						{
							fclose(File);
							free(OutBuffer);
							return -2;
						}
						OutBuffer = NewBuffer;
					}

					if(Size != 0)	// file is empty so there is nothing to do
						fread(buff, 1, Size, File);
					fclose(File);
				}
			}
			else
			{
				Size = Data->DataSize;
				if(Size != 0)
					memcpy(buff, Data->Data, Size);
			}
			buff += Size;
			RealSize += Size;
		}
		
		Ret = sSocket->WriteData(OutBuffer, RealSize);
		free(OutBuffer);
		UpdateSessionTimeOut();
	}
	else
	{
		for(i=0; i<(int)vDatas.size(); i++)
		{
			// We bufferize the ouput for the header part
			char *OutBuffer;
			int OutBufferSize = 0;

			NvData *Data = vDatas[i];

			// First calculate the size for OutputBuffer;
			OutBufferSize = (int)(Data->sInfo.length());
			OutBufferSize += (int)strlen(EndOfLine);
			map<string,string>::iterator pos;
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				// write header name
				OutBufferSize += (int)(pos->first.length());
				OutBufferSize += (int)strlen(HeaderSeparator);
				OutBufferSize += (int)(pos->second.length());
				OutBufferSize += (int)strlen(EndOfLine);
			}
			OutBufferSize += (int)strlen(EndOfLine);

			// If data is less than 1024
			// We add it in the unique write
			NvInt64 DataSize = (NvInt64)(Data->DataSize);
			if(Data->DataFile.length())
				DataSize = GetFileLength(Data->DataFile.c_str());
			if(DataSize <= 1024)
				OutBufferSize += (int)DataSize;

			
			OutBuffer = (char *)malloc(OutBufferSize);
			if(OutBuffer == NULL)
				return -2;

			char *buff = OutBuffer;

			// first write the Info line
			int Size = (int)(Data->sInfo.length());
			if(Size)
				memcpy(buff, Data->sInfo.data(), Size);
			buff += Size;
			
			Size = (int)strlen(EndOfLine);
			if(Size)
				memcpy(buff, EndOfLine, Size);
			buff += Size;
			
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				// write header name
				Size = (int)(pos->first.length());
				if(Size)
					memcpy(buff, pos->first.data(), Size);
				buff += Size;
			
				// write header separator
				Size = (int)(strlen(HeaderSeparator));
				if(Size)
					memcpy(buff, HeaderSeparator, Size);
				buff += Size;
			
				// write header value
				Size = (int)(pos->second.length());
				if(Size)
					memcpy(buff, pos->second.data(), Size);
				buff += Size;
			
				// end line
				Size = (int)(strlen(EndOfLine));
				if(Size)
					memcpy(buff, EndOfLine, Size);
				buff += Size;
			}
			
			
			// end line
			Size = (int)strlen(EndOfLine);
			if(Size)
				memcpy(buff, EndOfLine, Size);
			buff += Size;

			if((DataSize != 0) && DataSize <= 1024)
			{
				Size = (int)DataSize;
				if(Data->DataFile.length())
				{
					// we open the file
					FILE *File;
					File = Nvfopen(Data->DataFile.c_str(), "rb");
					if(File != NULL)
					{
						fread(buff, 1, Size, File);
						fclose(File);
					}
				}
				else
				{
					memcpy(buff, Data->Data, Size);
				}
				buff += Size;
			}
			
			Ret = sSocket->WriteData(OutBuffer, OutBufferSize);
			free(OutBuffer);
			UpdateSessionTimeOut();

			NvInt64 Size2 = DataSize;

			if(Size2 > 1024)
			{
				if(Data->DataFile.length())
				{
					// we open the file
					FILE *File;
					File = Nvfopen(Data->DataFile.c_str(), "rb");
					if(File != NULL)
					{
						NvInt64 TotalBytesWrite = 0;
						int BufSize = 4096;
						int SizePacketToWrite;
						char *DataBuffer = (char *)malloc(BufSize);
						if(DataBuffer == NULL)
						{
							fclose(File);
							return -2;
						}
						while(1)
						{
							SizePacketToWrite = BufSize;
							if((Size2-TotalBytesWrite) < SizePacketToWrite)
								SizePacketToWrite = (int)(Size2-TotalBytesWrite);
							fread(DataBuffer, 1, SizePacketToWrite, File);
							Ret = sSocket->WriteData(DataBuffer, SizePacketToWrite);
							if(Ret != 1)
							{
								free(DataBuffer);
								fclose(File);
								return Ret;		// error writting data
							}

							// we must write all bytes
							TotalBytesWrite += SizePacketToWrite;
							UpdateSessionTimeOut();

							if(TotalBytesWrite >= Size2)
								break;
						}
						free(DataBuffer);
						fclose(File);
					}
				}
				else
				{
					Ret = sSocket->WriteData(Data->Data, Data->DataSize);
					if(Ret != 1)
						return Ret;
					UpdateSessionTimeOut();
				}
			}

		}
	}


	return Ret;
}

// This is just a direct write test to discover a bug that was making nirva to loop
int NvHttpProtocol::WriteTest()
{
	int Ret = 1;
	Ret = sSocket->WaitWrite();
	if(Ret != 1)
		return Ret;

	char OutBuffer[] = "POST /nv_app_dm_factory/NVS?command&NV_SESSION_ID=B0548AA598&NVP_CONTEXT_ID=CB358253EA752DEC82B2A6B50605A1F6&NV_XML_XSL_ERROR=Error/maifa_error&NV_CMD=PILOT:RUN:DO_ACTION&NVP_ACTION=valid_check&NVP_VALIDATE=YES HTTP/1.1\r\nAccept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, application/x-shockwave-flash, */*\r\nReferer: http://www.dmfactory.fr/nv_app_dm_factory/NVS?command&NV_CMD=PILOT:RUN:SHOW_CONTEXT&NV_SESSION_ID=B0548AA598&NVP_CONTEXT_ID=CB358253EA752DEC82B2A6B50605A1F6&NV_KEEP_VAR=YES&NV_XML_XSL_ERROR=Error/maifa_error\r\nAccept-Language: fr\r\nContent-Type: multipart/form-data; boundary=---------------------------7da1e715190544\r\nUser-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322)\r\nContent-Length: 259\r\nPragma: no-cache\r\nHost: www.dmfactory.fr\r\nCache-Control: no-cache\r\n\r\n\0";

	Ret = sSocket->WriteData(OutBuffer, (int)(strlen(OutBuffer)));
	UpdateSessionTimeOut();

	return Ret;
}


// returns 0 if time out, 1 if OK and less than 0 if error
// -1 -> socket doesn't exist
// -2 -> data is null
// -3 -> nothing received
// -4 -> no boundary in multipart message
// negative value is socket error - 10000 if error comes from WaitRead function
// negative value is socket error - 20000 if error comes from ReadData function
int NvHttpProtocol::ReadLine(string& line, int max, bool CountCr)
{
	if(sSocket == NULL)
		return -1;

	line = "";
	int nbread = 0;

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::1");
#endif
	char c;
	int Ret;
	int NumTries = 0;
	while(1)
	{
		if(max >= 0)	// stop at max characters if required
			if(max == nbread)
				break;
		// read the line
		if((Ret = ReadData(&c, 1)) != 1)
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::2");
#endif
			if(Ret < 0)
				return Ret;		// error reading data
			if(Ret == 0)
			{
				// If nothing has been read we retry 9 times and we stop
				// Bug found by Maxime (infinite loop in some rare cases)
				if(NumTries == 9)
					return -3;
				NumTries++;
				SLEEP(20);
				continue;
			}
		}
		NumTries = 0;

		nbread++;

		if(c == '\n')
			break;
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::3");
#endif
		line += c;
	}

//	PM Modif, we return the number of bytes really read
//	if(!line.length())
//		return 0;
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::4");
#endif
	if(!line.length())
	{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::5");
#endif
		return nbread;
	}

	while(line.length() && line[line.length()-1] == '\r')
	{
		if(CountCr)
			NumCr++;
		line.resize(line.length()-1);
	}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadLine::6::");
sSocket->WriteLog(line.c_str());
#endif
	return nbread;

}


int NvHttpProtocol::ReadData(char *Data, int iDataLength, bool withtimeout)
{
	if(!iDataLength)	// nothing to read
		return 0;
	int Ret = 0;
	int TimeOutMicro = 0;
	if(withtimeout)
		TimeOutMicro = 1;
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::1");
#endif
	if(ReadBuf != NULL)
	{
		if(ReadBufDataSize >= iDataLength)
		{
			// All the data to get is in the read buffer
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::2");
#endif
			memcpy(Data, ReadBuf+ReadBufStart, iDataLength);
			ReadBufDataSize -= iDataLength;
			ReadBufStart += iDataLength;
			return iDataLength;
		}

		char *RealData = Data;
		int RealDataSize = iDataLength;
		if(ReadBufDataSize)
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::3");
#endif
			// The read buffer is not empty but doesn't contain all the necessary data
			// So we first reset it
			memcpy(Data, ReadBuf+ReadBufStart, ReadBufDataSize);
			Ret += ReadBufDataSize;
			RealData = Data + ReadBufDataSize;
			RealDataSize -= ReadBufDataSize;
			ReadBufDataSize = 0;
			ReadBufStart = 0;
		}
		if(ReadBufSize > RealDataSize)
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::4");
#endif
			// We bufferise
			int TotalBytesRead = 0;
			while(TotalBytesRead < RealDataSize)
			{
				if(withtimeout && (sSocket->WaitRead(0,1) != 1))
					break;	// Nothing more to read
				int NumBytesRead = sSocket->ReadData(ReadBuf+TotalBytesRead, ReadBufSize-TotalBytesRead, true);
				if(NumBytesRead < 0)
					return NumBytesRead;	// Error
				if(NumBytesRead == 0)
					break;	// Nothing more to read (peer socket closed)
				TotalBytesRead += NumBytesRead;
				UpdateSessionTimeOut();
			}
			ReadBufDataSize = TotalBytesRead;
			ReadBufStart = 0;
			int NumBytesToCopy = RealDataSize;
			if(ReadBufDataSize < RealDataSize)
				NumBytesToCopy = ReadBufDataSize;
			memcpy(RealData, ReadBuf, NumBytesToCopy);
			ReadBufDataSize -= NumBytesToCopy;
			ReadBufStart += NumBytesToCopy;
			Ret += NumBytesToCopy;
		}
		else
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::5");
#endif
			if(!withtimeout || (sSocket->WaitRead(0,1) == 1))
			{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::6");
#endif
				int ReallyRead = sSocket->ReadData(RealData, RealDataSize);	// We get data directly from socket
				if(ReallyRead < 0)
					return ReallyRead;	// Error
				Ret += ReallyRead;
				UpdateSessionTimeOut();
			}
		}
	}
	else
	{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::7");
#endif
		if(!withtimeout || (sSocket->WaitRead(0,1) == 1))
		{
#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::8");
#endif
			Ret = sSocket->ReadData(Data, iDataLength);
			UpdateSessionTimeOut();
		}
	}

#ifdef LOG_SOCKET
sSocket->WriteLog("\nNvHttpProtocol::ReadData::9");
#endif
	return Ret;

}


// this returns the size of the body
NvInt64 NvHttpProtocol::GetWriteBodySize()
{
	NvInt64 size = 0;
	int i=0;
	char EndOfLine[3];
	strcpy(EndOfLine, "\r\n");
	char HeaderSeparator[3];
	strcpy(HeaderSeparator, ": ");
	for(i=0; i<(int)vDatas.size(); i++)
	{
		NvData *Data = vDatas[i];
		// first count the first body if there is one
		if(i == 0)
		{
			NvInt64 DataSize = 0;
			DataSize = (NvInt64)(Data->DataSize);
			if(Data->DataFile.length())
				DataSize = GetFileLength(Data->DataFile.c_str());
			size += DataSize;
		}
		else
		{		// then count all the other part information

			// info
			size += (NvInt64)(Data->sInfo.length());
			// end of line
			size += (NvInt64)strlen(EndOfLine);

			// count part header sizes
			map<string,string>::iterator pos;
			for(pos = Data->mHeaders.begin(); pos != Data->mHeaders.end(); ++pos)
			{
				size += (NvInt64)(pos->first.length());
				size += (NvInt64)strlen(HeaderSeparator);
				size += (NvInt64)(pos->second.length());
				size += (NvInt64)strlen(EndOfLine);
			}

			// a blank line
			size += (NvInt64)strlen(EndOfLine);

			// and the part data size
			NvInt64 DataSize = Data->DataSize;
			if(Data->DataFile.length())
				DataSize = GetFileLength(Data->DataFile.c_str());
			size += DataSize;
		}
	}

	return size;
}


// takes the first data and extract parameters from its sInfo member
void NvHttpProtocol::SetParameters(string line)
{
	
	iMethod = HTTP_METHOD_NONE;
	sPath="";
	sService="";
	sCommand="";
	sHttpVersion="";
	mParameters.clear();		// we also remove all parameters
								// but not parts
	CleanString(line);

	// We must replace eventual &amp; strings with &
	string::size_type index;
	while((index = line.find("&amp;")) != string::npos)
		line.replace(index, 5, "&");
	while((index = line.find("&lt;")) != string::npos)
		line.replace(index, 4, "<");
	while((index = line.find("&gt;")) != string::npos)
		line.replace(index, 4, ">");
	while((index = line.find("&apos;")) != string::npos)
		line.replace(index, 6, "'");
	while((index = line.find("&quot;")) != string::npos)
		line.replace(index, 6, "\"");
	
	// First search for method
	if((index = line.find(' ')) != string::npos)
	{
		string provistr = line.substr(0, index);
		MakeUpper(provistr);
		CleanString(provistr);
		if(provistr == "GET")
			iMethod = HTTP_METHOD_GET;
		if(provistr == "PUT")
			iMethod = HTTP_METHOD_PUT;
		if(provistr == "POST")
			iMethod = HTTP_METHOD_POST;
		line.erase(0, ++index);
	}

	// Search for HTTP version
	if((index = line.rfind(' ')) != string::npos)
	{
		sHttpVersion = line.substr(index+1);
		MakeUpper(sHttpVersion);
		CleanString(sHttpVersion);
		line.erase(index);
	}

	// Resolve aliases
	GetAlias(line);
	// Search for service
	if((index = line.find('?')) != string::npos)
	{
		// service requested
		sService = line.substr(0, index);
		DecodeUrlString(sService);
		line.erase(0, ++index);

		// Search for path
		if((index = sService.rfind('/')) != string::npos)
		{
			sPath = sService.substr(0, index+1);
			DecodeUrlString(sPath);
			sService.erase(0, ++index);
		}

		// Search for command
		if((index = line.find('&')) != string::npos)
		{
			sCommand = line.substr(0, index);
			line.erase(0, index);	// we keep the & character
		}
		else
		{
			sCommand = line;	// this is a command with no parameters
			line = "";
		}
		if((index = sCommand.find('=')) != string::npos)
		{
			// The command is in fact a parameter so we must add it to parameters
			string Provistr = "&";
			Provistr += sCommand;
			Provistr += line;
			sCommand = "";
			line = Provistr;
		}
		DecodeUrlString(sCommand);


		// Now search all parameters
		bool bUrlUnicode = true;
		bool bAutoDetectEncoding = false;
#ifdef NV_SERVER
		bAutoDetectEncoding = GetMain()->GetAutoDetectEncoding();
#endif
		while((index = line.rfind('&')) != string::npos)
		{
			string sParameter = line.substr(index+1);
			line.erase(index);
			if((index = sParameter.find('=')) != string::npos)
			{
				string sValue = sParameter.substr(index+1);
				sParameter.erase(index);
				if(sParameter.size())
				{
					DecodeUrlString(sParameter, bUrlUnicode, true, bAutoDetectEncoding);
					DecodeUrlString(sValue, bUrlUnicode);
					MakeUpper(sParameter);	// the parameter name is case insensitive
					if(sParameter == "NV_URL_ENCODING")
					{
						MakeUpper(sValue);
						if(sValue == "ISO-8859-1")
						{
							bUrlUnicode = false;
							bAutoDetectEncoding = false;
						}
						if(sValue == "UTF-8")
						{
							bUrlUnicode = true;
							bAutoDetectEncoding = false;
						}
						if(sValue == "AUTO")
						{
							bAutoDetectEncoding = true;
						}
					}
					mParameters[sParameter] = sValue;
				}
			}
		}
	}
	else
	{
		// direct file requested
		// Search for path
		if((index = line.rfind('/')) != string::npos)
		{
			sPath = line.substr(0, index+1);
			line.erase(0, ++index);
			DecodeUrlString(sPath);
		}
		sCommand = line;	// this is a command with no parameters
		DecodeUrlString(sCommand);
	}

}

void NvHttpProtocol::GetAlias(string& Path)
{
	string PathHere = Path;
	CleanString(PathHere);
	MakeLower(PathHere);
	if(!PathHere.length())
		return;

	map<string,string>::iterator pos;
	string Alias = "";
	string PathInList = "";
	for(pos = Aliases.m_mList.begin(); pos != Aliases.m_mList.end(); ++pos)
	{
		if((PathHere.find(pos->first) == 0) && (pos->second.length() > Alias.length()))
		{
			PathInList = pos->first;
			Alias = pos->second;
		}
	}

	if(!Alias.length())
		return;
	
	string NewPath = Alias;
	if(Path.length() > PathInList.length())
		NewPath += Path.substr(PathInList.length());
	Path = NewPath;
}


void NvHttpProtocol::UpdateSessionTimeOut()
{
#ifdef NV_SERVER
	if(sSessionId.length())
		GetMain()->Sessions.UpdateSessionTimeOut(sSessionId.c_str());
#endif
}
