// ***************************************************************************************
// File: request.h
// Description: declarations of the Nirva client request classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: A request maintains a client context for sending and receiving
//           commands.
// ***************************************************************************************

#ifndef _REQUEST_H
#define _REQUEST_H

#include "task.h"
#include "container.h"
#include "obj.h"
#include "error.h"
#include "command.h"


#include <string>

// close session modes
#define NV_REQUEST_CLOSE_MANUAL					0
#define NV_REQUEST_CLOSE_COMMAND				1
#define NV_REQUEST_CLOSE_SESSION				2



/////////////////////////////////////////////////////////////////////////////
// NvRequest class
class NvRequest
{
public:

	NvError GlobalError;				// Global Error object that has to be tested after each command (except the command that gets the error code)

	std::string m_sServer;				// Server address and port
	std::string m_sServerAddress;		// Used for log to avoid giving proxy password in log or console
	std::string m_sUser;				// User name
	std::string m_sPassword;			// password
	std::string m_sNewPassword;			// optional new password
	std::string m_sNewPasswordConfirm;	// optional new password confirm
	std::string m_sAppli;				// Application name
	std::string m_sOpen;				// Session open procedure
	std::string m_sClose;				// Session close procedure
	std::string m_sHash;				// Used for security control
	bool b_UseSSL;
	bool b_ClientUnicode;				// true when client is unicode
	bool b_Hidden;						// Sends a NV_HIDDEN parameter for each command
	bool bWithNewPassword;				// Do we have to set a new password during login or not
	bool bWithNewPasswordConfirm;		// Set to true when a new password confirm entry has been given
	bool bControlPassword;				// Control the password or not
	bool bSensible;						// Set to true when sensible information must not be send (machine name, user, ...)

	std::string m_sSessionId;			// Session ID. If the Session ID is provided, the order tries to connect the given session
										// The session ID is also updated after an order successfully transmitted to the server
										// and also with successfull return (eventually with error)
	std::string	m_sSocketLogFile;		// Socket log file name
	void *CurrentConnection;			// Current connection
	int m_iAutoClose;					// Control how the distant session is closed
	NvSession *Session;					// Session owning this request
	
	int m_iSessionTimeOut;
	int m_iServerConnectionTimeOut;

	std::vector<void *> m_vCommands;	// command list

	NvRequest();
	virtual ~NvRequest();

	void Clear();

	// Global request command entry point
	bool RequestCommand(NvCommand *Command, char *Buffer = NULL, int BufferSize = 0, int *DataSize = NULL);

	// used for private communication between servers
	bool PrivateCommand(std::string Command, std::string& Result);

	// used for log only (extract the server real address from m_sServer)
	void SetServerAddress();
};


/////////////////////////////////////////////////////////////////////////////
// NvRequestList class
class NvRequestList : public NvObject
{

public:
	std::map<std::string, NvRequest*> m_mRequests;
	
	NvRequestList();
	virtual ~NvRequestList();

	// Removes all requests from the list
	void Clear();
	
	// Creates a new request and add it to the list
	NvRequest* AddRequest(const char* Name, bool Reopen = false);
	
	// Remove the given request from the list
	void RemoveRequest(const char* Name);
	
	// Check if a given request is in the list
	NvRequest* GetRequest(const char* Name);
};




/////////////////////////////////////////////////////////////////////////////
// NvHostList class
class NvHostList : public NvObject
{

public:
	NvObjTable m_tHosts;
	
	NvHostList();
	virtual ~NvHostList();

	// Removes all hosts from the list
	void Clear();

	// Load host list from registry
	void Load();

	// Save host list to registry
	void Save();
	
	// Creates a new host and add it to the list
	bool AddHost(const char* Name, const char* Description, const char* Addresses, const char* Mode = "FAILOVER");
	
	// Remove the given host from the list
	void RemoveHost(const char* Name);
	
	// Return the active address
	std::string GetActiveAddress(const char* Name);
	
	// Return the next address (following the active one or the first if no active address)
	// It then becomes the active one
	std::string GetNextAddress(const char* Name);
	
	// Return the number of addresses
	int GetNumAddresses(const char* Name);

	// Return the mode (FAILOVER or LOAD_BALANCING)
	std::string GetMode(const char* Name);

	// Get the entire host list
	void GetList(NvObjTable *Result);

};



#endif //_REQUEST_H
